(*  title     : An encoding of ITL in Isabelle/HOL
    Authors   : Antonio Cau     <cau.researcher at gmail.com>
                Ben Moszkowski
                Stefan kuhn <stefan.kuhn at dmu.ac.uk>
                James Hoey <jamesbhoey11 at gmail.com>
    Maintainer: Antonio Cau     <cau.researcher at gmail.com>        
    License   : BSD
*)

section \<open>Executability of ITL formulae\<close>

theory Executability
 imports Semantics TimeReversal Fuse 
begin

text \<open>
In this section we will discuss the notion of executability. 
It is used to determine whether an ITL formula represents a ``programming construct''. 
We first formalise the notion of forward executability of a formula which corresponds to 
generating a sequence of states in a particular fashion: we first generate the first state
 and then generate the next until the final state is generated. This sequence of state 
constitutes the behaviour of the system described by the formula. We then investigate the 
reflection of forward executable formula and this requires the introduction of the notion 
of backward executability. This notion corresponds to generating a sequence of states but 
now we first generate the final state  and then generate the previous state until we
 generate  the first state. This sequence corresponds to the reversed behaviour of the
 system described by the formula. Forward and backward executability are related by the 
reflection operator.


\<close>

subsection \<open>Forward Executability\<close>

subsubsection \<open>Common Prefix Value Trace Definitions\<close>


definition sat_d :: "('a :: world) formula \<Rightarrow> bool"
 where "sat_d F \<equiv> (\<exists>s. s \<Turnstile> F)" 

text \<open>
The following definition is a constraint on the intervals which satisfy a formula. 
Only intervals that share a common prefix of the value trace are allowed\<close>


definition commonpfx :: "('a :: world) formula \<Rightarrow> ('a,'b) stfun \<Rightarrow> bool"
where " commonpfx F v \<equiv> 
            (\<forall> s1 s. (s1 \<Turnstile> F) \<and> (s \<Turnstile> F)  \<and> ilen s \<le> ilen s1 \<longrightarrow> 
                 imap v s  =  (imap v (prefix (ilen s) s1))
            )"

text \<open>
The following definition introduces the notion of forward executability. 
\<close>

definition fexec :: "('a :: world) formula \<Rightarrow> ('a,'b) stfun \<Rightarrow> bool"
 where "fexec F v \<equiv> ( (sat_d F) \<and> (commonpfx F v) )"

syntax
 "_sat"        :: "lift \<Rightarrow> bool"                        ("(sat (_))" 5)
 "_commonpfx"  :: "lift \<Rightarrow> lift \<Rightarrow> bool"                ("(\<ddagger>[_]'_(_))" 5)
 "_fexec"      :: "lift \<Rightarrow> lift \<Rightarrow> bool"                ("(\<dagger>[_]'_(_))" 5)

translations
"_sat"       == "CONST sat_d"
"_commonpfx" == "CONST commonpfx" 
"_fexec"     == "CONST fexec" 


subsubsection \<open>Semantic lemmas\<close>

lemma imap_eq_inth_ilen:
 "(imap v s) = (imap v s1) \<longleftrightarrow> ilen s = ilen s1 \<and> (\<forall>i. i\<le>ilen s \<longrightarrow> v (inth s i) = v (inth s1 i))"
using interval_eq_inth_eq[of "(imap v s)" "(imap v s1)"]
by (simp add: inth_imap)

lemma interval_prefix_suffix_help:
 "n \<le> ilen s1 \<Longrightarrow> (suffix (min n (ilen s1)) (prefix n s1)) = \<langle>inth s1 n\<rangle>" 
unfolding min_def
by (metis (full_types) ilast_prefix prefix_ilen_good suffix_ilast)

lemma ExistsilenNhelp: "\<exists>sa. ilen sa = Suc (ilen s)"
by (metis Suc_eq_plus1_left ilen_imap ilen.simps(2))

lemma ExistsilenN:
 " \<exists>s. ilen s = n"
proof (induct n )
case 0
then show ?case by (meson ilen.simps(1))
next
case (Suc n)
then show ?case using ExistsilenNhelp by blast
qed

lemma unithelp1:
 "(v (inth s1 0) = c \<or> v (inth s1 0) = d) \<and> ilen s1 = 0 
   \<Longrightarrow> (imap v s1) = (INil c) \<or> (imap v s1) = (INil d)"
by (simp add: INil_ilen ifirst_imap)

lemma unithelp2:
 " v (inth (INil a) 0) = c \<or> v (inth (INil a) 0) = d \<Longrightarrow> 
   (imap v (INil a)) = (INil c) \<or> (imap v (INil a)) = (INil d)"
by auto

lemma skiphelp1:
 "v (inth s1 0) = c \<and> ilen s1 = 1 \<and> v (inth s1 (Suc 0)) = d \<Longrightarrow> (imap v s1) = (ICons c (INil d))" 
by (metis One_nat_def interval_hd_tail ilen_imap inth_imap suffix_ilast less_numeral_extra(1))

lemma skipnoteqnext:
 "v (inth s1 0) = c \<and> ilen s1 = (Suc n) \<and> v (inth s1 (Suc 0)) = d \<and>
  v (inth s 0) = c \<and> ilen s = (Suc n) \<and> v (inth s (Suc 0)) = d1 \<and> d\<noteq>d1 \<and> s \<noteq> s1 \<Longrightarrow> 
  (imap v s1) \<noteq> (imap v s)"
by (metis inth_imap) 

lemma skipnoteqfirst:
 "v (inth s1 0) = c \<and> ilen s1 = (Suc n) \<and> v (inth s1 (Suc 0)) = d \<and>
  v (inth s 0) = c1 \<and> ilen s = (Suc n) \<and> v (inth s (Suc 0)) = d \<and> c\<noteq>c1 \<and> s \<noteq> s1 \<Longrightarrow> 
  (imap v s1) \<noteq> (imap v s)"
by (metis inth_imap) 

lemma skipnoteqfirstnext:
 "v (inth s1 0) = c \<and> ilen s1 = (Suc n) \<and> v (inth s1 (Suc 0)) = d \<and>
  v (inth s 0) = c1 \<and> ilen s = (Suc n) \<and> v (inth s (Suc 0)) = d1 \<and> c\<noteq>c1 \<and> d \<noteq> d1 \<and> s \<noteq> s1 \<Longrightarrow> 
  (imap v s1) \<noteq> (imap v s)"
by (metis inth_imap) 

lemma skiphelp2:
 "v (inth s1 0) = c \<and> ilen s1 = 1 \<Longrightarrow> (imap v s1) = (ICons c (INil (v (inth s1 1))))"
by (metis interval_hd_tail ilen_imap inth_imap suffix_ilast less_numeral_extra(1))

lemma lenkhelp3:
 "surj v \<Longrightarrow> \<exists>s. v (inth s 0) = c \<and> ilen s =  n"
 by (metis ExistsilenN  inth.simps(1) Suc_eq_plus1_left inth_zero ilen.simps(1) ilen.simps(2) 
     not0_implies_Suc surjD)

lemma lenkhelp4:
 "surj v \<Longrightarrow> 0 < n \<Longrightarrow> \<exists>s. v (inth s 0) = c \<and> v (inth s 1) = d \<and> ilen s =  n"
 by (metis One_nat_def inth_Suc inth_zero ilen.simps(2) lenkhelp3 less_numeral_extra(3) 
     not0_implies_Suc plus_1_eq_Suc)


lemma lenkhelp1:
 "v (inth s1 0) = c \<and> ilen s1 = n \<and> v (inth s1 (Suc 0)) = d \<and> n = 1 \<Longrightarrow> 
    (imap v s1) = (ICons c (INil d))"
  by (metis One_nat_def skiphelp2)

lemma len0help:
assumes "surj v"
shows "  (\<exists>s. v (inth s 0) = c \<and> ilen s =  0 \<and> (\<forall>i. i \<le> ilen s \<longrightarrow> v (inth s i) = v (inth s 0)))" 
proof -
 have 0: "surj v \<Longrightarrow> 
          (\<exists>s. v (inth s 0) = c \<and> ilen s =  0 \<and> (\<forall>i. i \<le> ilen s \<longrightarrow> v (inth s i) = v (inth s 0))) =
          (\<exists>s. v (inth s 0) = c \<and> ilen s = (Suc 0) \<and>  v (inth s 0) = v (inth s 0) )"
      using lenkhelp3[of v c] by (metis  inth.simps(1) ilen.simps(1))
 have 01: "surj v \<Longrightarrow> (\<exists>s. v (inth s 0) = c \<and> ilen s =  (Suc 0) \<and>  v (inth s 0) = v (inth s 0) )" 
    by (simp add: lenkhelp3) 
 show ?thesis  using 0 01 assms by auto
qed

lemma getsmorehelp:
 assumes "\<forall>i<ilen s. v ( inth s (Suc i)) = Suc (v ( inth s i))"
         " v (ifirst s) = 0" 
         "j \<le> ilen s" 
 shows   " inth (imap v s) j = j " 
using assms 
proof (induct j arbitrary: s)
case 0
then show ?case by (simp add: ifirst_imap)
next
case (Suc j)
then show ?case by (simp add: Suc.prems(1) Suc.prems(3) Suc_leD Suc_le_lessD inth_imap)
qed 

lemma initonlyfinhelp:
 assumes "n \<le> ilen s1"
 shows "(\<forall>na\<le>ilen s1 - n. (na = 0) = w (suffix na (prefix na (suffix n s1)))) =
        (w \<langle>inth s1 n\<rangle> \<and> (\<forall>k. n<k \<and> k\<le>ilen s1 \<longrightarrow> \<not>w\<langle>inth s1 k\<rangle>))"
proof -
 have 1: "(\<forall>na\<le>ilen s1 - n. (na = 0) = w (suffix na (prefix na (suffix n s1)))) =
           (\<forall>na . na+n \<le> ilen s1 \<longrightarrow> (na = 0) = w (suffix na (prefix na (suffix n s1))))" 
        by (simp add: Nat.le_diff_conv2 assms)
 have 2: " (\<forall>na . na+n \<le> ilen s1 \<longrightarrow> (na = 0) = w (suffix na (prefix na (suffix n s1)))) = 
          (\<forall>na . na+n \<le> ilen s1 \<longrightarrow> (na = 0) = w (suffix na (suffix n (prefix (na +n) s1)))) " 
      by (simp add: suffix_prefix_swap)      
 have 3: "(\<forall>na . na+n \<le> ilen s1 \<longrightarrow> (na = 0) = w (suffix na (suffix n (prefix (na +n) s1)))) =
          (\<forall>na . na+n \<le> ilen s1 \<longrightarrow> (na = 0) = w (suffix (na +n)  (prefix (na +n) s1)))" 
         by simp
 have 4: "(\<forall>na . na+n \<le> ilen s1 \<longrightarrow> (na = 0) = w (suffix (na +n)  (prefix (na +n) s1))) =
          (\<forall>na . na+n \<le> ilen s1 \<longrightarrow> (na = 0) = w \<langle>inth s1 (na+n)\<rangle>)"
    by (metis interval_prefix_suffix_help min.orderE)      
 have 5: "(\<forall>na . na+n \<le> ilen s1 \<longrightarrow> (na = 0) = w \<langle>inth s1 (na+n)\<rangle>) =
         (\<forall>k. n\<le>k \<and> k \<le> ilen s1 \<longrightarrow> (k-n=0) = w \<langle>inth s1 k\<rangle>)"
   by (metis add_diff_cancel_right' le_add2 le_add_diff_inverse2)  
 have 6: "(\<forall>k. n\<le>k \<and> k \<le> ilen s1 \<longrightarrow> (k-n=0) = w \<langle>inth s1 k\<rangle>) =
          (w \<langle>inth s1 n\<rangle> \<and> (\<forall>k. n<k \<and> k \<le> ilen s1 \<longrightarrow> \<not> w \<langle>inth s1 k\<rangle>))"
          using assms le0 le_eq_less_or_eq less_numeral_extra(3) by auto
 show ?thesis using "1" "2" "4" "5" "6" by auto
qed



subsubsection \<open>Common prefix value trace theorems\<close>


lemma commonpfx_symmetric:
"(\<forall> s1 s. (s1 \<Turnstile> F) \<and> (s \<Turnstile> F)  \<and> ilen s \<le> ilen s1 \<longrightarrow> 
                imap v s  =  (imap v (prefix (ilen s) s1))
            ) =
  (\<forall> s1 s. (s1 \<Turnstile> F) \<and> (s \<Turnstile> F)  \<longrightarrow> 
              (ilen s \<le> ilen s1  \<longrightarrow> imap v s  =  (imap v (prefix (ilen s) s1))) \<and>
              (ilen s1 \<le> ilen s  \<longrightarrow> imap v s1 =  (imap v (prefix (ilen s1) s)))
            ) 
      "
by auto

lemma CommonPfxFixedTrace:
 assumes "basevars v" 
         "sat f \<and> len k" 
         "\<ddagger>[f]_v"
 shows " card { (imap v s) |s. (s \<Turnstile> f) \<and> ilen s = k } = 1" 
proof -
 obtain s0 where 1: "(s0 \<Turnstile> f) \<and> ilen s0 = k "
  by (metis (mono_tags, lifting) assms(2) len_defs sat_d_def unl_lift2) 
 have 3: "\<forall>s1. (s1 \<Turnstile> f) \<and> ilen s1 = k \<longrightarrow> (imap v s1) = (imap v s0)"
   by (metis "1" assms(3) le_cases commonpfx_def)
 have 4: " { (imap v s) |s. (s \<Turnstile> f) \<and> ilen s = k } = {(imap v s0)} "
   using 3 1 by blast
 show ?thesis by (simp add: "4")
qed

lemma NoTrace:
 assumes "basevars v" 
         "\<not>(sat f \<and> len k)" 
 shows " card { (imap v s) |s. (s \<Turnstile> f) \<and> ilen s = k } = 0"
proof -
 have 1: "\<forall> s. \<not>(s \<Turnstile> f) \<or> ilen s \<noteq> k" 
   using assms unfolding sat_d_def len_defs by auto
 have 2: "{ (imap v s) |s. (s \<Turnstile> f) \<and> ilen s = k } = {}"
    using 1 by blast 
 show ?thesis by (simp add: "2")
qed

text \<open>
The following theorem states that the combination of satisfiability with the notion of 
common prefix value trace can be used to determine whether a formula is executable or not,
 i.e., satisfiable and deterministic. 
\<close>

lemma CommonPfxTrace:
 assumes "basevars v" 
         "sat f " 
         "\<ddagger>[f]_v"
 shows " card { (imap v s) |s. (s \<Turnstile> f) \<and> ilen s = k } \<le> 1" 
proof -
 have 1: "(sat f) = (sat (\<exists>k. f \<and> len k)) "
   by (simp add: len_defs sat_d_def)
 have 2: "\<And>k. (sat f \<and> len k) \<or> \<not> (sat f \<and> len k)"
  by simp
 have 3: "\<And>k. \<not> (sat f \<and> len k) \<Longrightarrow> card {imap v s |s. f s \<and> ilen s = k} \<le> 1" 
   using assms  NoTrace[of v f ] by auto
 have 4: "\<And>k. sat f \<and> len k \<Longrightarrow> card {imap v s |s. f s \<and> ilen s = k} \<le> 1" 
  using  assms CommonPfxFixedTrace[of v f] by auto
 show ?thesis using 2 3 4 by auto
qed

lemma PairCommonPfx:
 assumes "basevars (v1, v2)"
         "\<ddagger>[f]_v1"
         "\<ddagger>[f]_v2"
  shows  "\<ddagger>[f]_(v1,v2)" 
 using assms unfolding commonpfx_def 
 by (simp add: imap_eq_inth_ilen)

lemma ProjCommonPfxA:
 assumes "basevars (v1, v2)"
         "\<ddagger>[f]_(v1,v2)"
  shows  "\<ddagger>[f]_v1" 
using assms unfolding commonpfx_def 
by (simp add: imap_eq_inth_ilen)

lemma ProjCommonPfxB:
 assumes "basevars (v1, v2)"
         "\<ddagger>[f]_(v1,v2)"
  shows  "\<ddagger>[f]_v2" 
using assms unfolding commonpfx_def 
by (simp add: imap_eq_inth_ilen)

lemma PairSplitCommonPfx:
 assumes "basevars (v1, v2)"
  shows  "(\<ddagger>[f]_(v1,v2)) = ((\<ddagger>[f]_v1) \<and> (\<ddagger>[f]_v2))" 
using PairCommonPfx ProjCommonPfxA ProjCommonPfxB assms by blast


subsubsection \<open>Common prefix value trace and len and assignment\<close>

lemma SatLenK:
 "sat (len k)"
by (simp add: Valid_def  itl_defs sat_d_def ExistsilenN)

lemma NotCommonPfxLenK:
assumes "basevars (v:: state \<Rightarrow> nat)"
shows "\<not>(\<ddagger>[len(k)]_v)"
proof - 
 have 1: "\<exists> s. v (ifirst s) = 0 \<and> ilen s  = k"
   by (meson assms basevars_def lenkhelp3)
 have 2: "\<exists> s. v (ifirst s) = 1 \<and> ilen s  = k"
   by (meson assms basevars_def lenkhelp3)       
 have 3: " \<not> (\<ddagger>[len k]_v)"
  using assms 1 2 unfolding commonpfx_def len_defs
    by (metis inth_imap prefix_ilen imap_prefix order.order_iff_strict zero_neq_one)
show ?thesis  using 3 by blast
qed 

lemma NotCommonPfxUnitAssignOrAndLenK:
assumes "basevars (v :: state \<Rightarrow> nat)"     
shows "\<not> (\<ddagger>[($v = #0 \<or> $v = #1) \<and> len k]_v)"
using assms lenkhelp3[of v 0 k]  lenkhelp3[of v 1 k]  
by ( simp add: commonpfx_def itl_defs basevars_def inth_imap  )
   (metis One_nat_def inth_imap prefix_ilen_bound prefix_ilen zero_neq_one)

lemma CommonPfxUnitAssignAndNextValAndSkip:
assumes "basevars (v :: state \<Rightarrow> nat)"    
shows " (\<ddagger>[$v = #c \<and> skip \<and> v$=#d]_v)"  
proof -
 have 2: " commonpfx (LIFT($v = #c \<and> skip \<and> v$ = #d)) v"
    using assms  lenkhelp1[of v _ c 1 d]  
      by (simp add: commonpfx_def  next_val_d_def itl_defs basevars_def) 
 show ?thesis using  2  by auto
qed

lemma NotCommonPfxUnitAssignAndLenK:
assumes "basevars (v :: state \<Rightarrow> nat)" 
        " 0 < k"     
shows " \<not> (\<ddagger>[$v = #0 \<and> len k ]_v)" 
using assms using lenkhelp4[of v k 0 0]  lenkhelp4[of v k 0 1]
    by  (simp add: commonpfx_def next_val_d_def itl_defs basevars_def) 
        (metis One_nat_def inth_imap prefix_ilen prefix_ilen_bound zero_neq_one)
   

lemma CommonPfxStableAndLenK:
assumes "basevars v"     
shows "  (\<ddagger>[$v = #(c::nat) \<and> len k \<and> stable v ]_v)" 
using assms by (simp add: commonpfx_def  itl_defs basevars_def)
               (metis  prefix_ilen imap_eq_inth_ilen)

subsubsection \<open>Tempura\<close>

lemma HaltAlways:
 "\<turnstile> (halt f) = \<box>(if\<^sub>i f then empty else (\<not>empty))"
proof -
 have 1: "\<turnstile> (empty=f) = (if\<^sub>i f then empty else (\<not>empty))" 
  unfolding ifthenelse_d_def by auto 
 have 2: " \<turnstile> \<box>(empty=f) = \<box>(if\<^sub>i f then empty else (\<not> empty) )" 
   using 1 BoxEqvBox[of  "LIFT(empty=f)" "LIFT(if\<^sub>i f then empty else (\<not>empty))"] by auto
 show ?thesis unfolding halt_d_def using 2 by auto
qed
 
lemma ChopEqvExistAndLenKChop:
 "\<turnstile> f;g = (\<exists>k. (f \<and> len k);g)"
by (simp add: Valid_def itl_defs)  

lemma ChopEqvExistChopAndLenK:
 "\<turnstile> f;g = (\<exists>k. f;(g \<and> len k))"
by (simp add: Valid_def itl_defs)  

lemma BiSkipImpEqv :
 "\<turnstile> bi (skip \<longrightarrow> f) = (more \<longrightarrow> (f \<and> skip);#True)"
by (auto simp add: Valid_def itl_defs itl_def min_def)

lemma GetsAlways:
 "\<turnstile> a gets b = \<box> (    (bi (skip \<longrightarrow> a$ = b)) )"
by (simp add: Valid_def itl_def)

lemma GetsAlwaysBox:
 "\<turnstile> a gets b = \<box> ( more \<longrightarrow> (a$ = b \<and> skip);#True)"
by (metis BiSkipImpEqv GetsAlways int_eq)

lemma "\<turnstile> bm ( a$ = b ) = \<box>( more \<longrightarrow> a$ = b)"
by (simp add: Valid_def itl_defs itl_def)

lemma "\<turnstile> (a gets $a+#(1::nat))  = (\<box>( more \<longrightarrow> a$ = $a+#1))"
by (auto simp add: Valid_def itl_defs )



subsubsection \<open>Basic theorems\<close> 

lemma CommonPfxEqv:
 assumes "basevars v" 
         "\<turnstile> f = g "
 shows   "(\<ddagger>[f]_v) = (\<ddagger>[g]_v)"
using assms inteq_reflection by blast

lemma CommonPfxAnd:
 assumes "basevars v"
         "\<ddagger>[f]_v"
         "\<ddagger>[g]_v"
 shows   "\<ddagger>[f \<and> g]_v "  
using assms unfolding commonpfx_def by simp

lemma CommonPfxEmptyOrMore:
 assumes "basevars v" 
 shows   "(\<ddagger>[f]_v) = (\<ddagger>[(f \<and> empty) \<or> (f \<and> more)]_v)"
proof -
 have 1: "\<turnstile> f = ((f \<and> empty) \<or> (f \<and> more)) "
   unfolding empty_d_def by fastforce
 show ?thesis using CommonPfxEqv 1 assms by blast
qed

lemma NotCommonPfxEmpty:
assumes "basevars (v :: state \<Rightarrow> nat)"
 shows  "\<not>(\<ddagger>[empty]_v)"
proof -
 obtain s0  where 2: "v (ifirst s0) = 0 \<and> ilen s0 = 0  "
 by (metis  inth.simps(1) assms baseE ilen.simps(1))
 obtain s1  where 3: "v (ifirst s1) = 1 \<and> ilen s1 = 0  "
 by (metis  inth.simps(1) assms baseE ilen.simps(1)) 
 show ?thesis unfolding commonpfx_def empty_defs 
 by (metis "2" "3" One_nat_def inth_imap n_not_Suc_n order.order_iff_strict)
qed    

lemma NotCommonPfxMore:
assumes "basevars (v :: state \<Rightarrow> nat)"
 shows  "\<not>(\<ddagger>[more]_v)"
proof -
  obtain s0  where 2: "v (ifirst s0) = 0 \<and> ilen s0 = 1  "
  by (metis One_nat_def assms baseE inth_zero ilen.simps(1) ilen.simps(2) plus_1_eq_Suc)
  obtain s1  where 3: "v (ifirst s1) = 1 \<and> ilen s1 = 1  "
  by (metis One_nat_def assms baseE inth_zero ilen.simps(1) ilen.simps(2) plus_1_eq_Suc)
  show ?thesis unfolding commonpfx_def more_defs
  by (metis "2" "3" One_nat_def ifirst_imap lessI n_not_Suc_n order_refl)
qed
   
lemma NotCommonPfxTrue:
assumes "basevars (v :: state \<Rightarrow> nat)"
 shows  "\<not>(\<ddagger>[#True]_v)" 
using assms by (metis NotCommonPfxEmpty commonpfx_def unl_con)


lemma CommonPfxChoice:
 assumes "basevars v" 
 shows   "(\<ddagger>[f]_v) = (\<ddagger>[(f \<and> g) \<or> (f \<and> \<not>g)]_v)"
proof -
 have 1: "\<turnstile> f = ((f \<and> g) \<or> (f \<and> \<not>g)) "
   by fastforce
 show ?thesis using CommonPfxEqv 1 assms by blast
qed

lemma CommonPfxDetChoice:
 assumes "basevars v" 
         "\<ddagger>[(f \<and> g) \<or> (f \<and> \<not>g)]_v" 
 shows   "(\<ddagger>[f \<and> g]_v) \<and> (\<ddagger>[(f \<and> \<not>g)]_v)"
using assms unfolding commonpfx_def  by simp

lemma CommonPfxBox:
 assumes "basevars v" 
 shows   "(\<ddagger>[\<box> f]_v) = (\<ddagger>[ f \<and> wnext (\<box> f) ]_v)" 
proof -
 have 1: "\<turnstile> \<box>f = (f \<and> wnext (\<box> f))"
   by (simp add: BoxEqvAndWnextBox) 
 show ?thesis  using assms 1 CommonPfxEqv[of v "LIFT(\<box>f)" "LIFT(f \<and> wnext (\<box> f))"] by simp
qed 

lemma CommonPfxBi:
 assumes "basevars v" 
         "\<ddagger>[f]_v" 
 shows   "\<ddagger>[bi f]_v"
using assms by (simp add: commonpfx_def bi_defs) force

lemma CommonPfxBox1:
 assumes "basevars v" 
         "\<ddagger>[f]_v" 
 shows   "\<ddagger>[\<box> f]_v"
using assms by (simp add: commonpfx_def itl_defs) force

lemma CommonPfxAssign:
 assumes "basevars (v :: state \<Rightarrow> nat)" 
 shows   "\<ddagger>[ $v = #c \<and> v:= #d \<and> skip]_v" 
proof -
 have 1: " \<turnstile> ($v = #c \<and> v:=#d \<and> skip) = ($v = #c \<and> skip \<and> v$=#d)"
      unfolding next_assign_d_def by fastforce
 have 2: " \<ddagger>[$v = #c \<and> skip \<and> v$=#d]_v"
   using CommonPfxUnitAssignAndNextValAndSkip[of v c d] assms by blast
 show ?thesis using assms 1 2 
    CommonPfxEqv[of v "LIFT($v = #c \<and> v:=#d \<and> skip)" "LIFT($v = #c \<and> skip \<and> v$=#d)"] 
   by blast
qed

lemma SatValid:
 assumes "\<turnstile> f"
 shows   "sat f" 
by (simp add: assms intD sat_d_def)

lemma SatValidNotA:
 " (\<turnstile> f) = (\<not>(sat \<not>f))"
unfolding sat_d_def Valid_def by auto

lemma SatValidNotB:
 " (sat f) = (\<not>(\<turnstile> \<not>f)) "
unfolding sat_d_def Valid_def by auto
  
lemma ValidAnd:
 "(\<turnstile> f \<and> g) = ((\<turnstile>f) \<and> (\<turnstile>g))"
unfolding Valid_def by auto

lemma ValidOrA:
 assumes "\<turnstile>f "
 shows   "\<turnstile>f \<or> g"
using assms unfolding Valid_def by auto

lemma ValidOrB:
 assumes "\<turnstile>g "
 shows   "\<turnstile>f \<or> g"
using assms unfolding Valid_def by auto

lemma SatAnd:
 assumes "sat f \<and> g"
 shows   "(sat f)  \<and> (sat g)"
using assms 
by (metis sat_d_def unl_lift2)

lemma SatOr:
 "(sat (f \<or> g)) = ((sat f) \<or> (sat g))" 
unfolding sat_d_def by auto

lemma SatAndLenK:
assumes "sat (f \<and> len k) "
 shows " sat f"
using assms SatAnd by blast

lemma SatChopHaltPrefix:
 assumes "sat (f \<and> halt w)" 
         "sat g"       
         "\<turnstile> init(w) \<longrightarrow> g" 
 shows  "sat (f \<and> halt w);g  "
proof -
 have 1: "\<exists> s. (s \<Turnstile> f) \<and> (\<forall>n\<le>ilen s. (ilen s = n) = w (suffix n s))"
   by (metis (mono_tags, lifting) assms(1) halt_defs sat_d_def unl_lift2)
 obtain s where 2: "(s \<Turnstile> f) \<and> (\<forall>n\<le>ilen s. (ilen s = n) = w (suffix n s))"
    using 1 by auto
 have 3: "w \<langle> inth s (ilen s)\<rangle>"
    using "2" by auto  
 obtain s1 where 4: " (s1 \<Turnstile> g) \<and> (inth s1 0) = inth s (ilen s) "
    by (metis "3"  inth.simps(1) Prop10 assms(3) init_defs inteq_reflection prefix_zero_ifirst 
        unl_lift2)
 have 5: " (fuse s s1) \<Turnstile> (f \<and> halt w);g"
   by (metis "2" "4" halt_defs chop_fuse unl_lift2) 
show ?thesis using "5" sat_d_def by auto
qed

lemma SatChopHaltInitonly:
 assumes "sat f \<and> halt w" 
         "sat g"       
         "\<turnstile> initonly(w) \<longrightarrow> g"      
 shows  "sat (f \<and> halt w);(g \<and> initonly w)  "
proof -
 have 1: "\<exists> s. (s \<Turnstile> f) \<and> (\<forall>n\<le>ilen s. (ilen s = n) = w (suffix n s))"
   by (metis (mono_tags, lifting) assms(1) halt_defs sat_d_def unl_lift2)
 obtain s where 2: "(s \<Turnstile> f) \<and> (\<forall>n\<le>ilen s. (ilen s = n) = w (suffix n s))"
    using 1 by auto
 have 3: "w \<langle> inth s (ilen s)\<rangle>"
    using "2" by auto 
 obtain s1 where 5: "(s1 \<Turnstile> g) \<and> (\<forall>n\<le>ilen s1. (n=0) = w (prefix n s1)) \<and> ifirst s1 = ilast s"
  using assms by (simp add: itl_defs Valid_def sat_d_def) 
   (metis "3" ifirst_suffix ilen.simps(1) le_zero_eq nat_le_linear prefix_ilen suffix_ilen_last) 
 have 6: "w \<langle>inth s1 0\<rangle>"
    using 5 by auto
 have 7: "(fuse s s1) \<Turnstile> (f \<and> halt w);(g \<and> initonly w)"
    using 2 5 3 6 by (auto simp add: itl_defs chop_fuse min_def)
      (metis fuse_ilen_a le_add1 prefix_fuse suffix_fuse suffix_ilen)
 show ?thesis using 7 unfolding sat_d_def  by auto
qed     
 
lemma SatChopL:
 assumes "sat f" 
         "sat g"
         "\<turnstile> f \<longrightarrow> fin w"
         "\<turnstile> init(w) \<longrightarrow> g" 
 shows  "sat f;g  "
using assms 
proof -
 have 1: "\<exists> s. (s \<Turnstile> f)  "
   by (metis (mono_tags, lifting) assms(1)  sat_d_def ) 
 obtain s where 2: "(s \<Turnstile> f) "
    using 1 by auto 
 have 3: " w \<langle> inth s (ilen s)\<rangle>"
    using "2" assms  by (auto simp add: Valid_def fin_defs init_defs)
 obtain s1 where 4: " (s1 \<Turnstile> g) \<and> (inth s1 0) = inth s (ilen s) "
    by (metis "3"  inth.simps(1) Prop10 assms(4) init_defs inteq_reflection prefix_zero_ifirst 
        unl_lift2) 
 have 5: " (fuse s s1) \<Turnstile> (f );g" 
   by (metis "2" "4" chop_fuse )
 show ?thesis using "5" sat_d_def by auto
qed

lemma SatChopR:
 assumes "sat f" 
         "sat g"
         "\<turnstile> fin w \<longrightarrow> f"
         "\<turnstile> g \<longrightarrow> init(w) " 
 shows  "sat f;g  "
proof -
 have 1: "\<exists> s. (s \<Turnstile> g)  "
   by (metis (mono_tags, lifting) assms(2)  sat_d_def ) 
 obtain s where 2: "(s \<Turnstile> g) "
    using 1 by auto
 have 3: " w \<langle> inth s 0\<rangle>"
    using 2 assms by (simp add: Valid_def init_defs)
 obtain s1 where 4: "(s1 \<Turnstile> f) \<and> (inth s1 (ilen s1)) = (inth s 0)"
    using 3 assms(3) by (simp add: Valid_def fin_defs)  (metis ilast_irev) 
 have 5: "(fuse s1 s) \<Turnstile> f;g"
    by (meson "2" "4" chop_fuse) 
 show ?thesis using 5 sat_d_def by auto
qed
 

lemma SatChopExistsL:
 " (sat f;g) = (sat (\<exists>k. (f \<and> len k);g))"
by (metis ChopEqvExistAndLenKChop inteq_reflection) 

lemma SatChopExistsR:
 " (sat f;g) = (sat (\<exists>k. f;(g \<and> len k)))"
by (metis ChopEqvExistChopAndLenK inteq_reflection) 

lemma SatExistsChopL:
 " (sat f;g) = (\<exists>k. (sat  (f \<and> len k);g))"
by (metis (mono_tags, lifting) SatChopExistsL  sat_d_def unl_Rex )

lemma SatExistsChopR:
 " (sat f;g) = (\<exists>k. (sat  f ;(g \<and> len k)))"
by (metis (mono_tags, lifting) SatChopExistsR  sat_d_def unl_Rex )

lemma CommonPfxChopFixed:
 assumes "basevars v" 
         "\<ddagger>[f]_v" 
         "\<ddagger>[g]_v"
         "\<turnstile> f \<longrightarrow> fin w"
         "\<turnstile> init(w) \<longrightarrow> g" 
 shows  "\<ddagger>[ (f \<and> len k);g ]_v "
proof -
 have 4: "\<ddagger>[f \<and> fin w]_v"
  using assms 
  by ( simp add:  itl_defs min_def interval_prefix_suffix_help commonpfx_def split: if_split_asm)  
 have 5: "\<ddagger>[init w \<and> g]_v"
      using assms 
  by ( simp add:  itl_defs min_def interval_prefix_suffix_help commonpfx_def split: if_split_asm)
 have 6:  "\<ddagger>[ (f \<and> len k);g ]_v "
  using assms 4  5 
  proof  (auto simp add:  itl_defs min_def Valid_def interval_prefix_suffix_help commonpfx_def)
    fix s1 :: "state interval"
    fix s  :: "state interval"
    assume a0: "basevars v"
    assume a1: "\<forall>s1 s. f s1 \<and> f s \<and> ilen s \<le> ilen s1 \<longrightarrow> imap v s = imap v (prefix (ilen s) s1) "
    assume a2: "\<forall>s1 s. g s1 \<and> g s \<and> ilen s \<le> ilen s1 \<longrightarrow> imap v s = imap v (prefix (ilen s) s1) "
    assume a3: "\<forall>wa. f wa \<longrightarrow> w \<langle>ilast wa\<rangle> "
    assume a4: "\<forall>wa. w \<langle>ifirst wa\<rangle> \<longrightarrow> g wa "
    assume a5: "ilen s \<le> ilen s1 "
    assume a6: "k \<le> ilen s "
    assume a7: " f (prefix k s1) "
    assume a8: " f (prefix k s) "
    assume a9: " g (suffix k s1) "
    assume a10: " g (suffix k s) "
    show  " imap v s = imap v (prefix (ilen s) s1) "
    proof -
     have 61: "ilen (prefix k s) \<le> ilen (prefix k s1)"
       using a5 by auto 
     have 62: "imap v (prefix k s) = imap v (prefix (ilen (prefix k s)) (prefix k s1))"
       using "61" a1 a7 a8 by blast 
     have 63: "ilen (suffix k s) \<le> ilen (suffix k s1)"
       by (simp add: a5 diff_le_mono) 
     have 64: "imap v (suffix k s) = imap v (prefix (ilen (suffix k s)) (suffix k s1))"
       using "63" a10 a2 a9 by blast 
     have 65: "fuse (prefix k s) (suffix k s) = s"
       by (simp add: a6 fuse_prefix_suffix) 
     have 66: "fuse (prefix k s1) (suffix k s1) = s1" 
       using a5 a6 fuse_prefix_suffix order_trans by blast
     have 67: "imap v (fuse (prefix k s) (suffix k s)) =
               fuse (imap v (prefix k s)) (imap v (suffix k s))"
        by (simp add: a6 fuse_prefix_suffix imap_prefix imap_suffix) 
     have 68: "fuse (imap v (prefix k s)) (imap v (suffix k s)) =
               fuse (imap v (prefix (ilen (prefix k s)) (prefix k s1)))
                    (imap v (prefix (ilen (suffix k s)) (suffix k s1)))" 
        by (simp add: "62" "64")
     have 69: "(fuse (imap v (prefix (ilen (prefix k s)) (prefix k s1)))
                     (imap v (prefix (ilen (suffix k s)) (suffix k s1)))) =
                (fuse (prefix (ilen (prefix k s)) (imap v (prefix k s1)))
                      (prefix (ilen (suffix k s)) (imap v (suffix k s1))))"
             by (metis "61" "63" imap_prefix)
     have 70: "(fuse (prefix k (imap v (prefix k s1)))
                     (prefix (ilen s -k) (imap v (suffix k s1)))) =
               (prefix (ilen (fuse (prefix k s) (suffix k s))) 
                       (fuse (imap v (prefix k s1)) (imap v (suffix k s1))))"
         proof -
          have f1: "\<forall>n na. \<not> (n::nat) \<le> na \<or> min n na = n"
            by (meson min_def_raw)
          have f2: "\<forall>n i. ilen (prefix n (i::state interval)) = min n (ilen i)"
            using prefix_ilen_min by blast
          have f3: "k + ilen (suffix k s) = ilen (fuse (prefix k s) (suffix k s))"
            using f1 by (simp add: a6 fuse_ilen_a)
          have f4: "k \<le> ilen s1"
            by (meson a5 a6 order.trans)
          have "\<forall>n na. \<not> (n::nat) \<le> na \<or> na - n + n = na"
            using le_add_diff_inverse2 by blast
          then have "ilen s - k + k = ilen s"
            by (meson a6)
          then show ?thesis
            using f4 f3 f2 f1 
            by (metis a5 a6 fuse_prefix_suffix ilen_imap imap_prefix 
                imap_suffix pref_pref_3 prefix_ilen suffix_prefix_swap)
         qed
    have 71: "(prefix (ilen (fuse (prefix k s) (suffix k s))) 
                       (fuse (imap v (prefix k s1)) (imap v (suffix k s1)))) =
              (imap v (prefix (ilen s) s1))"
     by (metis a5 a6 fuse_prefix_suffix ilen_imap imap_prefix imap_suffix le_trans) 
    show ?thesis using "65" "67" "68" "69" "70" "71"
    using a6 prefix_ilen_good by fastforce 
     qed   
   qed
 show ?thesis by (simp add: "6")
qed

lemma CommonPfxChopFixedExists:
 assumes "basevars v" 
         "\<ddagger>[f]_v" 
         "\<ddagger>[g]_v"
         "\<turnstile> f \<longrightarrow> fin w"
         "\<turnstile> init(w) \<longrightarrow> g" 
 shows  "(\<exists>k. (\<ddagger>[ (f \<and> len k);g ]_v)) "
using CommonPfxChopFixed assms by blast

lemma  CommonPfxChopExists:
 "(\<ddagger>[(\<exists>k. (f \<and> len k);g)]_v) \<Longrightarrow> (\<exists>k. (\<ddagger>[ (f \<and> len k);g ]_v))  "
unfolding commonpfx_def 
by auto
 
lemma CommonPfxExistsChop:
 "(\<ddagger>[(\<exists>k. (f \<and> len k);g)]_v) = (\<ddagger>[f;g]_v)"
by (metis ChopEqvExistAndLenKChop inteq_reflection)


        
            
subsubsection \<open>Forward Executability Theorems\<close>


lemma FExecStableInitial:
 assumes "basevars v" 
 shows "\<dagger>[$v = #c \<and> stable v]_v  "
proof -
 have 1: "sat ($v = #c \<and> stable v)"
   using assms by (simp add: sat_d_def  commonpfx_def  itl_defs basevars_def) (meson len0help)
 have 2: "\<ddagger>[$v = #c \<and> stable v]_v" 
   using assms by (simp add: min_def sat_d_def commonpfx_def itl_defs basevars_def imap_eq_inth_ilen)
 show ?thesis unfolding fexec_def using 1 2 by auto
qed

lemma FExecAlwaysConstant:
 assumes "basevars v" 
 shows "\<dagger>[\<box>($v = #c )]_v  "
proof -
 have 1: "sat \<box>($v = #c)"
   using assms  by (simp add: sat_d_def commonpfx_def itl_defs basevars_def) (metis len0help)
 have 2: "\<ddagger>[ \<box>($v = #c) ]_v" 
    using assms by (simp add: min_def sat_d_def commonpfx_def itl_defs basevars_def imap_eq_inth_ilen)
show ?thesis unfolding fexec_def using 1 2 by auto
qed
 

lemma FExecGetsAndMore:
assumes "basevars v"
shows   " \<dagger>[$v= #(0::nat) \<and> more \<and> v gets ($v +#1) ]_v"
proof -
 have 10: "sat $v= #0 \<and> more \<and> v gets ($v +#1)"
  using assms by (simp add: sat_d_def itl_defs)
    (metis  inth.simps(1) basevars inth_Suc inth_zero ilen.simps(1) ilen.simps(2) neq0_conv 
       not_less_eq plus_1_eq_Suc)
 have 20: "\<ddagger>[$v= #0 \<and> more \<and> v gets ($v +#1)]_v"
  using assms 
  proof (auto simp add: commonpfx_def itl_defs) 
    show "\<And>s1 s.
       basevars v \<Longrightarrow>
       v (ifirst s1) = 0 \<Longrightarrow>
       \<forall>i<ilen s1. v ( inth s1 (Suc i)) = Suc (v ( inth s1 i)) \<Longrightarrow>
       v (ifirst s) = 0 \<Longrightarrow>
       0 < ilen s \<Longrightarrow>
       \<forall>i<ilen s. v ( inth s (Suc i)) = Suc (v ( inth s i)) \<Longrightarrow>
       ilen s \<le> ilen s1 \<Longrightarrow> imap v s =  (imap v (prefix (ilen s) s1))"
    proof -
     fix s1 
     fix s
     assume a0: "basevars v"
       assume a1: "v (ifirst s1) = 0"  
       assume a2: "\<forall>i<ilen s1. v ( inth s1 (Suc i)) = Suc (v ( inth s1 i))"
       assume a3: "v (ifirst s) = 0"
       assume a4: "0 < ilen s"
       assume a5: "\<forall>i<ilen s. v ( inth s (Suc i)) = Suc (v ( inth s i))"
       assume a6: "ilen s \<le> ilen s1" 
     show "imap v s =  (imap v (prefix (ilen s) s1)) "
      proof -   
       have 1: " ilen(imap v s) = ilen  (imap v ((prefix (ilen s) s1))) "
          using a6 by auto
       have 2: "\<And>j. j \<le> ilen s \<longrightarrow> inth (imap v s) j = inth  (imap v ((prefix (ilen s) s1))) j"
         using a1 a2 a3 a5 a6 getsmorehelp le_trans by (metis imap_prefix inth_prefix)  
       show ?thesis by (metis "1" "2" interval_eq_inth_eq ilen_imap)
      qed
     qed
qed
show ?thesis unfolding fexec_def using 10 20 by simp
qed

lemma FExecGetsAndEmpty:
assumes "basevars v"
shows   " \<dagger>[$v= #(0::nat) \<and> empty \<and> v gets ($v +#1) ]_v"
proof -
 have 10: "sat $v= #0 \<and> empty \<and> v gets ($v +#1)"
  using assms by (simp add: sat_d_def itl_defs)
  (metis baseE ilen.simps(1) inth.simps(1) not_less_zero)
 have 20: "\<ddagger>[$v= #0 \<and> empty \<and> v gets ($v +#1)]_v"
  using assms 
  by (auto simp add: commonpfx_def itl_defs ifirst_imap INil_ilen)
show ?thesis unfolding fexec_def using 10 20 by simp
qed  

lemma FExecChopFixedExists:
 assumes "basevars v" 
         "\<dagger>[f]_v" 
         "\<dagger>[g]_v"
         "\<turnstile> f \<longrightarrow> fin w"
         "\<turnstile> init(w) \<longrightarrow> g" 
 shows  "(\<exists>k. (\<dagger>[ (f \<and> len k);g ]_v)) "
using assms SatChopL[of f  g w] 
unfolding fexec_def 
using SatExistsChopL CommonPfxChopFixed by blast

lemma FExecNextRule:
 assumes "basevars v"
         "\<dagger>[g]_v"
         "\<dagger>[init w \<and> empty]_v"
 shows   "\<dagger>[init w \<and> \<circle> g]_v"
proof -
 obtain s0 where 1: " w \<langle>ifirst s0\<rangle> \<and> ilen s0 = 0" 
  using assms unfolding fexec_def sat_d_def 
    by (auto simp add: itl_defs)
 obtain s1 where 2: " g s1"
  using assms unfolding fexec_def sat_d_def 
    by (auto simp add: itl_defs)
 have 3: " (\<langle>ifirst s0\<rangle>\<ominus> s1) \<Turnstile> init w \<and> \<circle> g" 
   using 1 2 by (auto simp add: itl_defs)
 have 4: "sat init w \<and> \<circle> g" 
  unfolding sat_d_def using 3 by blast
 have 5: "basevars v \<Longrightarrow>
    (sat g) \<and> (\<forall>s1 s. g s1 \<and> g s \<and> ilen s \<le> ilen s1 \<longrightarrow> imap v s = imap v (prefix (ilen s) s1)) \<Longrightarrow>
    (sat init w \<and> empty) \<and>
    (\<forall>s1 s.
        w \<langle>ifirst s1\<rangle> \<and> ilen s1 = 0 \<and> w \<langle>ifirst s\<rangle> \<and> ilen s = 0 \<and> ilen s \<le> ilen s1 \<longrightarrow>
        (\<exists>y. s = \<langle>y\<rangle> \<and> v y = v (ifirst s1))) \<Longrightarrow>
    \<forall>s1 s.
       w \<langle>ifirst s1\<rangle> \<and>
       0 < ilen s1 \<and> g (suffix (Suc 0) s1) \<and> w \<langle>ifirst s\<rangle> \<and> 0 < ilen s \<and> g (suffix (Suc 0) s) \<and> ilen s \<le> ilen s1 \<longrightarrow>
       imap v s = imap v (prefix (ilen s) s1)" 
   proof auto
    fix s1
    fix s 
    assume a0: "basevars v"
    assume a1: "\<forall>s1 s. g s1 \<and> g s \<and> ilen s \<le> ilen s1 \<longrightarrow> imap v s =  (imap v (prefix (ilen s) s1))" 
    assume a3: "\<forall>s1 s. w \<langle>ifirst s1\<rangle> \<and> ilen s1 = 0 \<and> w \<langle>ifirst s\<rangle> \<and> ilen s = 0 \<and> ilen s \<le> ilen s1 
                       \<longrightarrow> (\<exists>y. s = \<langle>y\<rangle> \<and> v y = v (ifirst s1))"
    assume a4: "w \<langle>ifirst s1\<rangle>"
    assume a5: "g (suffix (Suc 0) s1)"
    assume a6: "w \<langle>ifirst s\<rangle>"
    assume a7: "0 < ilen s"
    assume a8: "g (suffix (Suc 0) s)"
    assume a9: "ilen s \<le> ilen s1"
    show "imap v s =  (imap v (prefix (ilen s) s1))"
    proof -
     have 51: "imap v \<langle>ifirst s\<rangle> =  (imap v (prefix (ilen \<langle>ifirst s\<rangle>) \<langle>ifirst s1\<rangle>))"
     by (metis a3 a4 a6 ilast_prefix ilen.simps(1) ilen_gr_zero interval.simps(9) prefix_zero_ifirst) 
     have 52: "imap v (suffix (Suc 0) s) = 
               (imap v ( prefix (ilen (suffix (Suc 0) s)) (suffix (Suc 0) s1)))"
       by (metis a1 a5 a8 a9 diff_le_mono suffix_ilen)
     have 53: "s = (\<langle>ifirst s\<rangle> \<ominus> (suffix (Suc 0) s))"
       using a7 interval_hd_tail by auto 
     have 54: "imap v (\<langle>ifirst s\<rangle> \<ominus> (suffix (Suc 0) s)) =
               (imap v \<langle>ifirst s\<rangle>) \<ominus> (imap v (suffix (Suc 0) s))"
          by simp
     have 55: "(imap v \<langle>ifirst s\<rangle>) \<ominus> (imap v (suffix (Suc 0) s)) =
                (imap v ((prefix (ilen \<langle>ifirst s\<rangle>) \<langle>ifirst s1\<rangle>))) \<ominus>
                (imap v ((prefix (ilen (suffix (Suc 0) s)) (suffix (Suc 0) s1))))"
         using "51" "52" by auto 
     have 56: " (imap v ((prefix (ilen \<langle>ifirst s\<rangle>) \<langle>ifirst s1\<rangle>))) =
                (imap v ((prefix (ilen \<langle>ifirst s\<rangle>) (\<langle>ifirst s1\<rangle> \<ominus> (suffix (Suc 0) s1)))))"
          by auto 
     have 57: " (imap v (prefix (ilen (\<langle>ifirst s\<rangle> \<ominus> (suffix (Suc 0) s))) 
                                (\<langle>ifirst s1\<rangle> \<ominus> (suffix (Suc 0) s1)))) =
                (imap v ((prefix (ilen \<langle>ifirst s\<rangle>) \<langle>ifirst s1\<rangle>))) \<ominus>
                (imap v ((prefix (ilen (suffix (Suc 0) s)) (suffix (Suc 0) s1))))"
        by simp 
     have 58: "imap v (\<langle>ifirst s\<rangle> \<ominus> (suffix (Suc 0) s)) = 
                (imap v (prefix (ilen (\<langle>ifirst s\<rangle> \<ominus> (suffix (Suc 0) s))) 
                                (\<langle>ifirst s1\<rangle> \<ominus> (suffix (Suc 0) s1))))"
        using "55" by auto        
     have 59: " s1 = (\<langle>ifirst s1\<rangle> \<ominus> (suffix (Suc 0) s1)) "
          using a7 a9 interval_hd_tail by force 
     show ?thesis using "51" "52" "53" "54" "57" "59"  by metis
    qed
   qed
 have 6: "\<ddagger>[init w \<and> \<circle> g]_v"
   using assms 5 unfolding commonpfx_def itl_defs Valid_def fexec_def by force
  show ?thesis unfolding fexec_def using 6 4 by auto
qed

text \<open>
The rewriting process of Tempura corresponds to the notion of forward executability.\<close>

lemma FExecWNextRule:
 assumes "basevars v"
         "\<dagger>[g]_v"
         "\<dagger>[init w \<and> empty]_v"
 shows   "\<dagger>[init w \<and> wnext g]_v"
proof -
 have 1: "\<turnstile> (init w \<and> wnext g) = ( (init w \<and> empty) \<or> (init w \<and> \<circle> g))"
       using WnextEqvEmptyOrNext by fastforce 
 have 2: "sat init w \<and> wnext g"
   using 1 SatOr assms FExecNextRule[of v g w] unfolding fexec_def 
      by (metis int_eq)
 have 3: "\<ddagger>[init w \<and> \<circle> g]_v"
    using assms FExecNextRule[of v g w] unfolding fexec_def 
     by blast
 have 4: "\<ddagger>[init w \<and> wnext g]_v" 
     using 1 3 assms StateAndEmptyChop[of w "LIFT(\<circle>g)" ]  unfolding commonpfx_def fexec_def
     proof (auto simp add: itl_defs, metis INil_ilen ilen_gr_zero)
     fix s1 :: "state interval" and s :: "state interval"
     assume a1: "\<forall>s1 s. w \<langle>ifirst s1\<rangle> \<and> 0 < ilen s1 \<and> g (suffix (Suc 0) s1) \<and> w \<langle>ifirst s\<rangle> 
                  \<and> 0 < ilen s \<and> g (suffix (Suc 0) s) \<and> ilen s \<le> ilen s1 \<longrightarrow> 
                     imap v s = imap v (prefix (ilen s) s1)"
     assume a2: "ilen s \<le> ilen s1"
     assume a3: "w \<langle>ifirst s1\<rangle>"
     assume a4: "w \<langle>ifirst s\<rangle>"
     assume a5: "g (suffix (Suc 0) s1)"
     assume a6: "g (suffix (Suc 0) s)"
     assume "\<forall>s1 s. w \<langle>ifirst s1\<rangle> \<and> ilen s1 = 0 \<and> w \<langle>ifirst s\<rangle> \<and> ilen s = 0 \<and> ilen s \<le> ilen s1 \<longrightarrow>
                    (\<exists>y. s = \<langle>y\<rangle> \<and> v y = v (ifirst s1))"
     then have "\<forall>i ia. ((((imap v ia = imap v \<langle>ifirst i\<rangle> \<or> ilen i \<noteq> 0) \<or> ilen ia \<noteq> 0) \<or> 
                \<not> w \<langle>ifirst i\<rangle>) \<or> \<not> w \<langle>ifirst ia\<rangle>) \<or> \<not> ilen ia \<le> ilen i"
     by auto
     then show "imap v s = imap v (prefix (ilen s) s1)"
     using a6 a5 a4 a3 a2 a1 
     by (metis (no_types) ifirst_prefix le0 le_antisym neq0_conv prefix_ilen_good)
     qed  
 show ?thesis unfolding fexec_def using 2 4 by auto
qed

text \<open>
The following lemma states the conditions required to strengthen a forward executable specification.
\<close>

lemma FExecStrengthen:
 assumes "basevars v"
         "sat f \<and> g"
         "\<ddagger>[f]_v"
 shows   "\<dagger>[f \<and> g]_v"
using assms unfolding fexec_def commonpfx_def
by simp


 
      
subsection \<open>Backward Executability\<close>

subsubsection \<open>Common Suffix Value Trace Definitions\<close>

text \<open>
Reflection relates the notion of prefix intervals with that of suffix intervals. 
So we need to introduce the ``mirror image'' of common prefix value traces, i.e. the 
notion of common suffix value trace. 
\<close>

definition commonsfx :: "('a :: world) formula \<Rightarrow> ('a,'b) stfun \<Rightarrow> bool"
where " commonsfx F v \<equiv> 
            (\<forall> s1 s. (s1 \<Turnstile> F) \<and> (s \<Turnstile> F)  \<and> ilen s \<le> ilen s1 \<longrightarrow>
                imap v s  =  (imap v (suffix ((ilen s1) - (ilen s)) s1))
            )"

text \<open>
The notion of common suffix value trace 
corresponds to notion of generating a satisfying interval for a formula but it ''limits'' how 
 this is achieved, i.e., one proceeds in a backward manner. The following definition introduces the
 notion of backward executability. 
\<close>

definition bexec :: "('a :: world) formula \<Rightarrow> ('a,'b) stfun \<Rightarrow> bool"
 where "bexec F v \<equiv> ( (sat_d F) \<and> (commonsfx F v) )"

syntax 
 "_commonsfx"  :: "lift \<Rightarrow> lift \<Rightarrow> bool"                ("(\<natural>[_]'_(_))" 5)
 "_bexec"      :: "lift \<Rightarrow> lift \<Rightarrow> bool"                ("(\<flat>[_]'_(_))" 5)

translations
"_commonsfx" == "CONST commonsfx" 
"_bexec"     == "CONST bexec" 

subsubsection \<open>Common suffix value trace theorems\<close>

lemma CommonSfxFixedTrace:
 assumes "basevars v" 
         "sat f \<and> len k" 
         "\<natural>[f]_v"
 shows " card { (imap v s) |s. (s \<Turnstile> f) \<and> ilen s = k } = 1" 
proof -
 obtain s0 where 1: "(s0 \<Turnstile> f) \<and> ilen s0 = k "
  by (metis (mono_tags, lifting) assms(2) len_defs sat_d_def unl_lift2) 
 have 3: "\<forall>s1. (s1 \<Turnstile> f) \<and> ilen s1 = k \<longrightarrow> (imap v s1) = (imap v s0)"
 using 1 assms unfolding commonsfx_def by (metis le_less)
 have 4: " { (imap v s) |s. (s \<Turnstile> f) \<and> ilen s = k } = {(imap v s0)} "
   using 3 1 by blast
show ?thesis by (simp add: "4")
qed 

text \<open>
The following theorem states that the combination of satisfiability with the notion of common 
suffix value trace can be used to determine whether a formula is deterministic or not, i.e., is
 backward executable or not.
\<close>

lemma CommonSfxTrace:
 assumes "basevars v" 
         "sat f " 
         "\<natural>[f]_v"
 shows " card { (imap v s) |s. (s \<Turnstile> f) \<and> ilen s = k } \<le> 1" 
proof -
 have 1: "(sat f) = (sat (\<exists>k. f \<and> len k)) "
   by (simp add: len_defs sat_d_def)
 have 2: "\<And>k. (sat f \<and> len k) \<or> \<not> (sat f \<and> len k)"
  by simp
 have 3: "\<And>k. \<not> (sat f \<and> len k) \<Longrightarrow> card {imap v s |s. f s \<and> ilen s = k} \<le> 1" 
   using assms  NoTrace[of v f ] by auto
 have 4: "\<And>k. sat f \<and> len k \<Longrightarrow> card {imap v s |s. f s \<and> ilen s = k} \<le> 1" 
  using  assms CommonSfxFixedTrace[of v f] by auto
 show ?thesis using 2 3 4 by auto
qed

lemma PairCommonSfx:
 assumes "basevars (v1, v2)"
         "\<natural>[f]_v1"
         "\<natural>[f]_v2"
  shows  "\<natural>[f]_(v1,v2)" 
 using assms unfolding commonsfx_def 
 by (simp add: imap_eq_inth_ilen)

lemma ProjCommonSfxA:
 assumes "basevars (v1, v2)"
         "\<natural>[f]_(v1,v2)"
  shows  "\<natural>[f]_v1" 
using assms unfolding commonsfx_def 
by (simp add: imap_eq_inth_ilen)

lemma ProjCommonSfxB:
 assumes "basevars (v1, v2)"
         "\<natural>[f]_(v1,v2)"
  shows  "\<natural>[f]_v2" 
using assms unfolding commonsfx_def 
by (simp add: imap_eq_inth_ilen)

lemma PairSplitCommonSfx:
 assumes "basevars (v1, v2)"
  shows  "(\<natural>[f]_(v1,v2)) = ((\<natural>[f]_v1) \<and> (\<natural>[f]_v2))" 
using PairCommonSfx ProjCommonSfxA ProjCommonSfxB assms by blast


subsubsection \<open>Time reversal theorems\<close>

lemma SatRev:
  assumes "basevars v"
  shows    "(sat f) = (sat (f\<^sup>r))"
using assms unfolding  sat_d_def 
by (metis  irev_irev_ident reverse_d_def)

lemma RevCommonPfxSfx:
  assumes "basevars v"
  shows    "(\<natural>[f]_v) = (\<ddagger>[f\<^sup>r]_v)"
using assms unfolding commonpfx_def commonsfx_def reverse_d_def
by (metis (no_types, lifting)  irev_ilen irev_prefix irev_irev_ident irev_imap)

lemma RevCommonSfxPfx:
  assumes "basevars v"
  shows    "(\<ddagger>[f]_v) = (\<natural>[f\<^sup>r]_v)"
using assms 
by (simp add: RevCommonPfxSfx reverse_d_def)



subsubsection \<open>Common suffix value trace and len and assignment\<close>

lemma CommonSfxUnitAssignAndNextValAndSkip:
assumes "basevars (v :: state \<Rightarrow> nat)"    
shows " (\<natural>[$v = #c \<and> skip \<and> v$=#d]_v)"  
proof -
 have 3: "commonsfx (LIFT($v = #c \<and> skip \<and> v$ = #d)) v" 
    using assms  using lenkhelp1[of v _ c 1 d]  
    by (simp add: commonsfx_def  next_val_d_def itl_defs basevars_def) 
 show ?thesis using  3  by auto
qed

lemma NotCommonSfxUnitAssignAndLenK:
assumes "basevars (v :: state \<Rightarrow> nat)" 
        " 0 < k"     
shows " \<not> (\<natural>[$v = #0 \<and> len k ]_v)" 
using assms using lenkhelp4[of v k 0 0]  lenkhelp4[of v k 0 1]  
by  (simp add:  commonsfx_def next_val_d_def itl_defs basevars_def)
    (metis One_nat_def inth_imap order_refl zero_neq_one) 



lemma CommonSfxStableAndLenK:
assumes "basevars v"     
shows "  (\<natural>[$v = #(c::nat) \<and> len k \<and> stable v ]_v)" 
proof -
 have 3: "commonsfx (LIFT($v = #(c::nat) \<and> len k \<and> stable v)) v" 
   using assms by (simp add: commonsfx_def itl_defs basevars_def imap_eq_inth_ilen)
 show ?thesis using 3 by simp
qed


lemma NotCommonSfxGetsAndMore:
assumes "basevars v"
shows   " \<not>(\<natural>[$v= #(0::nat) \<and> more \<and> v gets ($v +#1) ]_v)"
proof -
 have 10: " \<And>k. \<exists>s. v (ifirst s) = 0 \<and> ilen s = k \<and> 
                     (\<forall>i<ilen s. v ( inth s (Suc i)) = Suc (v ( inth s i)))"
  using assms
    proof -
     fix k
     show "basevars v \<Longrightarrow> 
           \<exists>s. v (ifirst s) = 0 \<and> ilen s = k \<and> 
            (\<forall>i<ilen s. v ( inth s (Suc i)) = Suc (v ( inth s i))) "
       proof (induct k )
       case 0
       then show ?case 
       by (metis  inth.simps(1) basevars ilen.simps(1) less_nat_zero_code)
       next
       case (Suc k)
       then show ?case 
          proof -
           have 11: "\<exists>s. v (ifirst s) = 0 \<and> ilen s = k \<and> 
                         (\<forall>i<ilen s. v ( inth s (Suc i)) = Suc (v ( inth s i)))"
             using Suc.hyps Suc.prems by blast 
           obtain ss where 12: "v (ifirst ss) = 0 \<and> ilen ss = k \<and> 
                                (\<forall>i<ilen ss. v ( inth ss (Suc i)) = Suc (v ( inth ss i)))" 
              using 11 by auto
           have 121: " v (inth ss k) = k"
               by (metis "12" getsmorehelp inth_imap order_refl) 
           have 13: "\<exists> s. v (s) = (Suc k)"
                by (simp add: assms basevars)
           obtain ssl where 141: " v ssl = (Suc k)"
              using 13 by auto
           have 14: " v (ifirst (iapp ss (INil ssl))) = 0 \<and> ilen (iapp ss (INil ssl)) = Suc k " 
              by (simp add: "12")
           have 15: "(\<forall>i<ilen (iapp ss (INil ssl)). 
                       v (inth (iapp ss (INil ssl)) (Suc i)) = Suc (v (inth (iapp ss (INil ssl)) i))) =
                     ((\<forall>i<ilen ss. 
                       v (inth (iapp ss (INil ssl)) (Suc i)) = Suc (v (inth (iapp ss (INil ssl)) i))) \<and>
                       v (inth (iapp ss (INil ssl)) (Suc k)) = Suc (v (inth (iapp ss (INil ssl)) k))) "
                 using "14" less_Suc_eq by auto
           have 16: "v (inth (iapp ss (INil ssl)) (Suc k)) = v ssl "
                  by (metis "14" ilast_iapp) 
           have 17: "v (inth (iapp ss (INil ssl)) k) = v (inth ss k)" 
               by (simp add: "12" iapp_inth)
           have 18: "v (inth (iapp ss (INil ssl)) (Suc k)) = Suc (v (inth (iapp ss (INil ssl)) k))"
               by (simp add: "121" "141" "16" "17") 
           have 19: "(\<forall>i<ilen ss. 
                       v (inth (iapp ss (INil ssl)) (Suc i)) = Suc (v (inth (iapp ss (INil ssl)) i)))"
             by (simp add: "12" iapp_inth)          
           show ?thesis 
           using "14" "15" "18" "19" by blast
         qed
       qed
       qed  
 have 20: "\<not> (\<natural>[ $v= #0 \<and> more \<and> v gets ($v +#1) ]_v)"
  using assms 
  proof (simp add: commonsfx_def itl_defs)   
    obtain s0 where 1: "v (ifirst s0) = 0 \<and> 1 = ilen s0 \<and> 
                        (\<forall>i<ilen s0. v (inth s0 (Suc i)) = Suc (v (inth s0 i)))"
      using 10[of 1] 
      by fastforce  
    obtain s1 where 2: "v (ifirst s1) = 0 \<and> 2 = ilen s1 \<and> 
                        (\<forall>i<ilen s1. v (inth s1 (Suc i)) = Suc (v (inth s1 i)))"    
     using 10[of 2] by fastforce
     have 3: "ilen s0 \<le> ilen s1"  
       using 1 2 by auto
     have 4: "imap v s0 \<noteq>  (imap v (suffix (ilen s1 - ilen s0) s1))" 
         by (metis "1" "2" "3" One_nat_def Suc_1 ilast_suffix ilast_imap lessI less_le_trans n_not_Suc_n)
     have 5: " 0 < ilen s0"
       using "1" by linarith 
     have 6: " 0 < ilen s1"
       using "3" "5" order.strict_trans2 by blast
     show "basevars v \<Longrightarrow>
    \<exists>s1. v (ifirst s1) = 0 \<and>
         0 < ilen s1 \<and>
         (\<forall>i<ilen s1. v ( inth s1 (Suc i)) = Suc (v ( inth s1 i))) \<and>
         (\<exists>s. v (ifirst s) = 0 \<and>
              0 < ilen s \<and>
              (\<forall>i<ilen s. v ( inth s (Suc i)) = Suc (v ( inth s i))) \<and>
              ilen s \<le> ilen s1 \<and> imap v s \<noteq>  (imap v (suffix (ilen s1 - ilen s) s1))) "
               using "1" "2" "3" "4" "5" "6" by blast
  qed
 show ?thesis using 20 by auto
qed

lemma CommonSfxGetsAndLenK:
assumes "basevars v"
shows   " (\<natural>[$v= #(0::nat) \<and> len k \<and> v gets ($v +#1) ]_v)"
proof - 
have 10: " \<And>k. \<exists>s. v (ifirst s) = 0 \<and> ilen s = k \<and> 
                   (\<forall>i<ilen s. v ( inth s (Suc i)) = Suc (v ( inth s i)))"
  using assms
    proof -
     fix k
     show "basevars v \<Longrightarrow> 
            \<exists>s. v (ifirst s) = 0 \<and> ilen s = k \<and> 
                (\<forall>i<ilen s. v ( inth s (Suc i)) = Suc (v ( inth s i))) "
       proof (induct k )
       case 0
       then show ?case 
       by (metis  inth.simps(1) basevars ilen.simps(1) less_nat_zero_code)
       next
       case (Suc k)
       then show ?case 
          proof -
           have 11: "\<exists>s. v (ifirst s) = 0 \<and> ilen s = k \<and> 
                         (\<forall>i<ilen s. v ( inth s (Suc i)) = Suc (v ( inth s i)))"
             using Suc.hyps Suc.prems by blast 
           obtain ss where 12: "v (ifirst ss) = 0 \<and> ilen ss = k \<and> 
                                (\<forall>i<ilen ss. v ( inth ss (Suc i)) = Suc (v ( inth ss i)))" 
              using 11 by auto
           have 121: " v (inth ss k) = k"
               by (metis "12" getsmorehelp inth_imap order_refl) 
           have 13: "\<exists> s. v (s) = (Suc k)"
                by (simp add: assms basevars)
           obtain ssl where 141: " v ssl = (Suc k)"
              using 13 by auto
           have 14: " v (ifirst (iapp ss (INil ssl))) = 0 \<and> ilen (iapp ss (INil ssl)) = Suc k " 
              by (simp add: "12")
           have 15: "(\<forall>i<ilen (iapp ss (INil ssl)). 
                       v (inth (iapp ss (INil ssl)) (Suc i)) = Suc (v (inth (iapp ss (INil ssl)) i))) =
                     ((\<forall>i<ilen ss. 
                       v (inth (iapp ss (INil ssl)) (Suc i)) = Suc (v (inth (iapp ss (INil ssl)) i))) \<and>
                       v (inth (iapp ss (INil ssl)) (Suc k)) = Suc (v (inth (iapp ss (INil ssl)) k))) "
                 using "14" less_Suc_eq by auto
           have 16: "v (inth (iapp ss (INil ssl)) (Suc k)) = v ssl "
                  by (metis "14" ilast_iapp) 
           have 17: "v (inth (iapp ss (INil ssl)) k) = v (inth ss k)" 
               by (simp add: "12" iapp_inth)
           have 18: "v (inth (iapp ss (INil ssl)) (Suc k)) = Suc (v (inth (iapp ss (INil ssl)) k))"
               by (simp add: "121" "141" "16" "17") 
           have 19: "(\<forall>i<ilen ss. 
                      v (inth (iapp ss (INil ssl)) (Suc i)) = Suc (v (inth (iapp ss (INil ssl)) i)))"
             by (simp add: "12" iapp_inth)          
           show ?thesis 
           using "14" "15" "18" "19" by blast
         qed
       qed
       qed
have 20 : " (\<natural>[ $v= #0 \<and> len k \<and> v gets ($v +#1) ]_v)"
  using assms 
  by (simp add: commonsfx_def itl_defs imap_eq_inth_ilen inth_imap)
     (metis getsmorehelp inth_imap)
 show ?thesis using 20 by auto
qed

lemma RevCommonPfxStableAndLenK:
assumes "basevars v"     
shows "  (\<ddagger>[($v = #(c::nat) \<and> len k \<and> stable v)\<^sup>r ]_v)" 
using assms 
using RevCommonPfxSfx CommonSfxStableAndLenK by blast

lemma RevCommonSfxStableAndLenK:
assumes "basevars v"     
shows "  (\<natural>[($v = #(c::nat) \<and> len k \<and> stable v)\<^sup>r ]_v)" 
using assms 
using RevCommonSfxPfx CommonPfxStableAndLenK by blast

lemma StableAndLenKRev:
 "\<turnstile> ($v = #(c::nat) \<and> len k \<and> stable v) = ($v = #(c::nat) \<and> len k \<and> stable v)\<^sup>r "
by (simp add: Valid_def itl_defs current_val_d_def reverse_d_def)
   (metis diff_le_self ifirst_irev irev_ilen irev_inth 
    prefix_ilen_bound prefix_ilen irev_irev_ident)


subsubsection \<open>Basic theorems\<close>

lemma RevAndWeakNext:
 "\<turnstile>  (init w \<and> wnext f \<and> fin w1)\<^sup>r = (fin w \<and> wprev (f\<^sup>r) \<and> init w1) "
by (metis (no_types, lifting) RInitEqvFin RRAnd RRWPrevEqvWNext all_rev_eq(3) inteq_reflection)
 
lemma CommonSfxChopFixed:
 assumes "basevars v" 
         "\<natural>[f]_v" 
         "\<natural>[g]_v"
         "\<turnstile> fin w \<longrightarrow> f"
         "\<turnstile> g \<longrightarrow> init w" 
 shows  "\<natural>[ f;(g \<and> len k) ]_v "
proof -
 have 1: " \<ddagger>[g\<^sup>r]_v"
     using assms RevCommonPfxSfx by blast
 have 2: "\<ddagger>[f\<^sup>r]_v " 
     using assms RevCommonPfxSfx by blast
 have 3: "\<turnstile> g\<^sup>r \<longrightarrow> fin w"
   by (metis RImpRule RInitEqvFin assms(5) int_eq) 
 have 4: "\<turnstile> init w \<longrightarrow> f\<^sup>r"
   by (metis RFinEqvInit RImpRule assms(4) inteq_reflection) 
 have 5: "\<ddagger>[(g\<^sup>r \<and> len k);f\<^sup>r]_v" 
  using assms using CommonPfxChopFixed[of v "LIFT(g\<^sup>r)" "LIFT(f\<^sup>r)" w k]
    using "1" "2" "3" "4" by blast
 have 6: "\<turnstile> (g\<^sup>r \<and> len k)\<^sup>r = (g \<and> len k)"
    by (simp add: intI len_defs reverse_d_def)
 have 7: "\<turnstile> ((g\<^sup>r \<and> len k);f\<^sup>r)\<^sup>r = f;(g \<and> len k)"
   by (metis "6" EqvReverseReverse all_rev_eq(13) int_eq)
 have 8: "(\<ddagger>[(g\<^sup>r \<and> len k);f\<^sup>r]_v) = (\<natural>[ f;(g \<and> len k) ]_v)"
    using "7" RevCommonSfxPfx assms(1) int_eq by force
 show ?thesis using "5" "8" by auto
qed
 
lemma CommonSfxEqv:
 assumes "basevars v" 
         "\<turnstile> f = g "
 shows   "(\<natural>[f]_v) = (\<natural>[g]_v)"
using assms inteq_reflection by blast

lemma CommonSfxEmptyOrMore:
 assumes "basevars v" 
 shows   "(\<natural>[f]_v) = (\<natural>[(f \<and> empty) \<or> (f \<and> more)]_v)"
proof -
 have 1: "\<turnstile> f = ((f \<and> empty) \<or> (f \<and> more)) "
   unfolding empty_d_def by fastforce
 show ?thesis using CommonSfxEqv 1 assms by blast
qed

lemma CommonSfxChoice:
 assumes "basevars v" 
 shows   "(\<natural>[f]_v) = (\<natural>[(f \<and> g) \<or> (f \<and> \<not>g)]_v)"
proof -
 have 1: "\<turnstile> f = ((f \<and> g) \<or> (f \<and> \<not>g)) "
   by fastforce
 show ?thesis using CommonSfxEqv 1 assms by blast
qed


lemma CommonSfxDetChoice:
 assumes "basevars v" 
         "\<natural>[(f \<and> g) \<or> (f \<and> \<not>g)]_v" 
 shows   "(\<natural>[f \<and> g]_v) \<and> (\<natural>[(f \<and> \<not>g)]_v)"
using assms unfolding commonsfx_def by simp


lemma CommonSfxBox:
 assumes "basevars v" 
 shows   "(\<natural>[\<box> f]_v) = (\<natural>[ f \<and> wnext (\<box> f) ]_v)" 
proof -
 have 1: "\<turnstile> \<box>f = (f \<and> wnext (\<box> f))"
   by (simp add: BoxEqvAndWnextBox) 
 show ?thesis  using assms 1 CommonSfxEqv[of v "LIFT(\<box>f)" "LIFT(f \<and> wnext (\<box> f))"] by simp
qed


lemma CommonSfxBi:
 assumes "basevars v" 
         "\<natural>[f]_v" 
 shows   "\<natural>[bi f]_v"
using assms by (simp add: commonsfx_def itl_defs) force

lemma CommonSfxBox1:
 assumes "basevars v" 
         "\<natural>[f]_v" 
 shows   "\<natural>[\<box> f]_v"
using assms by (simp add: commonsfx_def itl_defs) force


lemma CommonSfxAssign:
 assumes "basevars (v :: state \<Rightarrow> nat)" 
 shows   "\<natural>[ $v = #c \<and> v:= #d \<and> skip]_v" 
proof -
 have 1: " \<turnstile> ($v = #c \<and> v:=#d \<and> skip) = ($v = #c \<and> skip \<and> v$=#d)"
      unfolding next_assign_d_def by fastforce
 have 2: " \<natural>[$v = #c \<and> skip \<and> v$=#d]_v"
   using CommonSfxUnitAssignAndNextValAndSkip[of v c d] assms by blast
 show ?thesis using assms 1 2 
   CommonSfxEqv[of v "LIFT($v = #c \<and> v:=#d \<and> skip)" "LIFT($v = #c \<and> skip \<and> v$=#d)"] 
   by blast
qed

subsubsection \<open>Backward Executability theorems\<close>

lemma RStableEqvStable:
"\<turnstile> (stable v)\<^sup>r = (stable v)"
proof -
 have 1: "\<turnstile> (v gets $v)\<^sup>r = ba (skip \<longrightarrow> ($v = ($v)\<^sup>r))"
   using RGetsEqvBaSkipImp by blast
 have 2: "\<turnstile> (skip \<longrightarrow> ($v = ($v)\<^sup>r)) = 
            (skip \<longrightarrow> (v$ = $v))"
    by (auto simp add: Valid_def all_rev_eq itl_defs)   
 have 3: "\<turnstile> ba (skip \<longrightarrow> ($v = ($v)\<^sup>r)) =
            ba (skip \<longrightarrow> (v$ = $v))"
   using 2 by (simp add: BaEqvBa)
 have 4: "\<turnstile> ba (skip \<longrightarrow> (v$ = $v)) = (v gets $v)" 
    using BaEqvBtBi GetsAlways by fastforce
 show ?thesis unfolding stable_d_def using "1" "3" "4" by fastforce
qed

lemma RStableInitialEqvStableFin:
 "\<turnstile> ($v = #c \<and> stable v)\<^sup>r = (!v = #c \<and> stable v)"

proof -
have f1: "\<turnstile> (!v = #c) = ($v = #c)\<^sup>r"
by (simp add: all_rev_eq(1) all_rev_eq(3) all_rev_eq(8)) (* 0.5 ms *)
have "\<turnstile>(stable v)\<^sup>r = (stable v)"
by (simp add: RStableEqvStable int_eq) (* 0.4 ms *)
then show ?thesis
using f1 RAnd by fastforce
qed  

lemma StableInitialEqvStableFin:
 "\<turnstile> ($v =#c \<and> stable v) = (!v = #c \<and> stable v)"
by (simp add: Valid_def itl_defs)
   (metis order_refl)
    

lemma BExecStableInitial:
 assumes "basevars v" 
 shows "\<flat>[$v = #c \<and> stable v]_v  "
proof -
 have 1: "sat ($v = #c \<and> stable v)"
   using assms  by (simp add: sat_d_def  commonpfx_def  itl_defs basevars_def)
   (meson len0help)
 have 2: "\<ddagger>[ $v = #c \<and> stable v]_v"
    by (meson FExecStableInitial assms fexec_def)
 have 3: "\<natural>[$v = #c \<and> stable v]_v" 
  using 2 
  by (metis RStableInitialEqvStableFin RevCommonSfxPfx StableInitialEqvStableFin assms int_eq)
 show ?thesis unfolding bexec_def using 1 3 by auto
qed

lemma RAlwaysConstantEqvAlwaysConstant:
 "\<turnstile> (\<box>($v = #c))\<^sup>r = \<box>($v = #c)"
by (auto simp add: Valid_def itl_defs reverse_d_def irev_inth)
   (metis diff_diff_cancel diff_le_self )


lemma BExecAlwaysConstant:
 assumes "basevars v" 
 shows "\<flat>[\<box>($v = #c )]_v  "
proof -
 have 1: "sat \<box>($v = #c)"
   using assms   by (simp add: sat_d_def commonpfx_def itl_defs basevars_def)
      (metis len0help)
 have 2: "\<ddagger>[ \<box>($v = #c) ]_v"
    by (meson FExecAlwaysConstant assms fexec_def)
 have 3: "\<natural>[ \<box>($v = #c) ]_v" 
   by (meson "2" RAlwaysConstantEqvAlwaysConstant RevCommonSfxPfx CommonSfxEqv assms)
show ?thesis unfolding bexec_def using 1 3 by auto
qed


lemma NotBExecGetsAndMore:
assumes "basevars v"
shows   "\<not>( \<flat>[$v= #(0::nat) \<and> more \<and> v gets ($v +#1) ]_v)"
proof -
 have 10: "sat $v= #0 \<and> more \<and> v gets ($v +#1)"
  using assms by (simp add: sat_d_def itl_defs)
    (metis  inth.simps(1) basevars inth_Suc inth_zero ilen.simps(1) ilen.simps(2) neq0_conv 
       not_less_eq plus_1_eq_Suc)
 have 20: "\<not>(\<natural>[$v= #0 \<and> more \<and> v gets ($v +#1)]_v)"
  using assms 
  using NotCommonSfxGetsAndMore by auto
show ?thesis unfolding bexec_def using 10 20 by simp
qed


lemma BExecGetsAndEmpty:
assumes "basevars v"
shows   " \<flat>[$v= #(0::nat) \<and> empty \<and> v gets ($v +#1) ]_v"
proof -
 have 10: "sat $v= #0 \<and> empty \<and> v gets ($v +#1)"
  using assms by (simp add: sat_d_def itl_defs)
     (metis baseE ilen.simps(1) inth.simps(1) not_less_zero)
 have 20: "\<natural>[$v= #0 \<and> empty \<and> v gets ($v +#1)]_v"
  using assms 
  by (auto simp add: commonsfx_def itl_defs imap_eq_inth_ilen )
show ?thesis unfolding bexec_def using 10 20 by simp
qed  


lemma BExecChopFixedExists:
 assumes "basevars v" 
         "\<flat>[f]_v" 
         "\<flat>[g]_v"
         "\<turnstile> fin w \<longrightarrow> f"
         "\<turnstile> g \<longrightarrow> init(w) " 
 shows  "(\<exists>k. (\<flat>[  f;(g \<and> len k) ]_v)) "
using assms SatChopR[of f  g w] 
unfolding bexec_def 
using SatExistsChopR CommonSfxChopFixed 
by metis


lemma BExecPrevRule:
 assumes "basevars v"
         "\<flat>[g]_v"
         "\<flat>[fin w \<and> empty]_v"
 shows   "\<flat>[fin w \<and> prev g]_v"
proof -
 have 1: "\<dagger>[g\<^sup>r]_v"
    using RevCommonPfxSfx SatRev assms fexec_def bexec_def by blast 
 have 2: "\<dagger>[(init w \<and> empty)\<^sup>r]_v" 
    by (metis RAndEmptyEqvAndEmpty REmptyEqvEmpty RInitEqvFin RevCommonPfxSfx all_rev_eq(3) 
        assms(1) assms(3) fexec_def bexec_def inteq_reflection) 
 have 3: "\<dagger>[(init w) \<and> empty]_v"
 using "2" RAndEmptyEqvAndEmpty SatRev CommonPfxEqv assms(1) fexec_def by blast
 have 4: "\<dagger>[(init w) \<and> (\<circle> (g\<^sup>r))]_v"
 by (simp add: "1" "3" FExecNextRule assms(1)) 
 have 5: "\<dagger>[(fin w \<and> prev g)\<^sup>r]_v"
 by (metis "4" RFinEqvInit RPrevEqvNext all_rev_eq(3) inteq_reflection)
 show ?thesis
 using "5" RevCommonPfxSfx SatRev assms(1) fexec_def bexec_def by blast
qed 


text \<open>
In Tempura we have unfortunately no rules for backward execution. But we can define a mirror image 
of lemma FExecWNextRule. So we first generate the last state of the interval and then proceed to 
determine the previous states if there are any. 
\<close>

lemma BExecWPrevRule:
 assumes "basevars v"
         "\<flat>[g]_v"
         "\<flat>[fin w \<and> empty]_v"
 shows   "\<flat>[fin w \<and> wprev g]_v"
proof -
 have 1: "\<turnstile> (fin w \<and> wprev g) = ( (fin w \<and> empty) \<or> (fin w \<and> prev g))"
 using WprevEqvEmptyOrPrev by fastforce     
 have 2: "sat fin w \<and> wprev g"
   using 1 SatOr assms BExecPrevRule[of v g w] unfolding bexec_def 
      by (metis int_eq)
 have 3: "\<natural>[fin w \<and> prev g]_v"
    using assms BExecPrevRule[of v g w] unfolding bexec_def 
     by blast
 have 5: "\<natural>[fin w \<and> wprev g]_v" 
    using assms 1 3 AndFinEqvChopAndEmpty[of "LIFT(prev g)" w]  unfolding bexec_def commonsfx_def
   by (simp add: itl_defs)
   (metis diff_self_eq_0 gr_implies_not0 le_less neq0_conv suffix_ilen suffix_ilen_last suffix_zero)    
 show ?thesis unfolding bexec_def using 2 5 by auto
qed

lemma BExecStrengthen:
 assumes "basevars v"
         "sat f \<and> g"
         "\<natural>[f]_v"
 shows   "\<flat>[f \<and> g]_v"
using assms unfolding bexec_def commonsfx_def
by simp

lemma RevBExecFExec:
 assumes "basevars v"
         "\<turnstile> f = f\<^sup>r"
 shows   "(\<flat>[f]_v) = (\<dagger>[f]_v)"
proof -
 have 1: "(sat f) = (sat f\<^sup>r)"
   using SatRev assms(1) by blast
 have 2: "(\<natural>[f]_v) = (\<ddagger>[f]_v)" 
    using RevCommonPfxSfx assms(1) assms(2) inteq_reflection by force 
 show ?thesis unfolding bexec_def fexec_def using 1 2 by auto
qed

subsection \<open>Reversing bad computations\<close>

text \<open>
The following lemmas gives the conditions necessary to ``undo'' a bad computation.
\<close>

lemma FExecChopLenKRev:
 assumes "basevars v"
         "sat f \<and> len k" 
         "\<ddagger>[f]_v" 
         "\<natural>[f]_v" 
 shows " \<dagger>[ (f \<and> len k); (f\<^sup>r \<and> len k)]_v" 
proof -
 have 1: "\<turnstile> (f \<and> len k)\<^sup>r = (f\<^sup>r \<and> len k)"
    by (simp add: intI len_defs reverse_d_def) 
 have 2: " sat (f\<^sup>r \<and> len k)"
    by (metis "1" SatRev assms(1) assms(2) int_eq)
 obtain s0 where 3: " s0 \<Turnstile> (f \<and> len k)"
     using assms(2) sat_d_def by blast
 obtain s1 where 4: "(s1 \<Turnstile> (f\<^sup>r \<and> len k)) \<and> ilast s0 = ifirst s1 " 
    using "2" 
    by (metis "1" "3" RRAnd all_rev_eq(3) ifirst_irev int_eq reverse_d_def)
 have 5: "fuse s0 s1 \<Turnstile> (f \<and> len k); (f\<^sup>r \<and> len k)" 
    by (meson "3" "4" chop_fuse)
 have 6: "sat  (f \<and> len k); (f\<^sup>r \<and> len k)"
    using "5" sat_d_def by auto
 have 7: "\<ddagger>[f \<and> len k]_v"
    using FExecStrengthen assms fexec_def by blast
 have 8: "\<ddagger>[f\<^sup>r \<and> len k]_v"
   using "2" FExecStrengthen RevCommonPfxSfx assms fexec_def by blast
 have 9: "\<ddagger>[(f \<and> len k);(f\<^sup>r \<and> len k)]_v"
    using 7 8 
  proof (auto simp add:  itl_defs min_def Valid_def reverse_d_def interval_prefix_suffix_help 
         commonpfx_def)
    fix s1 :: "state interval"
    fix s  :: "state interval"
    assume a0: "\<forall>s1 s. f s1 \<and> ilen s1 = k \<and> f s \<and> ilen s = k \<and> ilen s \<le> ilen s1 
                \<longrightarrow> imap v s = imap v (prefix k s1) "
    assume a1: "\<forall>s1 s. f (irev s1) \<and> ilen s1 = k \<and> f (irev s) \<and> ilen s = k \<and> ilen s \<le> ilen s1 
                \<longrightarrow> imap v s = imap v (prefix k s1) "
    assume a2: "ilen s \<le> ilen s1 "
    assume a3: "k \<le> ilen s "
    assume a4: "f (prefix k s1) "
    assume a5: "f (prefix k s) "
    assume a6: "f (irev (suffix k s1))"
    assume a7: "ilen s1 - k = k " 
    assume a8: "f (irev (suffix k s))"
    assume a9: "  ilen s - k = k "
    show "imap v s = imap v (prefix (ilen s) s1) " 
    proof -
     have 91: "imap v (prefix k s) = imap v (prefix k s1) "
          using a0 a3 a4 a5 a7 by force
     have 92: "imap v (suffix k s) = imap v (prefix k (suffix k s1))"
          by (simp add: a1 a6 a7 a8 a9)
     have 93: "fuse (prefix k s) (suffix k s) = s"
         by (simp add: a3 fuse_prefix_suffix) 
     have 94: "fuse (prefix k s1) (suffix k s1) = s1"
         using a2 a3 fuse_prefix_suffix le_trans by blast 
     have 95: "imap v (fuse (prefix k s) (suffix k s)) =
               fuse (imap v (prefix k s)) (imap v (suffix k s)) "
      by (simp add: a3 fuse_prefix_suffix imap_prefix imap_suffix)
     have 96: "fuse (imap v (prefix k s)) (imap v (suffix k s)) =
               fuse (imap v (prefix (ilen (prefix k s)) (prefix k s1)))
                    (imap v (prefix (ilen (suffix k s)) (suffix k s1)))"
       using "91" "92" a3 a9 prefix_ilen_good by fastforce  
     have 97: "(fuse (imap v (prefix (ilen (prefix k s)) (prefix k s1)))
                     (imap v (prefix (ilen (suffix k s)) (suffix k s1)))) =
                (fuse (prefix (ilen (prefix k s)) (imap v (prefix k s1)))
                      (prefix (ilen (suffix k s)) (imap v (suffix k s1))))"
       by (metis "91" a3 a7 a9 imap_prefix order_refl pref_pref prefix_ilen_good suffix_ilen)
     have 98: "(fuse (prefix (ilen (prefix k s)) (imap v (prefix k s1)))
                     (prefix (ilen (suffix k s)) (imap v (suffix k s1)))) =
               (prefix (ilen (fuse (prefix k s) (suffix k s))) 
                       (fuse (imap v (prefix k s1)) (imap v (suffix k s1))))"
       by (metis "91" "92" "95" a7 ilen_imap prefix_ilen suffix_ilen)      
     have 99: "(prefix (ilen (fuse (prefix k s) (suffix k s))) 
                       (fuse (imap v (prefix k s1)) (imap v (suffix k s1)))) =
              (imap v (prefix (ilen s) s1))"
            by (metis a2 a3 fuse_prefix_suffix ilen_imap imap_prefix imap_suffix le_trans) 
     show ?thesis using "93" "95" "96" "97" "98" "99" by auto
    qed
   qed
  show ?thesis by (simp add: "6" "9" fexec_def)
qed

lemma FExecRevChopLenKRev:
 assumes "basevars v"
         "sat f \<and> len k" 
         "\<ddagger>[f]_v" 
         "\<natural>[f]_v" 
 shows " \<dagger>[ (f\<^sup>r \<and> len k); (f \<and> len k)]_v" 
proof -
 have 1: "\<turnstile> (f \<and> len k)\<^sup>r = (f\<^sup>r \<and> len k)"
    by (simp add: intI len_defs reverse_d_def)
 have 2: "sat f\<^sup>r \<and> len k" 
    by (metis "1" SatRev assms(1) assms(2) int_eq)  
 have 3: "\<ddagger>[f\<^sup>r]_v"  
   using RevCommonPfxSfx assms by blast
 have 4: "\<natural>[f\<^sup>r]_v" 
    using RevCommonSfxPfx assms by auto
 show ?thesis using assms 1 2 3 4 
 by (metis EqvReverseReverse FExecChopLenKRev int_eq)
qed

lemma BExecChopLenKRev:
 assumes "basevars v"
         "sat f \<and> len k" 
         "\<ddagger>[f]_v" 
         "\<natural>[f]_v" 
 shows " \<flat>[ (f \<and> len k); (f\<^sup>r \<and> len k)]_v" 
using assms 
proof -
 have 1: "\<turnstile> (f \<and> len k)\<^sup>r = (f\<^sup>r \<and> len k)"
    by (simp add: intI len_defs reverse_d_def) 
 have 2: " sat (f\<^sup>r \<and> len k)"
    by (metis "1" SatRev assms(1) assms(2) int_eq)
  obtain s0 where 3: " s0 \<Turnstile> (f \<and> len k)"
     using assms(2) sat_d_def by blast
 obtain s1 where 4: "(s1 \<Turnstile> (f\<^sup>r \<and> len k)) \<and> ilast s0 = ifirst s1 " 
    using "2" 
    by (metis "1" "3" RRAnd all_rev_eq(3) ifirst_irev int_eq reverse_d_def)
 have 5: "fuse s0 s1 \<Turnstile> (f \<and> len k); (f\<^sup>r \<and> len k)" 
    by (meson "3" "4" chop_fuse)
 have 6: "sat  (f \<and> len k); (f\<^sup>r \<and> len k)"
    using "5" sat_d_def by auto
 have 7: "\<ddagger>[(f \<and> len k);(f\<^sup>r \<and> len k)]_v"
    using assms FExecChopLenKRev[of v f  k] unfolding fexec_def by blast
 have 8: "\<turnstile> ((f \<and> len k);(f\<^sup>r \<and> len k))\<^sup>r = (f \<and> len k);(f\<^sup>r \<and> len k)" 
    by (metis "1" LeftChopEqvChop RRAnd all_rev_eq(13) all_rev_eq(3) inteq_reflection)
 have 9: "\<natural>[(f \<and> len k);(f\<^sup>r \<and> len k)]_v"
   using "7" "8" RevCommonPfxSfx assms(1) int_eq by force
 show ?thesis by (simp add: "6" "9" bexec_def)
qed

lemma BExecRevChopLenKRev:
 assumes "basevars v"
         "sat f \<and> len k" 
         "\<ddagger>[f]_v" 
         "\<natural>[f]_v" 
 shows " \<flat>[ (f\<^sup>r \<and> len k); (f \<and> len k)]_v" 
proof -
have 1: "\<turnstile> (f \<and> len k)\<^sup>r = (f\<^sup>r \<and> len k)"
    by (simp add: intI len_defs reverse_d_def)
 have 2: "sat f\<^sup>r \<and> len k" 
    by (metis "1" SatRev assms(1) assms(2) int_eq)  
 have 3: "\<ddagger>[f\<^sup>r]_v"  
   using RevCommonPfxSfx assms by blast
 have 4: "\<natural>[f\<^sup>r]_v" 
    using RevCommonSfxPfx assms by auto
 show ?thesis using assms 1 2 3 4 
 by (metis BExecChopLenKRev EqvReverseReverse inteq_reflection)
qed




end
