(*  Title     : A Shallow Encoding of ITL in Isabelle/HOL
    Authors   : Antonio Cau     <cau.researcher at gmail.com>
                Ben Moszkowski
                David Smallwood <drs at dmu.ac.uk>
    Maintainer: Antonio Cau     <cau.researcher at gmail.com>        
    License   : BSD
*)

section \<open>Pi operator\<close>   

theory Pi
imports IFilter  UntilSince 
begin
(*
sledgehammer_params [minimize=true,preplay_timeout=10,timeout=60,verbose=true,
                    provers="
  cvc4 z3 e spass vampire
"]
*)

text \<open>
This theory introduces the Pi operator 
\cite{moszkowski15:_applic_tempor_projec_inter_concur, 
      moszkowski17}. The Pi operator is defined in terms of the ifilter operator introduced in 
IFilter.thy. We prove the soundness of the rules and axiom system. The until operator from 
UntilSince.thy is used as there is a striking similarity of the expressiveness of the until and 
the Pi operator \cite{moszkowski17}. 
\<close>

subsection \<open>Definitions\<close>



definition sfxfilt :: "'a interval \<Rightarrow> ('a:: world) formula \<Rightarrow> 'a interval interval"
 where "sfxfilt xs f = (ifilter (\<lambda> ys. ys \<Turnstile>f) (suffixes xs))"

definition pfxfilt :: "'a interval \<Rightarrow> ('a:: world) formula \<Rightarrow> 'a interval interval"
 where "pfxfilt xs f = (ifilter (\<lambda> ys. ys \<Turnstile>f) (prefixes xs))"

definition pifilt :: " 'a interval \<Rightarrow> ('a:: world) formula \<Rightarrow> 'a interval"
 where "pifilt xs f = (imap (\<lambda>s. inth s 0) (sfxfilt xs f))"

definition rpifilt :: " 'a interval \<Rightarrow> ('a:: world) formula \<Rightarrow> 'a interval"
 where "rpifilt xs f = (imap (\<lambda>s. ilast s) (pfxfilt xs f))"

definition pi_d :: "('a:: world) formula \<Rightarrow> 'a formula \<Rightarrow> 'a formula"
where "pi_d F G \<equiv> \<lambda>s. ( (\<exists> i \<le> ilen s. (suffix i s) \<Turnstile> F) \<and> ( (pifilt s F) \<Turnstile> G ) )"

definition rpi_d :: "('a:: world) formula \<Rightarrow> 'a formula \<Rightarrow> 'a formula"
where "rpi_d F G \<equiv> \<lambda>s. ( (\<exists> i \<le> ilen s. (prefix i s) \<Turnstile> F) \<and> ( (rpifilt s F) \<Turnstile> G ) )"

syntax
 "_pi_d"     :: "[lift,lift] \<Rightarrow> lift"        ("(_ \<Pi> _)" [84,84] 83)
 "_rpi_d"    :: "[lift,lift] \<Rightarrow> lift"        ("(_ \<Pi>\<^sup>p _)" [84,84] 83)

syntax (ASCII)
 "_pi_d"     :: "[lift,lift] \<Rightarrow> lift"        ("(_ PI _)" [84,84] 83)
 "_rpi_d"    :: "[lift,lift] \<Rightarrow> lift"        ("(_ RPI _)" [84,84] 83)

translations
 "_pi_d"    \<rightleftharpoons> "CONST pi_d"
 "_rpi_d"   \<rightleftharpoons> "CONST rpi_d"


definition upi_d :: "('a:: world) formula \<Rightarrow> 'a formula \<Rightarrow> 'a formula"
where "upi_d F G \<equiv> LIFT(\<not>(F \<Pi> (\<not> G)))"


syntax
 "_upi_d"     :: "[lift,lift] \<Rightarrow> lift"        ("(_ \<Pi>\<^sup>u _)" [84,84] 83)

syntax (ASCII)
 "_upi_d"     :: "[lift,lift] \<Rightarrow> lift"        ("(_ UPI _)" [84,84] 83)

translations
 "_upi_d"    \<rightleftharpoons> "CONST upi_d"


subsection \<open>Time reversal\<close>


lemma pifilt_irev_exists:
 "(\<exists> k \<le> ilen (irev \<sigma>). (suffix k (irev \<sigma>)) \<Turnstile> f) =
  (\<exists> k \<le> ilen \<sigma>. (prefix k \<sigma>) \<Turnstile> f\<^sup>r)"
by (simp add: reverse_d_def)
 (metis diff_diff_cancel irev_prefix suffix_ilen_bound suffix_ilen)

lemma PiRevsem:
 "(\<sigma> \<Turnstile> (f \<Pi> g)\<^sup>r) = (\<sigma> \<Turnstile> f\<^sup>r \<Pi>\<^sup>p g\<^sup>r)" 
proof -
 have 1: "(\<sigma> \<Turnstile> (f \<Pi> g)\<^sup>r) =
          ((\<exists>i\<le>ilen \<sigma>. f (suffix i (irev \<sigma>))) \<and> g (pifilt (irev \<sigma>) f))"
   by (simp add: pi_d_def reverse_d_def)
 have 2: "(\<exists>i\<le>ilen \<sigma>. f (suffix i (irev \<sigma>))) =
          (\<exists> k \<le> ilen \<sigma>. (prefix k \<sigma>) \<Turnstile> f\<^sup>r)" 
    using pifilt_irev_exists by auto
 have 3: "(pifilt (irev \<sigma>) f) = (imap (\<lambda>s. inth s 0) (sfxfilt (irev \<sigma>) f)) "
    by (simp add: pifilt_def)
 have 4: "(sfxfilt (irev \<sigma>) f) = (ifilter (\<lambda> ys. ys \<Turnstile>f) (suffixes (irev \<sigma>))) "
    by (simp add: sfxfilt_def ) 
 have 5: "(suffixes (irev \<sigma>)) = irev(imap irev (prefixes \<sigma>)) "
   using suffixes_irev by blast 
 have 6: "(ifilter (\<lambda> ys. ys \<Turnstile>f) (suffixes (irev \<sigma>))) =
          (ifilter (\<lambda> ys. ys \<Turnstile>f) (irev(imap irev (prefixes \<sigma>))) )" 
   using 5 by auto
 have 7: "(\<exists>i\<le>ilen \<sigma>. f (suffix i (irev \<sigma>))) \<Longrightarrow>
          (ifilter (\<lambda> ys. ys \<Turnstile>f) (irev(imap irev (prefixes \<sigma>))) ) =
          irev (ifilter (\<lambda> ys. ys \<Turnstile>f) (imap irev (prefixes \<sigma>)) ) "
    by (metis "5" ifilter_irev ilen_imap ilen_prefixes inth_iset inth_suffixes irev_ilen iset_irev) 
 have 8: "(\<exists>i\<le>ilen \<sigma>. f (suffix i (irev \<sigma>))) \<Longrightarrow>
          (ifilter (\<lambda> ys. ys \<Turnstile>f) (imap irev (prefixes \<sigma>))) =
          imap irev (ifilter ((\<lambda> ys. ys \<Turnstile>f)\<circ>irev) ( (prefixes \<sigma>))) "
      by (metis "5" ifilter_imap in_iset_suffixes  iset_irev osfx_suffix)
 have 9: "((\<lambda> ys. ys \<Turnstile>f)\<circ>irev) = (\<lambda> ys. ys \<Turnstile> f\<^sup>r)"
   by (auto simp add: reverse_d_def)
 have 10: "(\<exists>i\<le>ilen \<sigma>. f (suffix i (irev \<sigma>))) \<Longrightarrow>
           irev (ifilter (\<lambda> ys. ys \<Turnstile>f) (imap irev (prefixes \<sigma>)) ) =
        irev ( imap irev  (ifilter (\<lambda> ys. ys \<Turnstile> f\<^sup>r) ( (prefixes \<sigma>))) )" 
   using "8" "9" by auto 
 have 11: "(\<exists>i\<le>ilen \<sigma>. f (suffix i (irev \<sigma>))) \<Longrightarrow> 
           (pifilt (irev \<sigma>) f) =
           (imap (\<lambda>s. inth s 0) (irev ( imap irev  (ifilter (\<lambda> ys. ys \<Turnstile> f\<^sup>r) ( (prefixes \<sigma>))) )))
          " 
   by (simp add: "10" "3" "4" "6" "7")
 have 12: "(\<exists>i\<le>ilen \<sigma>. f (suffix i (irev \<sigma>))) \<Longrightarrow> 
           (imap (\<lambda>s. inth s 0) (irev ( imap irev  (ifilter (\<lambda> ys. ys \<Turnstile> f\<^sup>r) ( (prefixes \<sigma>))) ))) =
           irev (imap (\<lambda>s. inth s 0) ( imap irev  (ifilter (\<lambda> ys. ys \<Turnstile> f\<^sup>r) ( (prefixes \<sigma>))) ))"
   by (simp add: irev_imap) 
 have 13: "(imap (\<lambda>s. inth s 0) ( imap irev  (ifilter (\<lambda> ys. ys \<Turnstile> f\<^sup>r) ( (prefixes \<sigma>))) )) =
           (imap (\<lambda>s. ilast s) (ifilter (\<lambda> ys. ys \<Turnstile> f\<^sup>r) ( (prefixes \<sigma>))) )" 
       by (auto simp add: irev_inth)
 have 14: "(\<exists>i\<le>ilen \<sigma>. f (suffix i (irev \<sigma>))) \<Longrightarrow>  
           (pifilt (irev \<sigma>) f) = 
           irev (imap (\<lambda>s. ilast s) (ifilter (\<lambda> ys. ys \<Turnstile> f\<^sup>r) ( (prefixes \<sigma>))) )" 
   using "11" "12" "13" by auto
 have 15: "(\<exists>i\<le>ilen \<sigma>. f (suffix i (irev \<sigma>))) \<Longrightarrow> 
           irev (imap (\<lambda>s. ilast s) (ifilter (\<lambda> ys. ys \<Turnstile> f\<^sup>r) ( (prefixes \<sigma>))) ) =
           irev (rpifilt \<sigma> (LIFT(f\<^sup>r)))"
  by (simp add: rpifilt_def pfxfilt_def)
 have 16: "(\<exists>i\<le>ilen \<sigma>. f (suffix i (irev \<sigma>))) \<Longrightarrow>
           ((pifilt (irev \<sigma>) f) \<Turnstile> g) =
            ((rpifilt \<sigma> (LIFT(f\<^sup>r))) \<Turnstile> g\<^sup>r)"
  by (metis (mono_tags, lifting) "14" "15" reverse_d_def)
 have 17: "(\<sigma> \<Turnstile> f\<^sup>r \<Pi>\<^sup>p g\<^sup>r) =
           ((\<exists> k \<le> ilen \<sigma>. (prefix k \<sigma>) \<Turnstile> f\<^sup>r) \<and> ( irev(rpifilt \<sigma> (LIFT(f\<^sup>r))) \<Turnstile> g )) " 
   by (simp add: rpi_d_def reverse_d_def)
 show ?thesis 
 using "1" "14" "15" "17" "2" by force
qed

lemma PiRev:
 "\<turnstile> (f \<Pi> g)\<^sup>r = (f\<^sup>r \<Pi>\<^sup>p g\<^sup>r)" 
by (simp add: PiRevsem Valid_def)


subsection \<open>Semantic Lemmas\<close>


lemma sfxfilter_help:
 "(\<exists> ys \<in> iset (suffixes xs) . f ys) = (\<exists> i \<le> ilen xs. f (suffix i xs))"
using iset_suffixes_sfx by auto

lemma pifiltinit_help:
 "(\<exists> y \<in> iset (xs). w \<langle>y\<rangle>) = (\<exists> i \<le> ilen xs. w \<langle>inth xs i\<rangle>)"
by (metis inth_and_iset)

lemma sfxfilt_INil:
 " sfxfilt \<langle>x\<rangle> f = \<langle>\<langle>x\<rangle>\<rangle>"
by (auto simp:sfxfilt_def)

lemma pifilt_INil:
 "pifilt \<langle>x\<rangle> f = \<langle>x\<rangle>"
by (auto simp: pifilt_def sfxfilt_INil)

lemma sfxfilt_ICons:
shows "(sfxfilt (x\<odot>xs) f) =
  (if (\<exists>i\<le>ilen xs. f (suffix i xs)) then 
               (if f (x\<odot>xs) then 
                   (x\<odot>xs)\<odot> (sfxfilt (xs) f)
                else   (sfxfilt xs f) )
              else \<langle>x\<odot>xs\<rangle>)"
proof -
 have 1: "(sfxfilt (x\<odot>xs) f)  = 
          ifilter (\<lambda>ys. f ys) (suffixes (x\<odot>xs))
         " 
    by (simp add: sfxfilt_def)
 have 2: "suffixes (x\<odot>xs) = (x\<odot>xs) \<odot> suffixes xs"
   by simp
 have 3: "ifilter (\<lambda>ys. f ys) (suffixes (x\<odot>xs)) =
          (if (\<exists> ys \<in> iset (suffixes xs). f ys) then
            (if f (x\<odot>xs) then (x\<odot>xs)\<odot> (ifilter (\<lambda>ys. f ys) (suffixes (xs)))  
                          else (ifilter (\<lambda>ys. f ys) (suffixes (xs))) )
           else \<langle>x\<odot>xs\<rangle>)
         "
   by auto
 have 4: "(if (\<exists> ys \<in> iset (suffixes xs). f ys) then
            (if f (x\<odot>xs) then (x\<odot>xs)\<odot> (ifilter (\<lambda>ys. f ys) (suffixes (xs)))  
                          else (ifilter (\<lambda>ys. f ys) (suffixes (xs))) )
           else \<langle>x\<odot>xs\<rangle>) =
           (if (\<exists>i\<le>ilen xs. f (suffix i xs)) then 
               (if f (x\<odot>xs) then 
                   (x\<odot>xs)\<odot> (sfxfilt (xs) f)
                else   (sfxfilt xs f) )
              else \<langle>x\<odot>xs\<rangle>)
         " 
    by (auto simp add: sfxfilt_def iset_suffixes_sfx)
  show ?thesis by (simp add: "1" "4")   
qed

lemma pifilt_ICons:
shows "(pifilt (x\<odot>xs) f) =
  (if (\<exists>i\<le>ilen xs. f (suffix i xs)) then 
               (if f (x\<odot>xs) then 
                   (x)\<odot> (pifilt (xs) f)
                else   (pifilt xs f) )
              else \<langle>x\<rangle>)"
proof -
 have 1: "(pifilt (x\<odot>xs) f) = imap (\<lambda>s. (inth s 0)) (sfxfilt (x\<odot>xs) f)"
   by (simp add: pifilt_def) 
 have 2: "imap (\<lambda>s. (inth s 0)) (sfxfilt (x\<odot>xs) f) =
          (if (\<exists>i\<le>ilen xs. f (suffix i xs)) then 
               (if f (x\<odot>xs) then 
                   imap (\<lambda>s. (inth s 0))((x\<odot>xs)\<odot> (sfxfilt (xs) f))
                else  imap (\<lambda>s. (inth s 0)) (sfxfilt xs f) )
              else imap (\<lambda>s. (inth s 0)) \<langle>x\<odot>xs\<rangle>)
         "
    using "1" 
    by (metis sfxfilt_ICons)
 have 3: " imap (\<lambda>s. (inth s 0))((x\<odot>xs)\<odot> (sfxfilt (xs) f)) =
           (x)\<odot> (pifilt (xs) f)
         " 
    by (simp add: pifilt_def)
 have 4: "imap (\<lambda>s. (inth s 0)) \<langle>x\<odot>xs\<rangle> = \<langle>x\<rangle>" 
    by simp
 have 5: "imap (\<lambda>s. (inth s 0)) (sfxfilt xs f) = (pifilt (xs) f)"
    by (simp add: pifilt_def)
 have 6: "(if (\<exists>i\<le>ilen xs. f (suffix i xs)) then 
               (if f (x\<odot>xs) then 
                   imap (\<lambda>s. (inth s 0))((x\<odot>xs)\<odot> (sfxfilt (xs) f))
                else  imap (\<lambda>s. (inth s 0)) (sfxfilt xs f) )
              else imap (\<lambda>s. (inth s 0)) \<langle>x\<odot>xs\<rangle>) =
           (if (\<exists>i\<le>ilen xs. f (suffix i xs)) then 
               (if f (x\<odot>xs) then 
                   (x)\<odot> (pifilt (xs) f)
                else   (pifilt xs f) )
              else \<langle>x\<rangle>)"
    using "5" by auto
  show ?thesis by (simp add: "1" "2" "6")
qed   

lemma sfxfilt_inth_ICons:
 "inth (sfxfilt (x\<odot>xs) f) j =
             (if (\<exists>i\<le>ilen xs. f (suffix i xs)) then 
               (if f (x\<odot>xs) then 
                 (if j =0 then  (x\<odot>xs) else inth (sfxfilt (xs) f) (j-1)) 
                else  inth (sfxfilt (xs) f) j)
              else (x\<odot>xs))"
by simp
  (metis  inth.simps(1) One_nat_def inth_ICons_a inth_zero sfxfilt_ICons)

lemma pifilt_inth_ICons:
 " inth (pifilt (x \<odot> xs) f) i = 
             (if (\<exists>i\<le>ilen xs. f (suffix i xs)) then 
               (if f (x\<odot>xs) then 
                 (if i =0 then  x else inth (pifilt (xs) f) (i-1)) 
                else  inth (pifilt (xs) f) i)
              else x)"
by (metis  inth.simps(1) One_nat_def Suc_pred inth_Suc inth_zero not_gr0 pifilt_ICons)
 
lemma sfxfilt_inth:
assumes "(\<exists> i \<le> ilen \<sigma>. (suffix i \<sigma>) \<Turnstile> f)"
        " i\<le>ilen (sfxfilt \<sigma> f)"
shows   " (inth (sfxfilt \<sigma> f) i) \<Turnstile> f"
using assms in_iset_suffixes osfx_suffix sfxfilter_inth
by (simp add: sfxfilt_def ifilter_inth_aa sfxfilter_help)


lemma pifilt_exists:
 assumes "(\<exists>i\<le>ilen \<sigma>. f (suffix i \<sigma>) )"
 shows   "(\<exists>i \<le> ilen(sfxfilt \<sigma> f). (inth (sfxfilt \<sigma> f) i) \<Turnstile> f)"
using assms 
using sfxfilt_inth by blast

lemma sfxfilt_pifilt_ilen:
assumes "(\<exists> i \<le> ilen \<sigma>. (suffix i \<sigma>) \<Turnstile> f)"
shows   "ilen (pifilt \<sigma> f) = ilen (imap (\<lambda>s. inth s 0) (sfxfilt \<sigma> f))  "
using assms by (simp add: pifilt_def)

lemma sfxfilt_pifilt_inth:
assumes "(\<exists> i \<le> ilen \<sigma>. (suffix i \<sigma>) \<Turnstile> f)"
        "j\<le> ilen (pifilt \<sigma> f)"
shows   " inth (pifilt \<sigma> f) j = inth (imap (\<lambda>s. inth s 0) (sfxfilt \<sigma> f)) j  "
using assms
by (simp add: pifilt_def)

lemma sfxfilt_pifilt:
assumes "(\<exists> i \<le> ilen \<sigma>. (suffix i \<sigma>) \<Turnstile> f)"
shows   "(imap (\<lambda>s. inth s 0) (sfxfilt \<sigma> f)) = pifilt \<sigma> f"
using assms by (simp add: pifilt_def)

lemma sfxfilt_ilen_bound:
assumes "(\<exists>i\<le>ilen xs. f (suffix i xs))"
shows   "ilen (sfxfilt xs f) \<le> ilen xs "
using assms 
by (simp add: sfxfilt_def)
   (metis ilen_ifilter_le ilen_suffixes)

lemma pifilt_ilen_bound:
assumes "(\<exists>i\<le>ilen \<sigma>. f (suffix i \<sigma>))"
shows   "ilen (pifilt \<sigma> f) \<le> ilen \<sigma> "
using assms by (simp add: pifilt_def sfxfilt_ilen_bound)

lemma sfxfilt_ilen_inth_bound:
assumes "(\<exists>i\<le>ilen \<sigma>. f (suffix i \<sigma>))"
        " j\<le> ilen (sfxfilt \<sigma> f)"
shows   "ilen (inth (sfxfilt \<sigma> f) j) \<le> ilen \<sigma> "
using assms 
by (simp add: sfxfilt_def sfxfilter_help sfxfilter_inth_bound)

lemma sfxfilt_pifilt_inth_suffix:
assumes "(\<exists> i \<le> ilen \<sigma>. (suffix i \<sigma>) \<Turnstile> f)"
        "j\<le> ilen (pifilt \<sigma> f)"
shows   "  inth (sfxfilt \<sigma> f) j =  suffix (ilen \<sigma> - (ilen( inth (sfxfilt \<sigma> f) j))) \<sigma>  "
proof -
 have 1: "inth (sfxfilt \<sigma> f) j = inth (ifilter f (suffixes \<sigma>)) j" 
   by (simp add: sfxfilt_def)
 have 2: "suffix (ilen \<sigma> - ilen (inth (ifilter f (suffixes \<sigma>)) j)) \<sigma> =
          suffix (ilen \<sigma> - (ilen(inth (sfxfilt \<sigma> f) j))) \<sigma>" 
   by (simp add: sfxfilt_def)
 have 3: "j \<le> ilen (ifilter f (suffixes \<sigma>))" 
   using assms by (simp add: pifilt_def sfxfilt_def)
 have 4: "(\<exists>i\<le>ilen \<sigma>. f (suffix i \<sigma>))" 
    using assms by auto
 have 5: "(\<exists> ys \<in> iset (suffixes \<sigma>). f ys)"
   using "4" 
   using in_iset_suffixes osfx_suffix by blast 
 have 6: "inth (ifilter f (suffixes \<sigma>)) j =
          suffix (ilen \<sigma> - ilen (inth (ifilter f (suffixes \<sigma>)) j)) \<sigma>
         " 
   using "3" "5" sfxfilter_inth_suffix[of "\<sigma>" "f" "j"] by blast
 show ?thesis using "1" "6" by auto
qed

lemma pifilt_inth:
 assumes "(\<exists>i\<le>ilen \<sigma>. f (suffix i \<sigma>))"
         "i \<le> ilen (pifilt \<sigma> f)"
shows    "(\<exists> k \<le> ilen \<sigma>. inth (pifilt \<sigma> f) i = inth \<sigma> k)"
using assms sfxfilt_pifilt_inth_suffix[of "\<sigma>" "f" "i"] 
by (simp add: pifilt_def) 
   (metis diff_le_self ifirst_suffix  inth_imap )

lemma sfxfilt_ilen_imp:
 assumes "(\<exists>i\<le>ilen \<sigma>. f (suffix i \<sigma>) \<and> g (suffix i \<sigma>))"
 shows   " ilen (sfxfilt \<sigma> (LIFT(f \<and> g))) \<le> ilen (sfxfilt \<sigma> f)"
proof -
 have 1: "ilen (sfxfilt \<sigma> (LIFT(f \<and> g))) = 
          ilen (ifilter (\<lambda>ys. f ys \<and> g ys) (suffixes \<sigma>))" 
    by (simp add: sfxfilt_def)
 have 2: "ilen (ifilter f (suffixes \<sigma>)) = ilen (sfxfilt \<sigma> f) "
    by (simp add:  sfxfilt_def)
 have 3: "\<exists> x \<in> iset (suffixes \<sigma>). f x \<and> g x" 
    using assms by auto
 have 4: "ilen (ifilter (\<lambda>x. f x \<and> g x) (suffixes \<sigma>)) \<le> ilen (ifilter f (suffixes \<sigma>))" 
    using "3" ifilter_ilen_imp[of "suffixes \<sigma>" "f" "g"] by auto
 show ?thesis using "1" "2" "4" by auto
qed

lemma pifilt_ilen_imp:
 assumes "(\<exists>i\<le>ilen \<sigma>. f (suffix i \<sigma>) \<and> g (suffix i \<sigma>))"
 shows   " ilen (pifilt \<sigma> (LIFT(f \<and> g))) \<le> ilen (pifilt \<sigma> f)"
using assms
by (simp add: sfxfilt_ilen_imp pifilt_def)

lemma interval_iset_sfxfilt [simp]:
assumes "(\<exists> i \<le> ilen xs. f (suffix i xs))"
shows "(iset (sfxfilt xs f)) = {ys. ys \<in> iset (suffixes xs) \<and> f (ys)}"
using assms 
proof -
 have 1: "iset (sfxfilt xs f) = iset (ifilter f (suffixes xs))"
    by (simp add: sfxfilt_def) 
 have 2: "\<exists> x \<in> iset (suffixes xs). f x" 
    using assms using in_iset_suffixes osfx_suffix by blast
 have 3: "iset (ifilter f (suffixes xs)) = 
          {ys. ys \<in> iset (suffixes xs) \<and> f ys}
         " 
   using "2" iset_ifilter[of "suffixes xs" "f"] by auto
 show ?thesis by (simp add: "1" "3")
qed 


lemma interval_subset_sfxfilt [simp]:
assumes "(\<exists> i \<le> ilen xs. f (suffix i xs))"
shows "(iset (sfxfilt xs f)) \<le> (iset (sfxfilt xs (LIFT(f \<or> g))))"
proof -
 have 1: "\<exists> x \<in> iset (suffixes xs). f x"
   using assms using in_iset_suffixes osfx_suffix by blast 
 have 2: "iset (sfxfilt xs f) = iset (ifilter f (suffixes xs))"
     by (simp add: sfxfilt_def)
 have 3: " iset (sfxfilt xs (LIFT(f \<or> g))) = iset (ifilter (\<lambda>x. f x \<or> g x) (suffixes xs))"
   by (simp add: sfxfilt_def)
 have 4: "iset (ifilter f (suffixes xs)) \<le> iset (ifilter (\<lambda>x. f x \<or> g x) (suffixes xs))"
   using "1"  subset_ifilter[of "suffixes xs" "f" "g"] by auto
 show ?thesis using "2" "3" "4" by blast
qed


lemma interval_iset_pifilt [simp]:
assumes "(\<exists> i \<le> ilen xs. f (suffix i xs))"
shows "(iset (pifilt xs f)) = {(inth ys 0)| ys. ys \<in> iset(suffixes xs) \<and> f (ys)}"
proof -
 have 1: "(iset (pifilt xs f)) = (iset (imap (\<lambda>s. inth s 0) (sfxfilt xs f)))"
    by (simp add: pifilt_def)
 have 2: "(iset (imap (\<lambda>s. inth s 0) (sfxfilt xs f))) =
          {(inth ys 0) | ys. ys \<in> iset (sfxfilt xs f)}"
    by (induction xs) auto
 have 3: "{(inth ys 0) | ys. ys \<in> iset (sfxfilt xs f)} =
          {(inth ys 0)| ys. ys \<in> iset(suffixes xs) \<and> f (ys)}
         " 
    using assms by auto
 show ?thesis using "1" "2" "3" by auto
qed

lemma interval_inth_sfxfilt_in_iset:
"x \<in> iset (sfxfilt \<sigma> (LIFT(f \<or> g))) =
 (\<exists> i \<le> ilen (sfxfilt \<sigma> (LIFT(f \<or> g))). x =(inth (sfxfilt \<sigma> (LIFT(f \<or> g))) i))"
by (metis inth_and_iset)


lemma sfxfilt_inth_or:
assumes "(\<exists> i \<le> ilen \<sigma>. (suffix i \<sigma>) \<Turnstile> f)"
shows   "(\<exists> i \<le> ilen (sfxfilt \<sigma> (LIFT(f \<or> g))). (inth (sfxfilt \<sigma> (LIFT(f \<or> g)) ) i) \<Turnstile> f)"
proof -
 have 1: "\<exists> x \<in> iset (suffixes \<sigma>). f x" 
   using assms by auto
 have 2: "sfxfilt \<sigma> (LIFT(f \<or> g)) = ifilter (\<lambda> x. f x \<or> g x) (suffixes \<sigma>) "
   by (simp add: sfxfilt_def)
 have 3: "\<exists> x \<in> iset(ifilter (\<lambda> x. f x \<or> g x) (suffixes \<sigma>)). f x "
   using "1" ifilter_inth_or[of "suffixes \<sigma>" "f" "g"] by auto
 from "2" "3" show ?thesis by (metis interval_inth_sfxfilt_in_iset)
qed

lemma NotPiFalse:
 "\<sigma> \<Turnstile> \<not> ((#False) \<Pi> f)"
by (simp add: pi_d_def)

lemma pifilt_true:
 " pifilt \<sigma> (LIFT(#True)) = \<sigma> "
by (simp add: pifilt_def sfxfilt_def ifilter_True)

lemma pifilt_init_INil:
 "(pifilt \<langle>x\<rangle> (LIFT(init w))) = \<langle>x\<rangle>"
by ( auto simp add: pifilt_def sfxfilt_def)

lemma  pifilt_init_ICons:
 "(pifilt (x\<odot>xs) (LIFT(init w))) =
  (if (\<exists>i\<le>ilen xs. w \<langle>inth xs i\<rangle>)  then
           (if  w \<langle>x\<rangle> then x \<odot> (pifilt xs (LIFT(init w)))
                           else (pifilt xs (LIFT(init w))))
         else \<langle>x\<rangle>)"
by (simp add: pifilt_def sfxfilt_def init_defs)
 (metis ifirst_suffix in_iset_suffixes interval_sfx_1 osfx_suffix pifiltinit_help)
 

lemma PiStatesem: 
 " (\<sigma> \<Turnstile> (init w) \<Pi> f) =
   ((\<exists>i\<le>ilen \<sigma>. w \<langle>inth \<sigma> i\<rangle>) \<and> f (pifilt \<sigma> (LIFT(init w))))"
by (simp add: pi_d_def init_defs) 


subsection \<open>Soundness of Axioms\<close>

subsubsection \<open>PiK\<close>

lemma PiKsem:
 "\<sigma> \<Turnstile> f1 \<Pi>\<^sup>u ( f \<longrightarrow> g) \<longrightarrow> (  f1 \<Pi>\<^sup>u f \<longrightarrow> f1 \<Pi>\<^sup>u g ) "
by (simp add: upi_d_def init_defs pi_d_def) auto

lemma PiK:
 "\<turnstile> f1 \<Pi>\<^sup>u (f \<longrightarrow> g) \<longrightarrow> (  f1 \<Pi>\<^sup>u f \<longrightarrow> f1 \<Pi>\<^sup>u g )"
using PiKsem Valid_def by blast 

subsubsection \<open>PiDc\<close>

lemma PiDcsem:
 " \<sigma> \<Turnstile> f \<Pi> g \<longrightarrow> f \<Pi>\<^sup>u g"
by (simp add: upi_d_def init_defs pi_d_def)

lemma PiDc:
 "\<turnstile> f \<Pi> g \<longrightarrow> f \<Pi>\<^sup>u g"
using PiDcsem Valid_def by blast

subsubsection \<open>PiN\<close>

lemma PiN:
 assumes "\<turnstile> g"
 shows   "\<turnstile> f \<Pi>\<^sup>u g"
using assms by (simp add: Valid_def pi_d_def upi_d_def)

subsubsection \<open>PiTrueEqvDiamond\<close>

lemma PiTrueEqvDiamond:
 "\<turnstile> f \<Pi> #True = \<diamond> f"
by (simp add: Valid_def pi_d_def sometimes_defs)

subsubsection \<open>PiOr\<close>

lemma PiOr:
 "\<turnstile> f \<Pi> (g1 \<or> g2) = (f \<Pi> g1 \<or> f \<Pi> g2)"
by (simp add: Valid_def pi_d_def) blast

subsubsection \<open>UPiFalseEqvBoxNot:\<close>

lemma UPiFalseEqvBoxNot:
 "\<turnstile> f \<Pi>\<^sup>u #False = \<box> (\<not>f)"
by (simp add: Valid_def upi_d_def pi_d_def always_defs)

subsubsection \<open>BoxEqvImpPiEqv\<close>

lemma BoxEqvImpPiEqvsem:
assumes "(\<sigma> \<Turnstile> \<box> (f1 = f2))"
shows   "(\<sigma> \<Turnstile>(f1 \<Pi> g = f2 \<Pi> g))"
proof -
 show   "(\<sigma> \<Turnstile>(f1 \<Pi> g = f2 \<Pi> g))" 
 proof -
  have 1: "\<forall>n\<le>ilen \<sigma>. f1 (suffix n \<sigma>) = f2 (suffix n \<sigma>)" 
   using assms by (simp add: always_defs)
  have 2: "(\<sigma> \<Turnstile>(f1 \<Pi> g)) = ((\<exists>i\<le>ilen \<sigma>. f1 (suffix i \<sigma>)) \<and> g (pifilt \<sigma> f1))"
   by (simp add: pi_d_def)
  have 3: "(\<exists>i\<le>ilen \<sigma>. f1 (suffix i \<sigma>)) = (\<exists>i\<le>ilen \<sigma>. f2 (suffix i \<sigma>))"
    using "1" by blast  
  have 4: " (sfxfilt \<sigma> f1) =  (sfxfilt \<sigma> f2)"
     using "1" 
   proof (induct \<sigma>)
   case (INil x)
   then show ?case by (simp add: sfxfilt_INil)
   next
   case (ICons x1a \<sigma>)
   then show ?case 
     proof -
       have 41: "(sfxfilt (x1a \<odot> \<sigma>) f1) = 
         (if (\<exists>i\<le>ilen \<sigma>. (suffix i \<sigma>) \<Turnstile> f1) then
           (if ((x1a\<odot>\<sigma>) \<Turnstile> f1) then (x1a\<odot>\<sigma>) \<odot> (sfxfilt \<sigma> f1)
                             else (sfxfilt \<sigma> f1))
         else \<langle>x1a\<odot>\<sigma>\<rangle>)"
        using sfxfilt_ICons by blast
       have 42: " sfxfilt \<sigma> f1 = sfxfilt \<sigma> f2"
        by (metis ICons.hyps ICons.prems Suc_le_mono suffix_suc ilen.simps(2) 
            plus_1_eq_Suc)
       have 43: "((x1a\<odot>\<sigma>) \<Turnstile> f1) = ((x1a\<odot>\<sigma>) \<Turnstile> f2)"
        using ICons.prems by auto
       have 44: "(\<exists>i\<le>ilen \<sigma>. (suffix i \<sigma>) \<Turnstile> f1) =
                 (\<exists>i\<le>ilen \<sigma>. (suffix i \<sigma>) \<Turnstile> f2)"
        by (metis  Suc_le_mono suffix_suc ilen.simps(2) ICons(2) plus_1_eq_Suc) 
       have 45: "(if (\<exists>i\<le>ilen \<sigma>. (suffix i \<sigma>) \<Turnstile> f1) then
           (if ((x1a\<odot>\<sigma>) \<Turnstile> f1) then (x1a\<odot>\<sigma>) \<odot> (sfxfilt \<sigma> f1)
                             else (sfxfilt \<sigma> f1))
         else \<langle>x1a\<odot>\<sigma>\<rangle>) =
             (if (\<exists>i\<le>ilen \<sigma>. (suffix i \<sigma>) \<Turnstile> f2) then
           (if ((x1a\<odot>\<sigma>) \<Turnstile> f2) then (x1a\<odot>\<sigma>) \<odot> (sfxfilt \<sigma> f2)
                             else (sfxfilt \<sigma> f2))
         else \<langle>x1a\<odot>\<sigma>\<rangle>)" 
        using "42" "43" "44" by auto
       have 46: "(if (\<exists>i\<le>ilen \<sigma>. (suffix i \<sigma>) \<Turnstile> f2) then
           (if ((x1a\<odot>\<sigma>) \<Turnstile> f2) then (x1a\<odot>\<sigma>) \<odot> (sfxfilt \<sigma> f2)
                             else (sfxfilt \<sigma> f2))
         else \<langle>x1a\<odot>\<sigma>\<rangle>) = (sfxfilt (x1a \<odot> \<sigma>) f2)"
        by (simp add: sfxfilt_ICons)
       show ?thesis 
       using "41" "45" "46" by presburger 
      qed
   qed
  have 47: "(pifilt \<sigma> f1) =  (pifilt \<sigma> f2)" 
     by (simp add: "4" pifilt_def)
  have 5: "((\<exists>i\<le>ilen \<sigma>. f1 (suffix i \<sigma>)) \<and> g (pifilt \<sigma> f1)) =
           ((\<exists>i\<le>ilen \<sigma>. f2 (suffix i \<sigma>)) \<and> g (pifilt \<sigma> f2))"
   by (simp add: "3" "47")
  show ?thesis by (simp add: "5" pi_d_def)
  qed    
qed

lemma BoxEqvImpPiEqv:
 "\<turnstile> \<box> (f1 = f2) \<longrightarrow> (f1 \<Pi> g = f2 \<Pi> g)"
using BoxEqvImpPiEqvsem by (simp add: Valid_def,auto)      


subsubsection \<open>PiDiamondImpDiamond\<close>

lemma PiDiamondImpDiamondsem:
 "\<sigma> \<Turnstile> f \<Pi> (\<diamond> (init w)) \<longrightarrow> \<diamond> (init w)"
using pifilt_inth by (simp add: Valid_def pi_d_def sometimes_defs init_defs) fastforce

lemma PiDiamondImp:
 "\<turnstile> f \<Pi> (\<diamond> (init w)) \<longrightarrow> \<diamond> (init w)"
 using PiDiamondImpDiamondsem Valid_def by blast


subsubsection \<open>PiAssoc\<close>

                        
lemma PiAssocsem1:
assumes " i \<le> ilen \<sigma>"
        "f (suffix i \<sigma>)"
        "ia \<le> ilen (pifilt \<sigma> f)"
        "w \<langle>inth (pifilt \<sigma> f) ia\<rangle>" 
shows   "\<exists>i\<le>ilen \<sigma>. f (suffix i \<sigma>) \<and> w \<langle>inth (suffix i \<sigma>) 0\<rangle>"
proof -
 have 1: "(inth (pifilt \<sigma> f) ia) = (inth (imap (\<lambda>s. inth s 0) (sfxfilt \<sigma> f)) ia)"
   using assms(1) assms(2) assms(3) sfxfilt_pifilt_inth by blast 
 have 2: "(inth (imap (\<lambda>s. inth s 0) (sfxfilt \<sigma> f)) ia) =
          (\<lambda>s. inth s 0) (inth (sfxfilt \<sigma> f) ia)" 
  using inth_imap  by auto
 have 3: "f  (inth (sfxfilt \<sigma> f) ia)"
    using sfxfilt_inth 
    by (metis assms(1) assms(2) assms(3) ilen_imap pifilt_def)
 have 4: "inth (sfxfilt \<sigma> f) ia = suffix (ilen \<sigma> - ilen (inth (sfxfilt \<sigma> f) ia)) \<sigma>"
   using sfxfilt_pifilt_inth_suffix  assms(1) assms(2) assms(3) by blast
 have 5: "w \<langle>inth (suffix (ilen \<sigma> - ilen (inth (sfxfilt \<sigma> f) ia)) \<sigma>) 0\<rangle>  " 
   using "1" "2" "4" assms(4) by auto
 show ?thesis 
 by (metis "3" "4" "5" diff_le_self)
qed
  
lemma PiAssocsem2:
 assumes "i \<le> ilen \<sigma>"
         "f (suffix i \<sigma>)"
         " w \<langle> inth \<sigma> i\<rangle>"
shows "\<exists>j\<le>ilen (pifilt \<sigma> f). w \<langle>inth (pifilt \<sigma> f) j\<rangle>"
proof -
 have 1: "\<exists>j\<le>ilen (sfxfilt \<sigma> f). f (inth (sfxfilt \<sigma> f) j)"
     using assms pifilt_exists by blast
 have 2: " (LIFT (init w)) (suffix i \<sigma>)"
    by (simp add: assms init_defs)
 have 3: "\<exists>j\<le>ilen (sfxfilt \<sigma> (LIFT(init w))). (LIFT(init w)) (inth (sfxfilt \<sigma> (LIFT(init w))) j)"
   using pifilt_exists "2" assms by blast
 have 4: "(LIFT (f \<and> init w)) (suffix i \<sigma>)"
    by (simp add: assms init_defs)
 have 5: "\<exists>j\<le>ilen (sfxfilt \<sigma> (LIFT(f \<and> init w))). 
           (LIFT(f \<and> init w)) (inth (sfxfilt \<sigma> (LIFT(f \<and> init w))) j) "
   using pifilt_exists "4" assms by blast
 have 6: "\<exists>i\<le>ilen \<sigma>. suffix i \<sigma> \<Turnstile> f \<and> init w"
   using "4" assms by blast
 have 7: "\<exists>j\<le>ilen (sfxfilt \<sigma> (LIFT((f \<and> init w) \<or> (f \<and> \<not> (init w)))  )). 
           (LIFT(f \<and> init w)) (inth (sfxfilt \<sigma> (LIFT((f \<and> init w) \<or> (f \<and> \<not>(init w))) )) j)" 
  using "6" sfxfilt_inth_or[of "\<sigma>" "LIFT(f \<and> init w)" "LIFT(f \<and> \<not>(init w))"] 
      by auto
 have 8: "\<And>\<sigma> . (\<sigma> \<Turnstile> ((f \<and> init w) \<or> (f \<and> \<not> (init w)))) = (\<sigma> \<Turnstile> f)"
    by auto  
 have 9: "(sfxfilt \<sigma> (LIFT((f \<and> init w) \<or> (f \<and> \<not> (init w)))  )) = 
          (sfxfilt \<sigma> f)"
      using "8" by (simp add: sfxfilt_def)
 have 10: "\<exists>j\<le>ilen (sfxfilt \<sigma> f). 
           (LIFT(f \<and> init w)) (inth (sfxfilt \<sigma> f) j)"
   using "7" "9" by auto 
 have 11: "ilen (sfxfilt \<sigma> f) = ilen (pifilt \<sigma> f)" 
     by (simp add: pifilt_def)
 have 12: "\<exists>j\<le>ilen (pifilt \<sigma> f). 
           (LIFT(init w)) (inth (sfxfilt \<sigma> f) j)" 
    using "10" "11" by auto
 from 12 11 show ?thesis 
   by (simp add: init_defs)
      (metis inth_imap pifilt_def)
qed    
    
lemma PiAssocsema:
 "((\<exists>i\<le>ilen \<sigma>. f (suffix i \<sigma>)) \<and>
     (\<exists>i\<le>ilen (pifilt \<sigma> f). w \<langle>inth (suffix i (pifilt \<sigma> f)) 0\<rangle>)) =
  (\<exists>i\<le>ilen \<sigma>. f (suffix i \<sigma>) \<and> w \<langle>inth (suffix i \<sigma>) 0\<rangle>) "
using PiAssocsem1 PiAssocsem2 by fastforce

lemma PiAssocsemb:
 "((\<exists>i\<le>ilen \<sigma>. f (suffix i \<sigma>)) \<and>
     (\<exists>i\<le>ilen (pifilt \<sigma> f). (LIFT(init w)) (suffix i (pifilt \<sigma> f)) )) =
  (\<exists>i\<le>ilen \<sigma>. (LIFT(f \<and> init w)) (suffix i \<sigma>)) "
using PiAssocsem1 PiAssocsem2 
by (simp add: init_defs) fastforce

lemma pifilt_INil_help:
 "(\<exists> x \<in> iset (suffixes xs) . (LIFT(init w)) x) = (\<exists> x \<in> iset xs. w (\<langle>x\<rangle>))"
proof (auto simp add: init_defs)
 show "\<And>x. osfx x xs \<Longrightarrow> w \<langle> inth x 0\<rangle> \<Longrightarrow> \<exists>x\<in>iset xs. w \<langle>x\<rangle>" 
 using in_iset_suffixes interval_sfx_1 by blast
 show "\<And>x. x \<in> iset xs \<Longrightarrow> w \<langle>x\<rangle> \<Longrightarrow> \<exists>x\<in>iset (suffixes xs). w \<langle> inth x 0\<rangle>"
 by (metis in_iset_suffixes ifirst_suffix inth_and_iset osfx_suffix)
qed


lemma pifilt_init:
 assumes "(\<exists>i\<le>ilen xs. (LIFT(init w)) (suffix i xs))"
 shows   "(pifilt xs (LIFT(init w))) = ifilter (\<lambda>y. w (\<langle>y\<rangle>)) xs"
using assms
proof (induct xs)
case (INil x)
then show ?case 
by (simp add: pifilt_init_INil)
next
case (ICons x1a xs)
then show ?case 
  proof -
  have 1: "pifilt (x1a\<odot>xs) (LIFT(init w)) =
           imap (\<lambda>xs . (inth xs 0)) (sfxfilt (x1a\<odot>xs) (LIFT(init w)))
          " 
       using sfxfilt_pifilt by (simp add: pifilt_def)
  have 2: "sfxfilt (x1a\<odot>xs) (LIFT(init w)) =
           (ifilter (\<lambda> ys. (LIFT(init w)) ys) (suffixes (x1a \<odot>xs)))
          " 
       using sfxfilt_def by blast
  have 3: "suffixes (x1a \<odot>xs) = (x1a\<odot>xs)\<odot> (suffixes xs) "
       by simp
  have 4: "(ifilter (\<lambda> ys. (LIFT(init w)) ys) (suffixes (x1a \<odot>xs))) =
           (if (\<exists> x \<in> iset (suffixes xs) . (LIFT(init w)) x) then
                (if (LIFT(init w)) (x1a\<odot>xs) then (x1a\<odot>xs)\<odot> (ifilter (LIFT(init w)) (suffixes xs)) 
                 else (ifilter (LIFT(init w)) (suffixes xs)))
            else \<langle>x1a \<odot> xs\<rangle>)
          "  
       by simp
  have 5: " imap (\<lambda>xs . (inth xs 0)) (ifilter (\<lambda> ys. (LIFT(init w)) ys) (suffixes (x1a \<odot>xs))) =
            (if (\<exists> x \<in> iset (suffixes xs) . (LIFT(init w)) x) then
                (if (LIFT(init w)) (x1a\<odot>xs) 
                 then (x1a) \<odot> imap (\<lambda>xs . (inth xs 0)) (ifilter (LIFT(init w)) (suffixes xs)) 
                 else imap (\<lambda>xs . (inth xs 0)) (ifilter (LIFT(init w)) (suffixes xs)))
             else \<langle>x1a\<rangle>)
          " 
       by auto
  have 6: "ifilter (\<lambda>y. w \<langle>y\<rangle>) (x1a \<odot> xs) =
           (if (\<exists> x \<in> iset xs. w (\<langle>x\<rangle>)) then 
              (if w (\<langle>x1a\<rangle>) then x1a\<odot> (ifilter (\<lambda>y. w \<langle>y\<rangle>) xs) else (ifilter (\<lambda>y. w \<langle>y\<rangle>) xs))
            else \<langle>x1a\<rangle>)
          "
        by simp
  have 61: "(\<exists> x \<in> iset (suffixes xs) . (LIFT(init w)) x) =
             (\<exists> x \<in> iset xs. w (\<langle>x\<rangle>))
           " 
        by (auto simp: init_defs interval_sfx_1)
           (metis init_defs prefix_zero_ifirst pifilt_INil_help)
  have 62: "(LIFT(init w)) (x1a\<odot>xs) = w (\<langle>x1a\<rangle>)" 
    by (simp add: init_defs)
  have 63: " (\<exists> x \<in> iset xs. w (\<langle>x\<rangle>)) \<longrightarrow> 
             imap (\<lambda>xs . (inth xs 0)) (ifilter (LIFT(init w)) (suffixes xs)) =
             (ifilter (\<lambda>y. w \<langle>y\<rangle>) xs)" 
       by (auto simp add: inth_and_iset ) 
         (metis ICons.hyps interval.distinct(1) pifilt_ICons pifilt_def pifilt_init_ICons sfxfilt_def)
  have 7: "(if (\<exists> x \<in> iset (suffixes xs) . (LIFT(init w)) x) then
                (if (LIFT(init w)) (x1a\<odot>xs) 
                 then (x1a) \<odot> imap (\<lambda>xs . (inth xs 0)) (ifilter (LIFT(init w)) (suffixes xs)) 
                 else imap (\<lambda>xs . (inth xs 0)) (ifilter (LIFT(init w)) (suffixes xs)))
            else \<langle>x1a\<rangle>) =
           (if (\<exists> x \<in> iset xs. w (\<langle>x\<rangle>)) then 
              (if w (\<langle>x1a\<rangle>) then x1a\<odot> (ifilter (\<lambda>y. w \<langle>y\<rangle>) xs) else (ifilter (\<lambda>y. w \<langle>y\<rangle>) xs))
            else \<langle>x1a\<rangle>)
          " 
      by (simp add: "61" "62" "63")
     show ?thesis using "1" "2" "5" "7" by auto 
   qed       
qed

lemma pifilt_init_a:
 assumes "(\<exists>i\<le>ilen xs.  w \<langle> inth xs i\<rangle>)"
 shows   "(pifilt xs (\<lambda>s. w \<langle> inth s 0\<rangle>) ) = ifilter (\<lambda>y. w (\<langle>y\<rangle>)) xs"
using assms pifilt_init by (auto simp add: init_defs pifilt_def sfxfilt_def) 

lemma pifilt_pifilt :
assumes "(\<exists>i\<le>ilen xs. f (suffix i xs))"
        "(\<exists>i\<le>ilen (pifilt xs f). w \<langle>inth (suffix i (pifilt xs f)) 0\<rangle>)"
shows "(pifilt (pifilt xs f) (LIFT(init w)))  = pifilt xs (LIFT(f \<and> init w))"
proof -
 have 1: "\<exists>i\<le>ilen (pifilt xs f). (LIFT(init w)) (suffix i (pifilt xs f))"
      using assms by (simp add: init_defs)
 have 2: "(pifilt (pifilt xs f) (LIFT(init w))) =
          ifilter (\<lambda>y. w (\<langle>y\<rangle>)) (pifilt xs f)
         " 
      using "1" pifilt_init[of "(pifilt xs f)" "w" ] by auto 
 have 3: "(pifilt xs f) =
          imap (\<lambda>s. inth s 0) (sfxfilt xs f)" 
      by (simp add: assms sfxfilt_pifilt)
 have 4: "(sfxfilt xs f) = ifilter (\<lambda> ys. f ys) (suffixes xs)"
      using sfxfilt_def by blast
 have 5: " (pifilt xs f) = imap (\<lambda>s. inth s 0) (ifilter (\<lambda> ys. f ys) (suffixes xs))" 
      by (simp add: "3" "4")
 have 6: "ifilter (\<lambda>y. w (\<langle>y\<rangle>)) (pifilt xs f)  =
          ifilter (\<lambda>y. w (\<langle>y\<rangle>)) (imap (\<lambda>s. inth s 0) (ifilter (\<lambda> ys. f ys) (suffixes xs)))
         " 
   using "5" by simp
 have 7: "ifilter (\<lambda>y. w (\<langle>y\<rangle>)) (imap (\<lambda>s. inth s 0) (ifilter (\<lambda> ys. f ys) (suffixes xs))) =
      imap (\<lambda>s. inth s 0) (ifilter ((\<lambda>y. w (\<langle>y\<rangle>))\<circ>(\<lambda>s. inth s 0)) (ifilter (\<lambda> ys. f ys) (suffixes xs))) " 
   using assms by (metis "3" "4" ifilter_imap in_iset_suffixes interval_sfx_1 osfx_suffix)
 have 8: "\<exists>x\<in>iset (ifilter (\<lambda> ys. f ys) (suffixes xs)). ((\<lambda>y. w (\<langle>y\<rangle>))\<circ>(\<lambda>s. inth s 0)) x" 
      using assms "3" "4"  
      by simp_all
         (metis  ilen_imap inth_imap inth_iset)
 have 9: "\<exists>x\<in>iset (suffixes xs). (\<lambda> ys. f ys) x" 
      using assms in_iset_suffixes osfx_suffix by blast
 have 10: "\<exists>x\<in>iset (suffixes xs). ((\<lambda>y. w (\<langle>y\<rangle>))\<circ>(\<lambda>s. inth s 0)) x \<and> (\<lambda> ys. f ys) x" 
      using "8" "9" by auto
 have 11: "ifilter ((\<lambda>y. w (\<langle>y\<rangle>))\<circ>(\<lambda>s. inth s 0)) (ifilter (\<lambda> ys. f ys) (suffixes xs)) =
          ifilter (\<lambda> zs. ((\<lambda>y. w (\<langle>y\<rangle>))\<circ>(\<lambda>s. inth s 0)) zs \<and> (\<lambda> ys. f ys) zs) (suffixes xs)
         " 
      using ifilter_ifilter[of "(\<lambda> ys. f ys)" "(suffixes xs)" "((\<lambda>y. w (\<langle>y\<rangle>))\<circ>(\<lambda>s. inth s 0))" ]  
           "8" "10" by blast
 have 12: "\<exists>i\<le>ilen xs. (\<lambda> zs. ((\<lambda>y. w (\<langle>y\<rangle>))\<circ>(\<lambda>s. inth s 0)) zs \<and> (\<lambda> ys. f ys) zs) (suffix i xs) "
 by (metis "10" ilen_suffixes inth_and_iset inth_suffixes) 
 have 13: " (ifilter (\<lambda> zs. ((\<lambda>y. w (\<langle>y\<rangle>))\<circ>(\<lambda>s. inth s 0)) zs \<and> (\<lambda> ys. f ys) zs) (suffixes xs))
            = ( sfxfilt xs (\<lambda> zs. ((\<lambda>y. w (\<langle>y\<rangle>))\<circ>(\<lambda>s. inth s 0)) zs \<and> (\<lambda> ys. f ys) zs)) 
          " 
      by (simp add: sfxfilt_def)
 have 14: "\<exists>i\<le>ilen xs. (\<lambda> zs. ((\<lambda>y. w (\<langle>y\<rangle>))\<circ>(\<lambda>s. inth s 0)) zs \<and> (\<lambda> ys. f ys) zs) (suffix i xs)"
      using "12" by blast
 have 15: "imap (\<lambda>s. inth s 0) 
               (( sfxfilt xs (\<lambda> zs. ((\<lambda>y. w (\<langle>y\<rangle>))\<circ>(\<lambda>s. inth s 0)) zs \<and> (\<lambda> ys. f ys) zs)) ) = 
           pifilt xs (\<lambda> zs. ((\<lambda>y. w (\<langle>y\<rangle>))\<circ>(\<lambda>s. inth s 0)) zs \<and> (\<lambda> ys. f ys) zs)
          "  
      using "14" by (simp add: sfxfilt_pifilt)
 have 16: "\<And>xs . (\<lambda> zs. ((\<lambda>y. w (\<langle>y\<rangle>))\<circ>(\<lambda>s. inth s 0)) zs \<and> (\<lambda> ys. f ys) zs) xs =
            (LIFT(f \<and> init w)) xs
          " 
      by (auto simp add: init_defs)
 have 17: "pifilt xs (\<lambda> zs. ((\<lambda>y. w (\<langle>y\<rangle>))\<circ>(\<lambda>s. inth s 0)) zs \<and> (\<lambda> ys. f ys) zs) =
           pifilt xs (LIFT(f \<and> init w))
          " 
      using "16" by presburger
 show ?thesis 
 using "11" "13" "15" "17" "2" "3" "4" "7" by auto    
qed


lemma PiAssocsem:
 "\<sigma> \<Turnstile> f \<Pi> ((init w) \<Pi>  g) =   (f \<and> (init w)) \<Pi> g"
proof (auto simp add: pi_d_def init_defs)
 fix i 
 fix ia
 assume a0: "g (pifilt (pifilt \<sigma> f) (LIFT(init w)))"
 assume a1: "i \<le> ilen \<sigma>"  
 assume a2: "f (suffix i \<sigma>)"
 assume a3: "ia \<le> ilen (pifilt \<sigma> f)"
 assume a4: "w \<langle> inth (pifilt \<sigma> f) ia\<rangle>"   
 show "\<exists>i\<le>ilen \<sigma>. f (suffix i \<sigma>) \<and> w \<langle> inth \<sigma> i\<rangle>" 
  using a0 a1 a2 a3 a4 PiAssocsem1 by fastforce
next
 fix i
 fix ia
 assume a0: "g (pifilt (pifilt \<sigma> f) (LIFT(init w)))"
 assume a1: "i \<le> ilen \<sigma>"
 assume a2: "f (suffix i \<sigma>)"
 assume a3: "ia \<le> ilen (pifilt \<sigma> f)"
 assume a4: "w \<langle> inth (pifilt \<sigma> f) ia\<rangle>"      
 show "g (pifilt \<sigma> (LIFT(f \<and> init w)))" 
  using a0 a1 a2 a3 a4 by (metis ifirst_suffix  pifilt_pifilt)
next
 fix i
 assume a0: " g (pifilt \<sigma> (LIFT(f \<and> init w)))"
 assume a1: "i \<le> ilen \<sigma>"
 assume a2: "f (suffix i \<sigma>)"
 assume a3: "w \<langle> inth \<sigma> i\<rangle>"    
 show "\<exists>i\<le>ilen (pifilt \<sigma> f). w \<langle> inth (pifilt \<sigma> f) i\<rangle>" 
  using a0 a1 a2 a3 by (metis PiAssocsem2)
next
 fix i
 assume a0: "g (pifilt \<sigma> (LIFT(f \<and> init w)))"
 assume a1: " i \<le> ilen \<sigma>"  
 assume a2: "f (suffix i \<sigma>)"
 assume a3: "w \<langle> inth \<sigma> i\<rangle>"  
 show " g (pifilt (pifilt \<sigma> f) (LIFT(init w))) " 
  using a0 a1 a2 a3 
  by (metis PiAssocsem2 ifirst_suffix pifilt_pifilt)
qed

lemma  PiAssoc:
 "\<turnstile> f \<Pi> ((init w) \<Pi>  g) =   (f \<and> (init w)) \<Pi> g"
 using PiAssocsem Valid_def by blast

subsubsection \<open>PiNotEqvDiamondAndNotPi\<close>

lemma PiNotEqvDiamondAndNotPisem:
 "\<sigma> \<Turnstile> f \<Pi> (\<not> g) = (\<diamond>f \<and> \<not>(f \<Pi> g))"
by (simp add: pi_d_def sometimes_defs ) blast


lemma PiNotEqvDiamondAndNotPi:
 "\<turnstile> f \<Pi> (\<not> g) = (\<diamond>f \<and> \<not>(f \<Pi> g))"
using PiNotEqvDiamondAndNotPisem Valid_def by blast

subsubsection \<open>PiChopDist\<close>

lemma iset_fuse:
 assumes "ilast xs = ifirst ys"
 shows   " iset (fuse xs ys) = iset xs \<union> iset ys"
using assms
proof (induction xs arbitrary: ys)
case (INil x)
then show ?case 
by (metis fuse_INil fuse_rightneutral iapp_assoc iset_iapp opfx_code(1) opfx_def sup.idem)
next
case (ICons x1a xs)
then show ?case by simp
qed

lemma ifilter_chop:
assumes "ilast xs = ifirst ys \<and> P (ilast xs)"
        "(\<exists> x \<in> iset (fuse xs ys). P x)"
        "(\<exists> x \<in> iset xs. P x)"
        "(\<exists> x \<in> iset ys. P x)"
shows "ifilter P (fuse xs ys) = fuse (ifilter P xs)  (ifilter P ys)"
using assms 
proof (induction xs arbitrary: ys)
case (INil x)
then show ?case 
by simp
next
case (ICons x1a xs)
then show ?case 
   proof (cases "(\<exists> x \<in> iset xs. P x)")
   case True
   then show ?thesis 
      using ICons.IH ICons.prems(1) ICons.prems(4) iset_fuse by fastforce
   next
   case False
   then show ?thesis 
     using ICons.prems(1) inth_iset order_refl by force
   qed
qed

lemma ifilter_chop1:
assumes " n\<le> ilen xs \<and> P (ilast (prefix n xs))"
        "(\<exists> x \<in> iset xs. P x)"
        "(\<exists> x \<in> iset (prefix n xs). P x)"
        "(\<exists> x \<in> iset (suffix n xs). P x)"
shows "ifilter P xs = fuse (ifilter P (prefix n xs))  (ifilter P (suffix n xs))"
proof -
 have 1: "(\<exists> x \<in> iset xs. P x) = (\<exists> x \<in> iset (fuse (prefix n xs) (suffix n xs)). P x)" 
     by (simp add: assms(1) fuse_prefix_suffix)
 have 2: "ilast (prefix n xs) = ifirst (suffix n xs)"
     using ilast_ifirst by blast
 have 3: "xs = fuse (prefix n xs) (suffix n xs)"
   by (simp add: assms(1) fuse_prefix_suffix) 
 have 4: "ifilter P (fuse (prefix n xs) (suffix n xs)) =
          fuse (ifilter P (prefix n xs))  (ifilter P (suffix n xs))" 
  using assms 2 3 ifilter_chop[of "(prefix n xs)" "(suffix n xs)" "P"] 
  by auto  
 show ?thesis using "3" "4" by auto
qed


lemma ifilter_chop1_prefix:
assumes " n\<le> ilen xs"
        " P (ilast (prefix n xs))"
        "(\<exists> x \<in> iset xs. P x)"
        "(\<exists> x \<in> iset (prefix n xs). P x)"
        "(\<exists> x \<in> iset (suffix n xs). P x)"
shows "prefix (ilen (ifilter P (prefix n xs))) (ifilter P xs) = 
       (ifilter P (prefix n xs))"
proof -
 have 2: "ifilter P xs = fuse (ifilter P (prefix n xs))  (ifilter P (suffix n xs))" 
   using assms ifilter_chop1 by blast
 have 3: " ilast (ifilter P (prefix n xs)) = ifirst (ifilter P (suffix n xs))"
   using assms by (metis ilast_ifirst ifilter_ifirst ifilter_ilast)
 have 4: "prefix (ilen (ifilter P (prefix n xs))) 
                 (fuse (ifilter P (prefix n xs))  (ifilter P (suffix n xs)) ) = 
         (ifilter P (prefix n xs))
         " 
     using prefix_fuse using "3" by blast
 show ?thesis by (simp add: "2" "4")
qed

lemma ifilter_chop1_suffix:
assumes " n\<le> ilen xs "
        " P (ilast (prefix n xs))"
        "(\<exists> x \<in> iset xs. P x)"
        "(\<exists> x \<in> iset (prefix n xs). P x)"
        "(\<exists> x \<in> iset (suffix n xs). P x)"
shows "suffix (ilen (ifilter P (prefix n xs))) (ifilter P xs) = 
       (ifilter P (suffix n xs))"
proof -
 have 1: "ifilter P xs = fuse (ifilter P (prefix n xs))  (ifilter P (suffix n xs))" 
   using assms ifilter_chop1 by blast
 have 3: " ilast (ifilter P (prefix n xs)) = ifirst (ifilter P (suffix n xs))"
   using assms by (metis ilast_ifirst ifilter_ifirst ifilter_ilast)
 have 4: "suffix (ilen (ifilter P (prefix n xs)))  
                 (fuse (ifilter P (prefix n xs))  (ifilter P (suffix n xs))) =
          (ifilter P (suffix n xs))
         " 
     using suffix_fuse using "3" by blast
  show ?thesis by (simp add: "1" "4")
qed


lemma PiChopDistsema:
 assumes "(\<sigma> \<Turnstile> (init w) \<Pi> (g;h))"
 shows  " (\<sigma> \<Turnstile> ((init w) \<Pi> g);((init w) \<and> ((init w) \<Pi> h)))"
proof -
 have 1: "(\<sigma> \<Turnstile> (init w) \<Pi> (g;h))" 
   using assms by auto
 have 2: "((\<exists>i\<le>ilen \<sigma>. (LIFT(init w)) (suffix i \<sigma>)) \<and>
           ((pifilt \<sigma> (LIFT(init w))) \<Turnstile> g;h)
          )"
    using "1" by (simp add: pi_d_def)
 have 3: "(\<exists>i\<le>ilen \<sigma>. (LIFT(init w)) (suffix i \<sigma>))"  
   using "2" by auto
 have 4: "((pifilt \<sigma> (LIFT(init w))) \<Turnstile> g;h)"
   using "2" by auto 
 have 5: "ifilter (\<lambda>y. w (\<langle>y\<rangle>)) \<sigma> \<Turnstile> g;h"
    using pifilt_init 
    using "2" by fastforce 
 have 6: "\<exists>n\<le>ilen (ifilter (\<lambda>y. w \<langle>y\<rangle>) \<sigma>).
            g (prefix n (ifilter (\<lambda>y. w \<langle>y\<rangle>) \<sigma>)) \<and> 
            h (suffix n (ifilter (\<lambda>y. w \<langle>y\<rangle>) \<sigma>))"
  using "5" by (simp add: chop_defs)  
 obtain n where 7: "n\<le>ilen (ifilter (\<lambda>y. w \<langle>y\<rangle>) \<sigma>) \<and>
         g (prefix n (ifilter (\<lambda>y. w \<langle>y\<rangle>) \<sigma>)) \<and> 
            h (suffix n (ifilter (\<lambda>y. w \<langle>y\<rangle>) \<sigma>))"
   using "6" by auto
 have 8: "\<exists>i\<le>ilen \<sigma>. w \<langle>inth \<sigma> i\<rangle>"
   using "3" by (auto simp add: init_defs) 
 have 9: "\<exists> x \<in> iset \<sigma>. w \<langle>x\<rangle>"
    using "8" inth_iset by blast
 have 10: "(prefix n (ifilter (\<lambda>y. w \<langle>y\<rangle>) \<sigma>)) = 
           (ifilter (\<lambda>y. w \<langle>y\<rangle>) (prefix ((inth (nfilter (\<lambda>y. w \<langle>y\<rangle>) \<sigma> 0) n) ) \<sigma>) )"
     by (simp add: "7" "9" ifilter_nfilter_prefix_1) 
 have 11: "(suffix n (ifilter (\<lambda>y. w \<langle>y\<rangle>) \<sigma>)) =
           (ifilter (\<lambda>y. w \<langle>y\<rangle>) (suffix ((inth (nfilter (\<lambda>y. w \<langle>y\<rangle>) \<sigma> 0) n) ) \<sigma>) )"
   by (simp add: "7" "9" ifilter_nfilter_suffix_1)
 have 12: "g (ifilter (\<lambda>y. w \<langle>y\<rangle>) (prefix ((inth (nfilter (\<lambda>y. w \<langle>y\<rangle>) \<sigma> 0) n) ) \<sigma>) )"
   using "10" "7" by auto
 have 13: "h (ifilter (\<lambda>y. w \<langle>y\<rangle>) (suffix ((inth (nfilter (\<lambda>y. w \<langle>y\<rangle>) \<sigma> 0) n) ) \<sigma>) )" 
    using "11" "7" by auto
 have 14: "((inth (nfilter (\<lambda>y. w \<langle>y\<rangle>) \<sigma> 0) n) ) \<le> ilen \<sigma>" 
   by (metis "7" "9" add_cancel_right_left nfilter_ilen nfilter_upper_bound)
 have 15: "w \<langle>inth (suffix ((inth (nfilter (\<lambda>y. w \<langle>y\<rangle>) \<sigma> 0) n) ) \<sigma>) 0\<rangle>" 
    by (metis (no_types, lifting) "14" "7" "9" ifilter_inth_aa ifirst_suffix 
        nfilter_ifilter nfilter_ilen nfilter_inth_n_zero)
 have 16: "(\<exists>i\<le>ilen (prefix  ((inth (nfilter (\<lambda>y. w \<langle>y\<rangle>) \<sigma> 0) n) ) \<sigma>). 
            w \<langle>inth (suffix i (prefix  ((inth (nfilter (\<lambda>y. w \<langle>y\<rangle>) \<sigma> 0) n) ) \<sigma>)) 0\<rangle>)" 
   using "14" "15" by auto
 have 17: "(\<exists>i\<le>ilen (suffix ((inth (nfilter (\<lambda>y. w \<langle>y\<rangle>) \<sigma> 0) n) ) \<sigma>).
           w \<langle>inth (suffix (i + ((inth (nfilter (\<lambda>y. w \<langle>y\<rangle>) \<sigma> 0) n) )) \<sigma>) 0\<rangle>)"
    using "15" by auto
 have 18: "(\<exists>n\<le>ilen \<sigma>.
       (\<exists>i\<le>ilen (prefix n \<sigma>). w \<langle>inth (suffix i (prefix n \<sigma>)) 0\<rangle>) \<and>
       g (ifilter (\<lambda>y. w (\<langle>y\<rangle>)) (prefix n \<sigma>)) \<and>
       w \<langle>inth (suffix n \<sigma>) 0\<rangle> \<and>
       (\<exists>i\<le>ilen (suffix n \<sigma>). w \<langle>inth (suffix (i + n) \<sigma>) 0\<rangle>) \<and> 
        h (ifilter (\<lambda>y. w (\<langle>y\<rangle>)) (suffix n \<sigma>)) )" 
     using "12" "13" "14" "15" "16" "17" by blast
 have 181: "(\<exists>n\<le>ilen \<sigma>. (\<exists> i\<le> ilen (prefix n \<sigma>). w \<langle> inth (prefix n \<sigma>) i\<rangle> )  )"
   using "18" by auto
 have 182: "(\<exists>n\<le>ilen \<sigma>. (\<exists> i\<le> ilen (suffix n \<sigma>).  w \<langle> inth \<sigma> (i + n)\<rangle> )  )" 
  using "18" by auto
 have 19: "(\<exists>n\<le>ilen \<sigma>.
       (\<exists>i\<le>ilen (prefix n \<sigma>). w \<langle>inth (suffix i (prefix n \<sigma>)) 0\<rangle>) \<and>
       g (pifilt (prefix n \<sigma>) (LIFT(init w)) ) \<and>
       w \<langle>inth (suffix n \<sigma>) 0\<rangle> \<and>
       (\<exists>i\<le>ilen (suffix n \<sigma>). w \<langle>inth (suffix (i + n) \<sigma>) 0\<rangle>) \<and> 
        h (pifilt (suffix n \<sigma>) (LIFT(init w))))"    
   using 18 pifilt_init[of _ "w"] 
    proof -
     assume a1: "\<And>xs. \<exists>i\<le>ilen xs. (suffix i xs) \<Turnstile> (init w)  \<Longrightarrow> 
                       pifilt xs (LIFT(init w)) = ifilter (\<lambda>y. w \<langle>y\<rangle>) xs"
     obtain nn :: nat and nna :: nat and nnb :: nat where
     f2: "h (ifilter (\<lambda>a. w \<langle>a\<rangle>) (suffix nn \<sigma>)) \<and> 
          (w \<langle>inth (suffix (nna + nn) \<sigma>) 0\<rangle> \<and> nna \<le> ilen (suffix nn \<sigma>)) \<and> 
          w \<langle>inth (suffix nn \<sigma>) 0\<rangle> \<and> 
          g (ifilter (\<lambda>a. w \<langle>a\<rangle>) (prefix nn \<sigma>)) \<and> 
         (w \<langle> inth (suffix nnb (prefix nn \<sigma>)) 0\<rangle> \<and> nnb \<le> ilen (prefix nn \<sigma>)) \<and> 
         nn \<le> ilen \<sigma>"
     using \<open>\<exists>n\<le>ilen \<sigma>. 
             (\<exists>i\<le>ilen (prefix n \<sigma>). w \<langle>inth (suffix i (prefix n \<sigma>)) 0\<rangle>) \<and> 
             g (ifilter (\<lambda>y. w \<langle>y\<rangle>) (prefix n \<sigma>)) \<and> 
             w \<langle>inth (suffix n \<sigma>) 0\<rangle> \<and> 
             (\<exists>i\<le>ilen (suffix n \<sigma>). w \<langle>inth (suffix (i + n) \<sigma>) 0\<rangle>) \<and> 
             h (ifilter (\<lambda>y. w \<langle>y\<rangle>) (suffix n \<sigma>))\<close> by blast 
     then have "\<exists>n na. w (prefix 0 (suffix na (suffix nn \<sigma>))) \<and> 
                       h (pifilt (suffix nn \<sigma>) (LIFT(init w))) \<and> 
                       na \<le> ilen (suffix nn \<sigma>) \<and> 
                       g (pifilt (prefix nn \<sigma>) (LIFT(init w))) \<and> 
                       n \<le> ilen (prefix nn \<sigma>) \<and> 
                       w (prefix 0 (suffix n (prefix nn \<sigma>)))"
     using a1 by (metis (no_types) init_defs prefix_zero_ifirst suffix_suffix) 
     then show ?thesis
     using f2 by blast 
    qed 
 have 20: "((\<exists>n\<le>ilen \<sigma>.
       (\<exists>i\<le>ilen (prefix n \<sigma>). (LIFT(init w)) (suffix i (prefix n \<sigma>))) \<and>
       g (pifilt (prefix n \<sigma>) (LIFT(init w))) \<and>
       (LIFT(init w)) (suffix n \<sigma>) \<and> (\<exists>i\<le>ilen (suffix n \<sigma>). (LIFT(init w)) (suffix (i + n) \<sigma>)) 
         \<and> h (pifilt (suffix n \<sigma>) (LIFT(init w)))))" 
    by (metis "19" init_defs prefix_zero_ifirst)
 have 21: "(\<sigma> \<Turnstile> ((init w) \<Pi> g);((init w) \<and> ((init w) \<Pi> h)))"
      using "20" by (simp add: chop_defs pi_d_def) 
show ?thesis 
using "21" by auto
qed       

      
lemma PiChopDistsemb:
 assumes " (\<sigma> \<Turnstile> ((init w) \<Pi> g);((init w) \<and> ((init w) \<Pi> h)))"
 shows  "(\<sigma> \<Turnstile> (init w) \<Pi> (g;h))"
proof -  
 have 1: "(\<sigma> \<Turnstile> ((init w) \<Pi> g);((init w) \<and> ((init w) \<Pi> h)))" 
   using assms by auto
 have 2: "\<exists> n \<le> ilen \<sigma>. 
            ((prefix n \<sigma>) \<Turnstile> ((init w) \<Pi> g)) \<and>
            ((suffix n \<sigma>) \<Turnstile> ((init w) \<and> ((init w) \<Pi> h)))
         " 
    using assms chop_defs by blast
 obtain n where 3: "n \<le> ilen \<sigma> \<and> ((prefix n \<sigma>) \<Turnstile> ((init w) \<Pi> g)) \<and>
            ((suffix n \<sigma>) \<Turnstile> ((init w) \<and> ((init w) \<Pi> h)))" 
   using "2" by auto
 have 4: "((\<exists>i\<le>ilen (prefix n \<sigma>). (LIFT(init w)) (suffix i (prefix n \<sigma>))) \<and>
           ((pifilt (prefix n \<sigma>) (LIFT(init w))) \<Turnstile> g)
          )" 
    by (meson "3" pi_d_def)
 have 5: "(\<exists>i\<le>ilen (prefix n \<sigma>). (LIFT(init w)) (suffix i (prefix n \<sigma>)))" 
   using "4" by auto
 have 6: "g (pifilt (prefix n \<sigma>) (LIFT(init w)))" 
   using "4" by auto
 have 7: "g (ifilter (\<lambda>y. w (\<langle>y\<rangle>)) (prefix n \<sigma>)) "
    using "5" "6" pifilt_init by (metis)
 have 8: "((\<exists>i\<le>ilen (suffix n \<sigma>). (LIFT(init w)) (suffix i (suffix n \<sigma>))) \<and>
           ((pifilt (suffix n \<sigma>) (LIFT(init w))) \<Turnstile> h)
          )" 
   by (metis "3" intensional_rews(3) pi_d_def)
 have 9: "(\<exists>i\<le>ilen (suffix n \<sigma>). (LIFT(init w)) (suffix i (suffix n \<sigma>)))"
    using "8" by auto
 have 10: "h (pifilt (suffix n \<sigma>) (LIFT(init w)))"  
   using "8" by auto
 have 11: "h (ifilter (\<lambda>y. w (\<langle>y\<rangle>)) (suffix n \<sigma>))  "
    using "10" "9" pifilt_init by metis 
 have 12: "(\<lambda>y. w (\<langle>y\<rangle>)) (ilast (prefix n \<sigma>))"
    by (metis "3" ilast_ifirst init_defs intensional_rews(3) prefix_zero_ifirst)
 have 13: "\<exists> x \<in> iset \<sigma>. (\<lambda>y. w (\<langle>y\<rangle>)) x"
   using "12" "3" inth_iset using ilast_prefix by fastforce 
 have 14: "\<exists> x \<in> iset (prefix n \<sigma>) . (\<lambda>y. w (\<langle>y\<rangle>)) x" 
   using "12" inth_iset by (metis order_refl)
 have 15: "\<exists> x \<in> iset (suffix n \<sigma>) . (\<lambda>y. w (\<langle>y\<rangle>)) x" 
   by (metis "12" "3" ifirst_suffix ilast_prefix ilen_gr_zero inth_iset)
 have 16: "(ifilter (\<lambda>y. w (\<langle>y\<rangle>)) (prefix n \<sigma>)) =
           prefix (ilen (ifilter (\<lambda>y. w (\<langle>y\<rangle>)) (prefix n \<sigma>))) (ifilter (\<lambda>y. w (\<langle>y\<rangle>)) \<sigma>)
          " 
   using "12" "13" "14" "15" "3" 
          ifilter_chop1_prefix[of "n" "\<sigma>" "(\<lambda>y. w \<langle>y\<rangle>)"] by auto
 have 17: "(ifilter (\<lambda>y. w (\<langle>y\<rangle>)) (suffix n \<sigma>)) =
           suffix (ilen (ifilter (\<lambda>y. w (\<langle>y\<rangle>)) (prefix n \<sigma>))) (ifilter (\<lambda>y. w (\<langle>y\<rangle>)) \<sigma>)
          " 
   using "12" "13" "14" "15" "3" 
          ifilter_chop1_suffix[of "n" "\<sigma>" "(\<lambda>y. w \<langle>y\<rangle>)"] by auto
 have 18: "\<exists>n\<le>ilen (ifilter (\<lambda>y. w \<langle>y\<rangle>) \<sigma>).
            g (prefix n (ifilter (\<lambda>y. w \<langle>y\<rangle>) \<sigma>)) \<and> 
            h (suffix n (ifilter (\<lambda>y. w \<langle>y\<rangle>) \<sigma>))"
        by (metis "11" "16" "17" "7" prefix_ilen_bound)
 have 19: "ifilter (\<lambda>y. w (\<langle>y\<rangle>)) \<sigma> \<Turnstile> g;h"  
   by (simp add: "18" chop_defs)
 have 20: "((pifilt \<sigma> (LIFT(init w))) \<Turnstile> g;h)" 
   by (metis (mono_tags, lifting) "18" "3" ilast_ifirst intensional_rews(3) 
       chop_fuse fuse_prefix_suffix pifilt_init)
 have 21: "(\<exists>i\<le>ilen \<sigma>. (LIFT(init w)) (suffix i \<sigma>))" 
   using "3" by auto
 show ?thesis 
 by (simp add: "20" "21" pi_d_def)
qed

lemma PiChopDistsem:
 "\<sigma> \<Turnstile> (init w) \<Pi> (g;h) = ((init w) \<Pi> g);((init w) \<and> ((init w) \<Pi> h))"
using PiChopDistsema PiChopDistsemb unl_lift2 by blast

lemma PiChopDist:
 "\<turnstile> (init w) \<Pi> (g;h) = ((init w) \<Pi> g);((init w) \<and> ((init w) \<Pi> h))"
using PiChopDistsem Valid_def by blast

subsubsection \<open>PiProp\<close>

lemma Pistate:
 "(\<sigma> \<Turnstile> (init w) \<Pi> f) = 
  ((\<exists> x \<in> iset \<sigma>. w \<langle>x\<rangle>) \<and>  ( (ifilter (\<lambda>y. w \<langle>y\<rangle>) \<sigma>) \<Turnstile> f) )"
proof -
 have 1: "(\<sigma> \<Turnstile> (init w) \<Pi> f) =
          ( (\<exists> i\<le> ilen \<sigma>.  w \<langle>inth \<sigma> i\<rangle>) \<and> ((pifilt \<sigma> (LIFT(init w))) \<Turnstile> f) )"
  by (auto simp add: pi_d_def init_defs)
 have 2: "(\<exists>i\<le>ilen \<sigma>. (LIFT(init w)) (suffix i \<sigma>)) =
          (\<exists> i\<le> ilen \<sigma>.  w \<langle>inth \<sigma> i\<rangle>)"
  by (auto simp add: init_defs) 
 have 3: "(\<exists> i\<le> ilen \<sigma>.  w \<langle>inth \<sigma> i\<rangle>) \<longrightarrow>
           (pifilt \<sigma> (LIFT(init w))) = (ifilter (\<lambda>y. w \<langle>y\<rangle>) \<sigma>)" 
   using pifilt_init using "2" by blast
 have 4: "(\<exists> i\<le> ilen \<sigma>.  w \<langle>inth \<sigma> i\<rangle>) =(\<exists> x \<in> iset \<sigma>. w \<langle>x\<rangle>) "
    using inth_and_iset by force
 show ?thesis 
 using "1" "3" "4" by auto
qed


lemma PiPropsem1a:
 "(\<sigma> \<Turnstile> f \<Pi> $p) = 
   ((\<exists>i\<le>ilen \<sigma>. f (suffix i \<sigma>)) \<and> p (inth \<sigma> (inth (nfilter f (suffixes \<sigma>) 0) 0)) )
  "
using inth_imap[of "(\<lambda>s. inth s 0)" "(ifilter f (suffixes \<sigma>))" "0"] 
using nfilter_ifilter[of "suffixes \<sigma>" "f" "0" "0" ] 
by (simp add: pi_d_def current_val_d_def pifilt_def sfxfilt_def )
   (metis in_iset_suffixes inth_imap imap_first_suffixes osfx_suffix)

lemma PiPropsem2a:
 " (\<sigma> \<Turnstile> (\<not> f) \<U> (f \<and> $p)) =
   (\<exists>k\<le>ilen \<sigma>. f (suffix k \<sigma>) \<and> p (inth \<sigma> k) \<and> (\<forall>j<k. \<not> f (suffix j \<sigma>)))"
by (simp add: until_d_def current_val_d_def)

lemma PiPropsem3a:
 assumes "(\<sigma> \<Turnstile> f \<Pi> $p)"
 shows   " (\<sigma> \<Turnstile> (\<not> f) \<U> (f \<and> $p))"
proof -
 have 1: "((\<exists>i\<le>ilen \<sigma>. f (suffix i \<sigma>)) \<and> p (inth \<sigma> (inth (nfilter f (suffixes \<sigma>) 0) 0) ) )"
    using assms  PiPropsem1a by auto
 have 2: "\<exists> x \<in> iset(suffixes \<sigma>). f x" 
   using "1" in_iset_suffixes osfx_suffix by blast
 have 3: "\<forall> x \<in> iset(nfilter f (suffixes \<sigma>) 0). f  (inth (suffixes \<sigma>) x)"
   by (simp add: "2")
 have 4: "f (suffix (inth (nfilter f (suffixes \<sigma>) 0) 0)  \<sigma>) " 
   by (metis "2" "3" add.left_neutral ilen_gr_zero nfilter_upper_bound inth_iset 
       inth_suffixes)
 have 5: "(\<forall>j< (inth (nfilter f (suffixes \<sigma>) 0) 0). \<not> f (suffix j \<sigma>)) " 
   using nfilter_not_before[of "suffixes \<sigma>" "f" ] 2 
   proof -
   have f1: "\<forall>n.\<not> n <  inth (nfilter f (suffixes \<sigma>) 0) 0 \<or> \<not> f ( inth (suffixes \<sigma>) n)"
   using \<open>\<And>i. \<lbrakk>\<exists>x\<in>iset (suffixes \<sigma>). f x; i < inth (nfilter f (suffixes \<sigma>) 0) 0\<rbrakk> \<Longrightarrow>
        \<not> f (inth (suffixes \<sigma>) i)\<close> \<open>\<exists>x\<in>iset (suffixes \<sigma>). f x\<close> by blast 
   obtain ii :: "'a interval" where
   f2: "ii \<in> iset (suffixes \<sigma>) \<and> f ii"
   using \<open>\<exists>x\<in>iset (suffixes \<sigma>). f x\<close> by blast 
   obtain nn :: "'a interval interval \<Rightarrow> 'a interval \<Rightarrow> nat" where
   "\<forall>x0 x1. (\<exists>v2\<le>ilen x0. inth x0 v2 = x1) = (nn x0 x1 \<le> ilen x0 \<and> inth x0 (nn x0 x1) = x1)"
   by moura 
   then have "nn (suffixes \<sigma>) ii \<le> ilen (suffixes \<sigma>) \<and> inth (suffixes \<sigma>) (nn (suffixes \<sigma>) ii) = ii"
   using f2 by (meson inth_and_iset) 
   then have "inth (nfilter f (suffixes \<sigma>) 0) 0 \<le> ilen (suffixes \<sigma>)"
   using f2 f1 by (metis (no_types) dual_order.trans le_less_linear) 
   then show ?thesis
   using f1 by (metis (no_types) dual_order.trans less_or_eq_imp_le inth_suffixes) 
   qed 
 have 6: "(\<exists>k\<le>ilen \<sigma>. f (suffix k \<sigma>) \<and> p (inth \<sigma> k) \<and> (\<forall>j<k. \<not> f (suffix j \<sigma>)))"
   by (metis "1" "4" "5" interval_suf_first neqE) 
 show ?thesis using 6 PiPropsem2a by metis
qed

lemma PiPropsem3b:
 assumes "(\<sigma> \<Turnstile> (\<not> f) \<U> (f \<and> $p))"
 shows   " (\<sigma> \<Turnstile> f \<Pi> $p)"
proof -
 have 1: "(\<exists>k\<le>ilen \<sigma>. f (suffix k \<sigma>) \<and> p (inth \<sigma> k) \<and> (\<forall>j<k. \<not> f (suffix j \<sigma>)))"
  using assms  PiPropsem2a by auto
 obtain k where 2: "k\<le>ilen \<sigma> \<and> f (suffix k \<sigma>) \<and> p (inth \<sigma> k) \<and> (\<forall>j<k. \<not> f (suffix j \<sigma>))"
   using 1 by auto
 have 3: "(\<exists>i\<le>ilen \<sigma>. f (suffix i \<sigma>))"
   using "2" by blast  
 have 31: "\<exists>x \<in> iset(suffixes \<sigma>). f x"
   using "2" by auto 
 have 32: "f (suffix (inth (nfilter f (suffixes \<sigma>) 0) 0) \<sigma>)"
   using nfilter_holds[of "suffixes \<sigma>" "f" "0"]  nfilter_not_before[of "suffixes \<sigma>" "f"]
    by (metis "31" diff_zero ilen_gr_zero nfilter_upper_bound inth_iset inth_suffixes 
       ordered_cancel_comm_monoid_diff_class.add_diff_inverse)
 have 4: "p (inth \<sigma> (inth (nfilter f (suffixes \<sigma>) 0) 0) )"
   by (metis "1" "31" "32" ilen_suffixes linorder_neqE_nat nfilter_not_before inth_suffixes)
 show ?thesis using 4 3 by (simp add: PiPropsem1a)
qed

lemma PiPropsema:
 "\<sigma> \<Turnstile> f \<Pi> $p = (\<not> f) \<U> (f \<and> $p)"
using PiPropsem3a PiPropsem3b unl_lift2 by blast


lemma PiProp:
 "\<turnstile> f \<Pi> $p = (\<not> f) \<U> (f \<and> $p)"
using PiPropsema Valid_def by blast

subsubsection \<open>PiNext\<close>



lemma PiNextsem1:
 "(\<sigma> \<Turnstile> f \<Pi> (\<circle> g)) =
  ((\<exists>i\<le>ilen \<sigma>. f (suffix i \<sigma>)) \<and>
    0 < ilen (ifilter f (suffixes \<sigma>)) \<and>
    g (suffix (Suc 0) (imap (\<lambda>s. inth s 0) (ifilter f (suffixes \<sigma>)))))"
by (simp add: pi_d_def next_defs pifilt_def sfxfilt_def)  

lemma PiNextsem2:
 "(\<sigma> \<Turnstile> (\<not> f) \<U> (f \<and> \<circle>(f \<Pi> g))) =
  (\<exists>k\<le>ilen \<sigma>.
       f (suffix k \<sigma>) \<and>
       k < ilen \<sigma> \<and>
       (\<exists>i\<le>ilen \<sigma> - Suc k. f (suffix (Suc (i + k)) \<sigma>)) \<and>
       g (imap (\<lambda>s. inth s 0) (ifilter f (suffixes (suffix (Suc k) \<sigma>)))) \<and> (\<forall>j<k. \<not> f (suffix j \<sigma>)))
  "
by (simp add: until_d_def next_defs pi_d_def pifilt_def sfxfilt_def) 


lemma PiNextsem3:
 assumes "(\<sigma> \<Turnstile> f \<Pi> (\<circle> g))" 
 shows    "(\<sigma> \<Turnstile> (\<not> f) \<U> (f \<and> \<circle>(f \<Pi> g)))" 
proof -
 have 1: "((\<exists>i\<le>ilen \<sigma>. f (suffix i \<sigma>)) \<and>
    0 < ilen (ifilter f (suffixes \<sigma>)) \<and>
    g (suffix (Suc 0) (imap (\<lambda>s. inth s 0) (ifilter f (suffixes \<sigma>)))))" 
   using assms PiNextsem1 by auto
 have 2: "\<exists> x \<in> iset(suffixes \<sigma>). f x" 
   using "1" in_iset_suffixes osfx_suffix by blast
 have 3: "\<forall> x \<in> iset(nfilter f (suffixes \<sigma>) 0). f  (inth (suffixes \<sigma>) x)"
   by (simp add: "2")
 have 4: "f (suffix (inth (nfilter f (suffixes \<sigma>) 0) 0)  \<sigma>) "
  by (metis "2" add.left_neutral ilen_gr_zero nfilter_ifilter nfilter_inth_n_zero 
      nfilter_upper_bound inth_suffixes sfxfilter_inth)
 have 41: "(inth (nfilter f (suffixes \<sigma>) 0) 1) \<in> iset(nfilter f (suffixes \<sigma>) 0)"
   by (metis "1" "2" One_nat_def Suc_leI nfilter_ilen inth_iset)
 have 42: "  (inth (nfilter f (suffixes \<sigma>) 0) 1) \<le> ilen (suffixes \<sigma>) "
    by (metis "1" "2" One_nat_def Suc_leI add_cancel_right_left nfilter_ilen nfilter_upper_bound) 
 have 5: "f (suffix (inth (nfilter f (suffixes \<sigma>) 0) 1)  \<sigma>)" 
    using "3" "41" "42" inth_suffixes by fastforce
 have 6: "(inth (nfilter f (suffixes \<sigma>) 0) 0) < (inth (nfilter f (suffixes \<sigma>) 0) 1)" 
   by (simp add: "1" "2" idx_nfilter_mono nfilter_ilen)
 have 7: " Suc (inth (nfilter f (suffixes \<sigma>) 0) 0) \<le> ilen \<sigma>" 
   by (metis "1" "3" ilen_ifilter_le Suc_diff_Suc diff_is_0_eq' 
       ifilter_nfilter_suffix ilen_gr_zero suffix_ilen ilen_suffixes 
       not_less_eq_eq inth_iset)
 have 8: "0 < ilen (nfilter f (suffixes \<sigma>) 0)" 
   by (simp add: "1" "2" nfilter_ilen)
 have 9: "(inth (nfilter f (suffixes \<sigma>) 0) 1) \<le> ilen \<sigma>" 
    using nfilter_upper_bound[of "suffixes \<sigma>" "f" "1" "0"] 
    by (simp add: "2" "8" Suc_leI)  
 have 10: "(Suc (inth (nfilter f (suffixes \<sigma>) 0) 0)) \<le> (inth (nfilter f (suffixes \<sigma>) 0) 1)"
    using "6" Suc_leI by blast 
 have 11: "(inth (nfilter f (suffixes \<sigma>) 0) 1) = 
            (inth (nfilter f (suffixes \<sigma>) 0) 1) - (Suc (inth (nfilter f (suffixes \<sigma>) 0) 0)) +
              (Suc (inth (nfilter f (suffixes \<sigma>) 0) 0)) " 
        using "10" by auto
  have 12: "(inth (nfilter f (suffixes \<sigma>) 0) 1) - (Suc (inth (nfilter f (suffixes \<sigma>) 0) 0)) \<le>
            ilen \<sigma> - Suc (inth (nfilter f (suffixes \<sigma>) 0) 0)"
     using "9" diff_le_mono by blast 
 have 13: "\<exists> i \<le> ilen \<sigma> - Suc (inth (nfilter f (suffixes \<sigma>) 0) 0). 
             (inth (nfilter f (suffixes \<sigma>) 0) 1) = (Suc (i + (inth (nfilter f (suffixes \<sigma>) 0) 0)))" 
  using 11 12 by auto
 have 14: "(\<exists>i\<le>ilen \<sigma> - Suc (inth (nfilter f (suffixes \<sigma>) 0) 0). 
            f (suffix (Suc (i + (inth (nfilter f (suffixes \<sigma>) 0) 0))) \<sigma>))" 
     using "13" "5" by auto 
 have 15: "(suffix (Suc 0) (imap (\<lambda>s. inth s 0) (ifilter f (suffixes \<sigma>)))) =
           (imap (\<lambda>s. inth s 0) 
                (ifilter f (suffixes (suffix (Suc (inth (nfilter f (suffixes \<sigma>) 0) 0)) \<sigma>))))" 
   using 2 8 by (simp add: "1" Suc_leI ifilter_suffixes_imap)
 have 16: "g (imap (\<lambda>s .inth s 0) 
                  (ifilter f (suffixes (suffix (Suc (inth (nfilter f (suffixes \<sigma>) 0) 0)) \<sigma>))))  "
     using "1" "15" by auto    
 have 17: "\<forall> j< (inth (nfilter f (suffixes \<sigma>) 0) 0). \<not> f (suffix j \<sigma>)"
   by (metis "7" Suc_leD Suc_leI in_iset_suffixes ilen_suffixes le_trans nfilter_not_before
        inth_suffixes osfx_suffix) 
 have 18: "(\<exists>k\<le>ilen \<sigma>.
       f (suffix k \<sigma>) \<and>
       k < ilen \<sigma> \<and>
       (\<exists>i\<le>ilen \<sigma> - Suc k. f (suffix (Suc (i + k)) \<sigma>)) \<and>
       g (imap (\<lambda>s. inth s 0) (ifilter f (suffixes (suffix (Suc k) \<sigma>)))) \<and> (\<forall>j<k. \<not> f (suffix j \<sigma>)))"
  using "14" "16" "17" "4" "7" Suc_leD Suc_le_lessD by blast 
 show ?thesis using 18 by (simp add: PiNextsem2)
qed


lemma PiNextsem4:
 assumes "(\<sigma> \<Turnstile> (\<not> f) \<U> (f \<and> \<circle>(f \<Pi> g)))" 
 shows    "(\<sigma> \<Turnstile> f \<Pi> (\<circle> g))"
proof -
 have 1: "(\<exists>k\<le>ilen \<sigma>.
       f (suffix k \<sigma>) \<and>
       k < ilen \<sigma> \<and>
       (\<exists>i\<le>ilen \<sigma> - Suc k. f (suffix (Suc (i + k)) \<sigma>)) \<and>
       g (imap (\<lambda>s. inth s 0) (ifilter f (suffixes (suffix (Suc k) \<sigma>)))) \<and> (\<forall>j<k. \<not> f (suffix j \<sigma>)))"
   using assms   by (simp add: PiNextsem2)
 obtain k where 2: "k\<le>ilen \<sigma> \<and> f (suffix k \<sigma>) \<and> k < ilen \<sigma> \<and> (\<exists>i\<le>ilen \<sigma> - Suc k. 
                    f (suffix (Suc (i + k)) \<sigma>)) \<and>
       g (imap (\<lambda>s. inth s 0) (ifilter f (suffixes (suffix (Suc k) \<sigma>)))) \<and> (\<forall>j<k. \<not> f (suffix j \<sigma>))"
   using 1 by auto 
 have 3: "\<exists> x \<in> iset (suffixes \<sigma>). f x"
   using "1" in_iset_suffixes osfx_suffix by blast 
 have 4: "\<forall> x \<in> iset(nfilter f (suffixes \<sigma>) 0). f  (inth (suffixes \<sigma>) x)"   
   by (simp add: "3")
 have 5: "f (suffix (inth (nfilter f (suffixes \<sigma>) 0) 0)  \<sigma>) "
  by (metis "3" "4" add.left_neutral ilen_gr_zero nfilter_upper_bound inth_iset 
      inth_suffixes)
 have 6: "0 < ilen (ifilter f (suffixes \<sigma>))"
  using ifilter_ilen_zero_conv_a[of "suffixes \<sigma>" "f"] 
  by (metis "2" "3" Nat.le_diff_conv2 Suc_leI Suc_n_not_le_n add_Suc_right ilen_suffixes le_add2 
      neq0_conv inth_suffixes)
 have 61: "(inth (nfilter f (suffixes \<sigma>) 0) 1) \<in> iset(nfilter f (suffixes \<sigma>) 0)"
   by (metis "3" "6" One_nat_def Suc_leI nfilter_ilen inth_iset) 
 have 62: "(inth (nfilter f (suffixes \<sigma>) 0) 1) \<le> ilen (suffixes \<sigma>)"
 by (metis "3" "6" One_nat_def Suc_leI add_cancel_right_left nfilter_ilen nfilter_upper_bound) 
 have 7: "f (suffix (inth (nfilter f (suffixes \<sigma>) 0) 1)  \<sigma>)"
   using "4" "61" "62" inth_suffixes by fastforce
 have 8: "(\<exists>i\<le>ilen \<sigma>. f (suffix i \<sigma>))" 
    using "1" by blast
 have 9: "g (suffix (Suc 0) (imap (\<lambda>s. inth s 0) (ifilter f (suffixes \<sigma>))))"
  by (metis "2" "3" "5" "6" Suc_leI ilen_suffixes linorder_neqE_nat 
      nfilter_not_before inth_suffixes ifilter_suffixes_imap) 
 have 10: "((\<exists>i\<le>ilen \<sigma>. f (suffix i \<sigma>)) \<and>
           0 < ilen (ifilter f (suffixes \<sigma>)) \<and>
           g (suffix (Suc 0) (imap (\<lambda>s. inth s 0) (ifilter f (suffixes \<sigma>)))))" 
  using "6" "8" "9" by blast
 show ?thesis using 10 
 by (simp add: PiNextsem1)
qed


lemma PiNextsem:
 " (\<sigma> \<Turnstile> f \<Pi> (\<circle> g) = (\<not> f) \<U> (f \<and> \<circle>(f \<Pi> g)))"
using PiNextsem3 PiNextsem4 
using unl_lift2 by blast


lemma PiNext:
 "\<turnstile> f \<Pi> (\<circle> g) = (\<not> f) \<U> (f \<and> \<circle>(f \<Pi> g))"
using PiNextsem Valid_def by blast

subsubsection \<open>PiUntil\<close>


lemma PiUntilDistsem1:
  "(\<sigma> \<Turnstile> f \<Pi> (g \<U> h)) =
   ((\<exists>i\<le>ilen \<sigma>. f (suffix i \<sigma>)) \<and>
    (\<exists>k\<le>ilen (ifilter f (suffixes \<sigma>)).
        h (suffix k (imap (\<lambda>s. inth s 0) (ifilter f (suffixes \<sigma>)))) \<and>
        (\<forall>j<k. g (suffix j (imap (\<lambda>s. inth s 0) (ifilter f (suffixes \<sigma>)))))))"
by (simp add: pi_d_def pifilt_def sfxfilt_def until_d_def)


lemma PiUntilDistsem2:
 " (\<sigma> \<Turnstile> ( f \<Pi> g ) \<U> ( f \<Pi> h ) ) =
   (\<exists>k\<le>ilen \<sigma>.
       (\<exists>i\<le>ilen \<sigma> - k. f (suffix (i + k) \<sigma>)) \<and>
       h (imap (\<lambda>s. inth s 0) (ifilter f (suffixes (suffix k \<sigma>)))) \<and>
       (\<forall>j<k. (\<exists>i\<le>ilen \<sigma> - j. f (suffix (i + j) \<sigma>)) \<and> 
              g (imap (\<lambda>s. inth s 0) (ifilter f (suffixes (suffix j \<sigma>))))))"
by (simp add: until_d_def pi_d_def pifilt_def sfxfilt_def)


lemma cover:
 assumes "(\<exists>i<k. ff(i) <(j::nat) \<and> j\<le> ff(Suc i))"
         "(\<forall>i<k. ff(i)<ff(Suc i))"
 shows   "ff(0)<j \<and> j \<le>ff(k)  "
using assms 
proof (induct k arbitrary:j)
case 0
then show ?case by simp
next
case (Suc k)
then show ?case 
  proof -
   have 1: "(\<exists>i<k. ff(i) <(j::nat) \<and> j\<le> ff(Suc i)) \<or> (ff (k) <j \<and> j\<le>ff(Suc k))"
     using Suc.prems(1) less_SucE by blast
   have 2: "(\<forall>i<k. ff(i)<ff(Suc i))" 
     using Suc.prems(2) by auto
   have 3: "ff k < ff (Suc k)"
    by (simp add: Suc.prems(2)) 
   have 4: "(ff(0)<j \<and> j \<le>ff(k)) \<or>  (ff (k) <j \<and> j\<le>ff(Suc k))"
      using "1" "2" Suc.hyps by blast 
   have 41: "ff(0)<j"
      using "2" "4" Suc.hyps order_refl by force 
   have 42: "j \<le>ff(Suc k)"
     using "3" "4" by linarith 
   have 5: "ff(0)<j \<and> j \<le>ff(Suc k) " 
     by (simp add: "41" "42")
   show ?thesis 
   by (simp add: "5")
  qed    
qed


lemma cover_a:
 assumes "(\<forall> j.  
          (j\<le> ff 0) \<or> (\<exists>i<k. ff(i) <(j::nat) \<and> j\<le> ff(Suc i)) 
            \<longrightarrow> gg j)
          "
         "(\<forall>i<k. ff(i)<ff(Suc i))"
 shows   "(\<forall> j< ff k. gg j)  "
proof -
 have 1: "(\<forall> j< ff 0. gg j)" 
   by (simp add: assms(1))
 have 2: "(\<forall> j. ff 0 < j \<and> j\<le> ff k \<longrightarrow> gg j)" 
   proof 
     fix j
     show "ff 0 < j \<and> j \<le> ff k \<longrightarrow> gg j"  
      using assms 
       proof (induct k arbitrary: j)
       case 0
       then show ?case by simp
       next
       case (Suc k)
       then show ?case 
         proof -
          have 21: "(\<forall> j.  
          (j\<le> ff 0) \<or> (\<exists>i<k. ff(i) <(j::nat) \<and> j\<le> ff(Suc i))
             \<or> (ff k < j \<and> j \<le> ff (Suc k)    ) 
            \<longrightarrow> gg j)" 
             using Suc.prems(1) less_SucI by blast
          have 22: "(\<forall>i<k. ff(i)<ff(Suc i))" 
            using Suc.prems(2) by auto
          have 23: "ff k < ff (Suc k)"
             by (simp add: Suc.prems(2))
          have 24: "(\<forall> j.  
                    (j\<le> ff 0) \<or> (ff 0 <j \<and> j \<le> ff k) 
                     \<or> (ff k < j \<and> j \<le> ff (Suc k)    ) 
                   \<longrightarrow> gg j)"  
            using "21" "22" Suc.hyps by blast
          have 25: "ff 0 < j \<and> j \<le> ff (Suc k) \<longrightarrow> gg j"
          using "24" not_less by blast 
          show ?thesis using "25" by blast
         qed
       qed
    qed
  show ?thesis 
  by (metis "2" assms(1) less_or_eq_imp_le linorder_neqE_nat)
qed


lemma PiUntilDistsem3:
 assumes "(\<sigma> \<Turnstile> f \<Pi> (g \<U> h))"
 shows  " (\<sigma> \<Turnstile> ( f \<Pi> g ) \<U> ( f \<Pi> h ) )"
proof -
  have 1: "((\<exists>i\<le>ilen \<sigma>. f (suffix i \<sigma>)) \<and>
    (\<exists>k\<le>ilen (ifilter f (suffixes \<sigma>)).
        h (suffix k (imap (\<lambda>s. inth s 0) (ifilter f (suffixes \<sigma>)))) \<and>
        (\<forall>j<k. g (suffix j (imap (\<lambda>s. inth s 0) (ifilter f (suffixes \<sigma>)))))))"
   using assms PiUntilDistsem1 by blast
  have 2: "\<exists> x \<in> iset(suffixes \<sigma>). f x"
     using "1" in_iset_suffixes osfx_suffix by blast
  have 3: "\<forall> x \<in> iset(nfilter f (suffixes \<sigma>) 0). f  (inth (suffixes \<sigma>) x)"
   by (simp add: "2")
  have 4: "(\<exists>k\<le>ilen (ifilter f (suffixes \<sigma>)).
        h (suffix k (imap (\<lambda>s. inth s 0) (ifilter f (suffixes \<sigma>)))) \<and>
        (\<forall>j<k. g (suffix j (imap (\<lambda>s. inth s 0) (ifilter f (suffixes \<sigma>))))))"
    using 1 by auto
  obtain k where 5: "k\<le>ilen (ifilter f (suffixes \<sigma>)) \<and> 
                     h (suffix k (imap (\<lambda>s. inth s 0) (ifilter f (suffixes \<sigma>)))) \<and>
                     (\<forall>j<k. g (suffix j (imap (\<lambda>s. inth s 0) (ifilter f (suffixes \<sigma>)))))"
    using 4 by auto
  have 6: "f (suffix (inth (nfilter f (suffixes \<sigma>) 0) k)  \<sigma>) "   
    using 2 3 5 nfilter_upper_bound[of "suffixes \<sigma>" "f" "k" "0"]    
    by (metis (no_types, lifting) "2" "3" "5" add.left_neutral nfilter_ilen inth_iset inth_suffixes)
  have 7: "k=0 \<longrightarrow> (\<exists>i\<le>ilen \<sigma> - k. f (suffix (i + k) \<sigma>)) \<and>
       h (imap (\<lambda>s. inth s 0) (ifilter f (suffixes (suffix k \<sigma>)))) \<and>
       (\<forall>j<k. (\<exists>i\<le>ilen \<sigma> - j. f (suffix (i + j) \<sigma>)) \<and> 
              g (imap (\<lambda>s. inth s 0) (ifilter f (suffixes (suffix j \<sigma>)))))"
   using "1" "5" by auto
  have 71:" k>0 \<longrightarrow> (inth (nfilter f (suffixes \<sigma>) 0) (k -1)) \<in> iset(nfilter f (suffixes \<sigma>) 0)"
    by (metis "2" "5" diff_le_self le_trans nfilter_ilen inth_iset) 
  have 8: "k>0 \<longrightarrow> f (suffix (inth (nfilter f (suffixes \<sigma>) 0) (k -1))  \<sigma>)" 
   by (metis "2" "3" "71" add.left_neutral inth_and_iset nfilter_upper_bound inth_suffixes)
  have 9: "k>0 \<longrightarrow> (inth (nfilter f (suffixes \<sigma>) 0) (k -1)) < (inth (nfilter f (suffixes \<sigma>) 0) k)"    
     by (metis "2" "5" One_nat_def Suc_diff_Suc Suc_le_lessD diff_zero idx_nfilter_mono 
         nfilter_ilen)
  have 10: "k>0 \<longrightarrow>  (inth (nfilter f (suffixes \<sigma>) 0) (k -1)) \<le> ilen \<sigma>"
    using  nfilter_upper_bound[of "suffixes \<sigma>" "f" "k-1" "0"] 
    by (simp add: "2" "5" Suc_leD nfilter_ilen) 
  have 11: "k>0 \<longrightarrow> (inth (nfilter f (suffixes \<sigma>) 0) k) \<le> ilen \<sigma>" 
    using nfilter_upper_bound[of "suffixes \<sigma>" "f" "k" "0"] 
    by (simp add: "2" "5" nfilter_ilen)
  have 12: "k>0 \<longrightarrow> 
             h (imap (\<lambda>s. inth s 0) 
                    (ifilter f (suffixes (suffix (inth (nfilter f (suffixes \<sigma>) 0) k) \<sigma>))))" 
    using ifilter_nfilter_suffix[of "f" "(suffixes \<sigma>)" "k"] 
     by (metis "11" "5" "6" ilen_suffixes imap_suffix inth_suffixes 
         suffix_suffixes)  
  have 121: "k>0 \<longrightarrow> 
             h (imap (\<lambda>s. inth s 0) 
                    (ifilter f (suffixes (suffix (Suc (inth (nfilter f (suffixes \<sigma>) 0) (k-1))) \<sigma>))))" 
     by (metis "2" "5" Suc_diff_1 ifilter_suffixes_imap)
  have 13: "k>0 \<longrightarrow> (\<exists>i\<le>ilen \<sigma> - (inth (nfilter f (suffixes \<sigma>) 0) k). 
                      f (suffix (i + (inth (nfilter f (suffixes \<sigma>) 0) k)) \<sigma>))"
        using "6" by auto     
  have 131: "k>0 \<longrightarrow> (\<exists>i\<le>ilen \<sigma> - (Suc (inth (nfilter f (suffixes \<sigma>) 0) (k-1))). 
                       f (suffix (i + (Suc (inth (nfilter f (suffixes \<sigma>) 0) (k-1)))) \<sigma>))"
    using "11" "6" "9" diff_le_mono by fastforce 
  have 14: "k>0 \<longrightarrow> (\<forall> j< (inth (nfilter f (suffixes \<sigma>) 0) k). 
                      (\<exists>i\<le>ilen \<sigma> - j. f (suffix (i + j) \<sigma>)))" 
      using "11" "6" diff_le_mono by fastforce
  have 141: "k>0 \<longrightarrow> (\<forall> j< (Suc (inth (nfilter f (suffixes \<sigma>) 0) (k-1))). 
                        (\<exists>i\<le>ilen \<sigma> - j. f (suffix (i + j) \<sigma>)))" 
    using "14" "9" by auto
  have 15: "(\<forall>j<k. g (suffix j (imap (\<lambda>s. inth s 0) (ifilter f (suffixes \<sigma>)))))" 
     using "5" by blast
  have 151: "(\<forall>j<k. g (imap (\<lambda>s. inth s 0) (suffix j  (ifilter f (suffixes \<sigma>)))))" 
   by (simp add: "5" less_le_trans less_or_eq_imp_le imap_suffix)
  have 152: "(\<forall>j<k. (suffix j  (ifilter f (suffixes \<sigma>))) = 
                    (ifilter f (suffix (inth (nfilter f (suffixes \<sigma>) 0) j) (suffixes \<sigma>))))"
     by (meson "2" "5" ifilter_nfilter_suffix_1 le_trans less_imp_le_nat)
  have 16: "(\<forall>j<k. g  (imap (\<lambda>s. inth s 0) 
                           (ifilter f (suffix (inth (nfilter f (suffixes \<sigma>) 0) j) (suffixes \<sigma>))))) "
  using 151 152 ifilter_nfilter_suffix by simp
  have 1610: " (inth (nfilter f (suffixes \<sigma>) 0) k) \<le> ilen(suffixes \<sigma>)" 
    by (metis "2" "5" diff_zero ilen_gr_zero nfilter_ilen 
        nfilter_upper_bound ordered_cancel_comm_monoid_diff_class.add_diff_inverse)
  have 1611: "(\<forall>j<k.  (inth (nfilter f (suffixes \<sigma>) 0) j) <   (inth (nfilter f (suffixes \<sigma>) 0) k)) "
   by (simp add: "2" "5" idx_nfilter_gr nfilter_ilen) 
  have 1612: "(\<forall>j<k.  (inth (nfilter f (suffixes \<sigma>) 0) j) \<le> ilen(suffixes \<sigma>))"
    using "1610" "1611" by auto
  have 161: "(\<forall>j<k. ( (suffix (inth (nfilter f (suffixes \<sigma>) 0) j) (suffixes \<sigma>))) =
                    ( (suffixes (suffix (inth (nfilter f (suffixes \<sigma>) 0) j) \<sigma>))))"
    using suffix_suffixes using "1612" by blast
  have 17: "(\<forall>j<k. g  (imap (\<lambda>s. inth s 0) 
                      (ifilter f (suffixes (suffix (inth (nfilter f (suffixes \<sigma>) 0) j) \<sigma>)))))" 
       using "16" "161" by auto
  have 18: "k>0 \<longrightarrow>  
            g  (imap (\<lambda>s. inth s 0) 
               (ifilter f (suffixes (suffix (inth (nfilter f (suffixes \<sigma>) 0) (k-1)) \<sigma>))))" 
   using 17 by simp
  have 19: "k>0 \<longrightarrow> 
            g  (imap (\<lambda>s. inth s 0) 
                    (ifilter f (suffixes (suffix (inth (nfilter f (suffixes \<sigma>) 0) 0) \<sigma>))))"
     using "17" by blast 
  have 20: "k>0 \<longrightarrow> (ifilter f (suffixes (suffix (inth (nfilter f (suffixes \<sigma>) 0) 0) \<sigma>))) =
                    (ifilter f (suffix (inth (nfilter f (suffixes \<sigma>) 0) 0) (suffixes \<sigma>)))" 
       using "161" by auto 
  have 21: "k>0 \<longrightarrow> (\<forall> j\<le> (inth (nfilter f (suffixes \<sigma>) 0 ) 0). 
                        ( (suffixes (suffix j \<sigma>))) =
                        ( (suffix j (suffixes \<sigma>)))) "
       using "1612" le_trans suffix_suffixes by fastforce
  have 22: "k>0 \<longrightarrow> (\<forall> j\<le> (inth (nfilter f (suffixes \<sigma>) 0 ) 0). 
                        (ifilter f (suffixes (suffix j \<sigma>))) =
                        (ifilter f (suffix j (suffixes \<sigma>))))" 
        using "21" by auto
  have 23: " k>0 \<longrightarrow> 
              (\<forall> j\<le> (inth (nfilter f (suffixes \<sigma>) 0 ) 0).  
                (imap (\<lambda>s. inth s 0) 
                     (ifilter f (suffixes (suffix (inth (nfilter f (suffixes \<sigma>) 0) 0) \<sigma>)))) =
                (imap (\<lambda>s. inth s 0) (ifilter f (suffixes (suffix j \<sigma>))))
               ) "
     by (simp add: "2" "21" ifilter_suffixes_imap_help_0)
  have 24: "k>0 \<longrightarrow> 
              (\<forall> j\<le> (inth (nfilter f (suffixes \<sigma>) 0 ) 0).  
                 g (imap (\<lambda>s. inth s 0) 
                        (ifilter f (suffixes (suffix (inth (nfilter f (suffixes \<sigma>) 0) 0) \<sigma>)))) =
                 g (imap (\<lambda>s. inth s 0) (ifilter f (suffixes (suffix j \<sigma>))))
              )" 
    using "23" by auto
  have 241: "k>0 \<longrightarrow> (\<forall> j\<le> (inth (nfilter f (suffixes \<sigma>) 0 ) 0).  
                      g (imap (\<lambda>s. inth s 0) (ifilter f (suffixes (suffix j \<sigma>)))) ) "
     using "19" "24" by blast 
  have 25: "k>0 \<longrightarrow> (\<forall> i < k -1. 
                      (\<forall> l. l\<le>  (inth (nfilter f (suffixes \<sigma>) 0) (Suc i) ) \<and>
                            (inth (nfilter f (suffixes \<sigma>) 0) i ) < l \<longrightarrow>  
                       (ifilter f (suffix (inth(nfilter f (suffixes \<sigma>) 0) (Suc i)) (suffixes \<sigma>)) =
                       (ifilter f (suffix l (suffixes  \<sigma>)))     ) )  ) "
       proof 
        assume "k>0"
        show "(\<forall> i < k -1. 
                      (\<forall> l. l\<le>  (inth (nfilter f (suffixes \<sigma>) 0) (Suc i) ) \<and>
                            (inth (nfilter f (suffixes \<sigma>) 0) i ) < l \<longrightarrow>  
                       (ifilter f (suffix (inth(nfilter f (suffixes \<sigma>) 0) (Suc i)) (suffixes \<sigma>)) =
                       (ifilter f (suffix l (suffixes  \<sigma>)))     ) )  )"
        proof 
         fix i
         show "i < k - 1 \<longrightarrow>
                 (\<forall>l. l \<le> inth (nfilter f (suffixes \<sigma>) 0) (Suc i) \<and> 
                      inth (nfilter f (suffixes \<sigma>) 0) i < l \<longrightarrow>
                      ifilter f (suffix (inth (nfilter f (suffixes \<sigma>) 0) (Suc i)) (suffixes \<sigma>)) =
                      ifilter f (suffix l (suffixes \<sigma>))) "
         proof -
          have 251: "k=1 \<longrightarrow> i < k - 1 \<longrightarrow>
                     (\<forall>l. l \<le> inth (nfilter f (suffixes \<sigma>) 0) (Suc i) \<and> 
                          inth (nfilter f (suffixes \<sigma>) 0) i < l \<longrightarrow>
                        ifilter f (suffix (inth (nfilter f (suffixes \<sigma>) 0) (Suc i)) (suffixes \<sigma>)) =
                        ifilter f (suffix l (suffixes \<sigma>)))" 
             by auto
          have 252: "k>1 \<longrightarrow> i < k - 1 \<longrightarrow>
                      (\<forall>l. l \<le> inth (nfilter f (suffixes \<sigma>) 0) (Suc i) \<and> 
                           inth (nfilter f (suffixes \<sigma>) 0) i < l \<longrightarrow>
                          ifilter f (suffix (inth (nfilter f (suffixes \<sigma>) 0) (Suc i)) (suffixes \<sigma>)) =
                          ifilter f (suffix l (suffixes \<sigma>)))"
            proof
             assume "k>1"
             show "i < k - 1 \<longrightarrow> 
                          (\<forall>l. l \<le> inth (nfilter f (suffixes \<sigma>) 0) (Suc i) \<and> 
                                inth (nfilter f (suffixes \<sigma>) 0) i < l \<longrightarrow>
                           ifilter f (suffix (inth (nfilter f (suffixes \<sigma>) 0) (Suc i)) (suffixes \<sigma>)) =
                           ifilter f (suffix l (suffixes \<sigma>))
                          )"   
             proof
               assume "i < k - 1"
               show "(\<forall>l. l \<le> inth (nfilter f (suffixes \<sigma>) 0) (Suc i) \<and> 
                          inth (nfilter f (suffixes \<sigma>) 0) i < l \<longrightarrow>
                       ifilter f (suffix (inth (nfilter f (suffixes \<sigma>) 0) (Suc i)) (suffixes \<sigma>)) =
                       ifilter f (suffix l (suffixes \<sigma>)))"
               proof
               fix l         
                show "l \<le> inth (nfilter f (suffixes \<sigma>) 0) (Suc i) \<and> 
                      inth (nfilter f (suffixes \<sigma>) 0) i < l \<longrightarrow>
                       ifilter f (suffix (inth (nfilter f (suffixes \<sigma>) 0) (Suc i)) (suffixes \<sigma>)) =
                       ifilter f (suffix l (suffixes \<sigma>))"
                 using  ifilter_suffixes_imap_help_j[of "l" "f" "suffixes \<sigma>" "i"] 
                  using "2" "5" \<open>i < k - 1\<close> by linarith  
                qed
            qed
          qed
        show ?thesis 
        by (simp add: "252")
      qed
     qed
   qed  
  have 261: "k>0 \<longrightarrow> (\<forall> i < k -1. 
                      (\<forall> l. l\<le>  (inth (nfilter f (suffixes \<sigma>) 0) (Suc i) ) \<and>
                            (inth (nfilter f (suffixes \<sigma>) 0) i ) < l \<longrightarrow> 
                            l \<le> ilen (suffixes \<sigma>) ))
            "
       by (metis "1612" Suc_diff_1 Suc_mono le_eq_less_or_eq less_le_trans)
 have 262: "k>0 \<longrightarrow> (\<forall> i < k -1. 
                      (\<forall> l. l\<le>  (inth (nfilter f (suffixes \<sigma>) 0) (Suc i) ) \<and>
                            (inth (nfilter f (suffixes \<sigma>) 0) i ) < l \<longrightarrow> 
                            (suffix l (suffixes  \<sigma>)) = (suffixes (suffix l \<sigma>)) ))"
     using "261" suffix_suffixes by blast 
  have 26: "k>0 \<longrightarrow> (\<forall> i < k -1. 
                      (\<forall> l. l\<le>  (inth (nfilter f (suffixes \<sigma>) 0) (Suc i) ) \<and>
                            (inth (nfilter f (suffixes \<sigma>) 0) i ) < l \<longrightarrow>  
                       (imap (\<lambda>s. inth s 0) 
                         (ifilter f (suffix (inth(nfilter f (suffixes \<sigma>) 0) (Suc i)) (suffixes \<sigma>))))=
                       (imap (\<lambda>s. inth s 0) (ifilter f (suffixes  (suffix l \<sigma>)))     ) )  ) "
     using "25" "262" by auto
  have 27: "k>0 \<longrightarrow> (\<forall> i < k -1. 
                      (\<forall> l. l\<le>  (inth (nfilter f (suffixes \<sigma>) 0) (Suc i) ) \<and>
                            (inth (nfilter f (suffixes \<sigma>) 0) i ) < l \<longrightarrow>  
                      g (imap (\<lambda>s. inth s 0) 
                         (ifilter f 
                            (suffix (inth(nfilter f (suffixes \<sigma>) 0) (Suc i)) (suffixes \<sigma>))))))"
     by (simp add: "16")
  have 28:  "k>0 \<longrightarrow> (\<forall> i < k -1. 
                      (\<forall> l. l\<le>  (inth (nfilter f (suffixes \<sigma>) 0) (Suc i) ) \<and>
                            (inth (nfilter f (suffixes \<sigma>) 0) i ) < l \<longrightarrow>  
                      g (imap (\<lambda>s. inth s 0) (ifilter f (suffixes  (suffix l \<sigma>)))     )))  "
    using "25" "262" "27" by auto
  have 281: "k>0 \<longrightarrow> 
             (\<forall>j. 
                (j \<le> (inth (nfilter f (suffixes \<sigma>) 0) 0 ) \<or>
                (\<exists> i. i < k -1 \<and> 
                     j\<le>  (inth (nfilter f (suffixes \<sigma>) 0) (Suc i) ) \<and>
                     (inth (nfilter f (suffixes \<sigma>) 0) i ) < j)  ) 
                     \<longrightarrow>  g (imap (\<lambda>s. inth s 0) (ifilter f (suffixes  (suffix j \<sigma>)))     )) "
       using 241 28 by blast
  have 282: "k>0 \<longrightarrow>  (\<forall>i<k - 1. 
                        inth (nfilter f (suffixes \<sigma>) 0) i < inth (nfilter f (suffixes \<sigma>) 0) (Suc i))"
     by (metis "2" "5" Suc_diff_1 Suc_leI idx_nfilter_gr le_SucI le_trans lessI nfilter_ilen)
  have 29: "k>0 \<longrightarrow> (\<forall>j< (Suc (inth (nfilter f (suffixes \<sigma>) 0) (k-1))).  
                        g (imap (\<lambda>s. inth s 0) (ifilter f (suffixes (suffix j \<sigma>)))))"     
     using 281 282 cover_a[of "\<lambda> x. inth (nfilter f (suffixes \<sigma>) 0) x " "k-1" 
                              "(\<lambda> j. g (imap (\<lambda>s. inth s 0) (ifilter f (suffixes (suffix j \<sigma>)))))"]
     using "18" less_antisym by blast
  have 30: "k>0 \<longrightarrow> (\<exists>k\<le>ilen \<sigma>.
       (\<exists>i\<le>ilen \<sigma> - k. f (suffix (i + k) \<sigma>)) \<and>
       h (imap (\<lambda>s. inth s 0) (ifilter f (suffixes (suffix k \<sigma>)))) \<and>
       (\<forall>j<k. (\<exists>i\<le>ilen \<sigma> - j. f (suffix (i + j) \<sigma>)) \<and> 
              g (imap (\<lambda>s. inth s 0) (ifilter f (suffixes (suffix j \<sigma>))))))" 
    using 29 121 131 141 
    by (meson "11" "9" Suc_leI less_le_trans)
 have 31: "(\<exists>k\<le>ilen \<sigma>.
       (\<exists>i\<le>ilen \<sigma> - k. f (suffix (i + k) \<sigma>)) \<and>
       h (imap (\<lambda>s. inth s 0) (ifilter f (suffixes (suffix k \<sigma>)))) \<and>
       (\<forall>j<k. (\<exists>i\<le>ilen \<sigma> - j. f (suffix (i + j) \<sigma>)) \<and> 
              g (imap (\<lambda>s. inth s 0) (ifilter f (suffixes (suffix j \<sigma>))))))" 
  using 30 7 by blast
  show ?thesis using 31
  by (simp add: PiUntilDistsem2)
qed 



lemma PiUntilDistsem4:
 assumes "(\<sigma> \<Turnstile> ( f \<Pi> g ) \<U> ( f \<Pi> h ) )"
 shows  "(\<sigma> \<Turnstile> f \<Pi> (g \<U> h)) "
proof -
 have 1: "(\<exists>k\<le>ilen \<sigma>.
       (\<exists>i\<le>ilen \<sigma> - k. f (suffix (i + k) \<sigma>)) \<and>
       h (imap (\<lambda>s. inth s 0) (ifilter f (suffixes (suffix k \<sigma>)))) \<and>
       (\<forall>j<k. (\<exists>i\<le>ilen \<sigma> - j. f (suffix (i + j) \<sigma>)) \<and> 
              g (imap (\<lambda>s. inth s 0) (ifilter f (suffixes (suffix j \<sigma>))))))" 
  using assms by (simp add: PiUntilDistsem2)
 obtain k where 2: "k\<le>ilen \<sigma> \<and> (\<exists>i\<le>ilen \<sigma> - k. f (suffix (i + k) \<sigma>)) \<and>
     h (imap (\<lambda>s. inth s 0) (ifilter f (suffixes (suffix k \<sigma>)))) \<and>
     (\<forall>j<k. (\<exists>i\<le>ilen \<sigma> - j. f (suffix (i + j) \<sigma>)) \<and> 
        g (imap (\<lambda>s. inth s 0) (ifilter f (suffixes (suffix j \<sigma>)))))" 
   using 1 by auto
 have 3: "(\<exists>i\<le>ilen \<sigma> - k. f (suffix (i + k) \<sigma>))"
    using 2 by auto
 have 4: "k \<le> ilen \<sigma>"
   using 2 by auto
 obtain i where 5: "i\<le>ilen \<sigma> - k \<and> f (suffix (i + k) \<sigma>)"
   using 3 by auto 
 have 6: " i+k \<le> ilen \<sigma>"
    using 4 5 Nat.le_diff_conv2 by blast
 have 61: "\<exists>x \<in> iset (suffixes (suffix k \<sigma>)). f x"
   by (metis  "5" in_iset_suffixes suffix_suffix osfx_suffix)
 have 7: "(\<exists>i\<le>ilen \<sigma>. f (suffix i \<sigma>))"
   using "5" "6" by blast
 have 71: "\<exists>x \<in> iset (suffixes \<sigma>). f x"
   using "5" "6" in_iset_suffixes osfx_suffix by blast 
 have 72: "\<exists> x \<in> iset(nfilter f (suffixes \<sigma>) 0). f (inth (suffixes \<sigma>) x) " 
   by (metis "5" "6" cancel_comm_monoid_add_class.diff_cancel ilen_suffixes le_refl 
       nfilter_holds_not_a inth_iset inth_suffixes 
       ordered_cancel_comm_monoid_diff_class.add_diff_inverse)
 have 73: "f (suffix (inth (nfilter f (suffixes (suffix k \<sigma>)) 0) 0)  (suffix k \<sigma>)) "
      using nfilter_holds[of "suffixes (suffix k \<sigma>)" "f" "0"]
      by (metis "61" add.left_neutral diff_zero inth_iset inth_suffixes nfilter_upper_bound zero_le) 
 have 74: "f (suffix ((inth (nfilter f (suffixes (suffix k \<sigma>)) 0) 0)+k)  \<sigma>) "
     using "73" by auto 
 have 75: "f (suffix k (suffix (inth (nfilter f (suffixes (suffix k \<sigma>)) 0) 0) \<sigma>))  "
    by (metis "73" add.commute suffix_suffix) 
 have 8: "h (imap (\<lambda>s. inth s 0) (ifilter f (suffixes (suffix k \<sigma>))))"  
    using 2 by auto
 have 9: "(ifilter f (suffixes (suffix k \<sigma>))) = (ifilter f (suffix k (suffixes \<sigma>)))"
   by (simp add: "4" suffix_suffixes)
 have 10: "h (imap (\<lambda>s. inth s 0) 
             (suffix (ilen(ifilter f (suffixes \<sigma>)) - ilen (ifilter f (suffixes (suffix k \<sigma>))))
                     (ifilter f (suffixes \<sigma>))))"
   using "2" "61" sfxfilter_suffix_suffix by fastforce 
 have 11: "h (suffix (ilen(ifilter f (suffixes \<sigma>)) - ilen (ifilter f (suffixes (suffix k \<sigma>))))
                     (imap (\<lambda>s. inth s 0) (ifilter f (suffixes \<sigma>))))" 
  by (metis "10" diff_le_self imap_suffix)
 have 12: "(\<forall>j<k. (\<exists>i\<le>ilen \<sigma> - j. f (suffix (i + j) \<sigma>)) \<and> 
              g (imap (\<lambda>s. inth s 0) (ifilter f (suffixes (suffix j \<sigma>)))))"
  using "2" by blast   
 have 13: "(\<forall>j<k. (\<exists>x \<in> iset(suffixes (suffix j \<sigma>)). f x) \<and> 
              g (imap (\<lambda>s. inth s 0) (ifilter f (suffixes (suffix j \<sigma>)))))"
    by (metis "2" suffix_ilen suffix_suffix ilen_suffixes inth_iset inth_suffixes) 
 have 14: "(\<forall>j<k. (\<exists>x \<in> iset(suffixes (suffix j \<sigma>)). f x) \<and> 
              g (imap (\<lambda>s. inth s 0) 
                 (suffix (ilen(ifilter f (suffixes \<sigma>)) - ilen (ifilter f (suffixes (suffix j \<sigma>))))
                     (ifilter f (suffixes \<sigma>)))))"
    using 13 sfxfilter_suffix_suffix 
    by (metis (no_types, lifting) "2" dual_order.strict_iff_order less_le_trans)
 have 15: "(\<forall>j<k. (\<exists>x \<in> iset(suffixes (suffix j \<sigma>)). f x) \<and>
              g (suffix (ilen(ifilter f (suffixes \<sigma>)) - ilen (ifilter f (suffixes (suffix j \<sigma>))))
                     (imap (\<lambda>s. inth s 0)  (ifilter f (suffixes \<sigma>)))))" 
  by (metis "14" diff_le_self imap_suffix)
 have 150: "k=0 \<longrightarrow>(\<forall>jj<(ilen(ifilter f (suffixes \<sigma>)) - ilen (ifilter f (suffixes (suffix k \<sigma>)))). 
             g (suffix jj (imap (\<lambda>s. inth s 0) (ifilter f (suffixes \<sigma>))))) "
     by auto
 have 151: "(\<forall>jj<(ilen(ifilter f (suffixes \<sigma>)) - ilen (ifilter f (suffixes (suffix k \<sigma>)))). 
               (suffix jj (ifilter f (suffixes \<sigma>)))  =
               ifilter f (suffixes (suffix ((ilen \<sigma>) - ilen (inth (ifilter f (suffixes \<sigma>)) jj)) \<sigma>))
             )"
      by (simp add: "71" sfxfilter_suffix_suffix_a)
 have 153: "(\<forall>jj<(ilen(ifilter f (suffixes \<sigma>)) - ilen (ifilter f (suffixes (suffix k \<sigma>)))).
            g (imap (\<lambda>s. inth s 0) 
                 (ifilter f 
                   (suffixes (suffix ((ilen \<sigma>) - ilen (inth (ifilter f (suffixes \<sigma>)) jj))   \<sigma>))))
            )" 
       using "13" sfx_suffix_upperbound by blast 
 have 1530: "(\<forall>jj<(ilen(ifilter f (suffixes \<sigma>)) - ilen (ifilter f (suffixes (suffix k \<sigma>)))).
              g (imap (\<lambda>s. inth s 0) (suffix jj  (ifilter f (suffixes \<sigma>)))))"
        by (simp add: "151" "153")
 have 1531: "(\<forall>jj<(ilen(ifilter f (suffixes \<sigma>)) - ilen (ifilter f (suffixes (suffix k \<sigma>)))).
              g  (suffix jj  (imap (\<lambda>s. inth s 0) (ifilter f (suffixes \<sigma>)))))"
     by (metis "1530" diff_le_self suffix_ilen suffix_ilen_code less_le_trans 
         less_numeral_extra(3) imap_suffix zero_less_diff)
 have 154: "((\<exists>i\<le>ilen \<sigma>. f (suffix i \<sigma>)) \<and>
    (\<exists>k\<le>ilen (ifilter f (suffixes \<sigma>)).
        h (suffix k (imap (\<lambda>s. inth s 0) (ifilter f (suffixes \<sigma>)))) \<and>
        (\<forall>j<k. g (suffix j (imap (\<lambda>s. inth s 0) (ifilter f (suffixes \<sigma>)))))))" 
    using "11" "1531" "7" diff_le_self by blast
 show ?thesis 
 by (simp add: "154" PiUntilDistsem1)
qed

lemma  PiUntilDistsem:
 " \<sigma> \<Turnstile>  f \<Pi> (g \<U> h) = ( f \<Pi> g ) \<U> ( f \<Pi> h ) "
using PiUntilDistsem3  PiUntilDistsem4 using unl_lift2 by blast

lemma  PiUntilDist:
 " \<turnstile>  f \<Pi> (g \<U> h) = ( f \<Pi> g ) \<U> ( f \<Pi> h ) "
using PiUntilDistsem Valid_def by blast

subsubsection \<open>PiChopstar\<close>


lemma wnextboxnotstatesem:
assumes " k \<le> ilen \<sigma>"
 shows  " (\<forall> j \<le> ilen \<sigma>. k<j \<longrightarrow> \<not> (\<lambda>y. w \<langle>y\<rangle>) (inth \<sigma> j)) =
             (LIFT(wnext (\<box> (init (\<not> w))))) (suffix k \<sigma>)  
             "
using assms  
 proof  (auto simp add: always_defs init_defs wnext_defs)
  show "\<And>j. j \<le> ilen \<sigma> \<Longrightarrow> k < j \<Longrightarrow> w \<langle>inth \<sigma> j\<rangle> \<Longrightarrow> 
        \<forall>n\<le>ilen \<sigma> - Suc k. \<not> w \<langle>inth \<sigma> (Suc (n + k))\<rangle> \<Longrightarrow> False"  
    proof (cases k)
     show "\<And>j. j \<le> ilen \<sigma> \<Longrightarrow> k < j \<Longrightarrow> w \<langle>inth \<sigma> j\<rangle> \<Longrightarrow> 
            \<forall>n\<le>ilen \<sigma> - Suc k. \<not> w \<langle>inth \<sigma> (Suc (n + k))\<rangle> \<Longrightarrow> k = 0 \<Longrightarrow> False" 
      by simp
         (metis Suc_le_mono Suc_pred less_le_trans)
     show "\<And>j nat. j \<le> ilen \<sigma> \<Longrightarrow> k < j \<Longrightarrow> w \<langle>inth \<sigma> j\<rangle> \<Longrightarrow> 
           \<forall>n\<le>ilen \<sigma> - Suc k. \<not> w \<langle>inth \<sigma> (Suc (n + k))\<rangle> \<Longrightarrow> k = Suc nat \<Longrightarrow> False "  
      proof -
      fix j :: nat and nat :: nat
      assume a1: "k < j"
      assume a2: "\<forall>n\<le>ilen \<sigma> - Suc k. \<not> w \<langle> inth \<sigma> (Suc (n + k))\<rangle>"
      assume a3: "j \<le> ilen \<sigma>"
      assume a4: "w \<langle>inth \<sigma> j\<rangle>"
      have "Suc (k + (j - Suc k)) = j"
      using a1 by simp 
      then show False
      using a4 a3 a2 by (metis diff_add_inverse2 diff_le_mono 
                         linordered_semidom_class.add_diff_inverse not_add_less2) (* 20 ms *)
      qed
   qed
qed

lemma NotStateUntilStateAndsem:
 "(\<sigma> \<Turnstile> (init (\<not> w)) \<U> ((init w) \<and> f) ) =
   (\<exists>k\<le>ilen \<sigma>. w \<langle>inth \<sigma> k\<rangle> \<and> f (suffix k \<sigma>) \<and> (\<forall>j<k. \<not> w \<langle>inth \<sigma> j\<rangle>))
  "
by (auto simp add: until_d_def init_defs) 


lemma StateUntilEqvWPrevChopsem:
  "\<sigma> \<Turnstile> (init w) \<U> f = (wprev (\<box> (init w)));f"
by (auto simp add: until_d_def wprev_defs always_defs init_defs chop_defs)


lemma StateUntilEqvWPrevChop:
 "\<turnstile> (init w) \<U> f = (wprev (\<box> (init w)));f"
using StateUntilEqvWPrevChopsem Valid_def by blast

lemma UntilChopDist:
 " \<turnstile> (init w) \<U> (g;h) = ( (init w) \<U> g) ;h"
by (metis ChopAssoc StateUntilEqvWPrevChop inteq_reflection)



lemma PiEmptysem:
 " \<sigma> \<Turnstile> (init w) \<Pi> empty = (init (\<not> w)) \<U> ((init w) \<and> wnext (\<box> (init (\<not> w))))"
proof -
have 1: "(\<sigma> \<Turnstile> (init w) \<Pi> empty) = 
         ((\<exists>x\<in>iset \<sigma>. w \<langle>x\<rangle>) \<and> ilen (ifilter (\<lambda>y. w \<langle>y\<rangle>) \<sigma>) = 0)"
 by (simp add: init_defs empty_defs Pistate)
have 2: "((\<exists>x\<in>iset \<sigma>. w \<langle>x\<rangle>) \<and> ilen (ifilter (\<lambda>y. w \<langle>y\<rangle>) \<sigma>) = 0) =
        (\<exists> k \<le> ilen \<sigma> . (\<lambda>y. w \<langle>y\<rangle>) (inth \<sigma> k) \<and> 
             (\<forall> j. j<k \<longrightarrow> \<not> (\<lambda>y. w \<langle>y\<rangle>) (inth \<sigma> j)) \<and>
             (\<forall> j \<le> ilen \<sigma>. k<j \<longrightarrow> \<not> (\<lambda>y. w \<langle>y\<rangle>) (inth \<sigma> j)))
        " 
   by (simp add: ifilter_ilen_zero_conv_2)
have 3: "(\<exists> k \<le> ilen \<sigma> . (\<lambda>y. w \<langle>y\<rangle>) (inth \<sigma> k) \<and> 
             (\<forall> j. j<k \<longrightarrow> \<not> (\<lambda>y. w \<langle>y\<rangle>) (inth \<sigma> j)) \<and>
             (\<forall> j \<le> ilen \<sigma>. k<j \<longrightarrow> \<not> (\<lambda>y. w \<langle>y\<rangle>) (inth \<sigma> j))) =
           (\<exists>k\<le>ilen \<sigma>.
       w \<langle>inth \<sigma> k\<rangle> \<and>
       (LIFT(wnext (\<box> (init (\<not> w))))) (suffix k \<sigma>) \<and>
       (\<forall>j<k. \<not> w \<langle>inth \<sigma> j\<rangle>))
        " using wnextboxnotstatesem
     by metis
have 4: " (\<exists>k\<le>ilen \<sigma>.
       w \<langle>inth \<sigma> k\<rangle> \<and>
       (LIFT(wnext (\<box> (init (\<not> w))))) (suffix k \<sigma>) \<and>
       (\<forall>j<k. \<not> w \<langle>inth \<sigma> j\<rangle>)) =
         (\<sigma> \<Turnstile> (init (\<not> w)) \<U> ((init w) \<and> wnext (\<box> (init (\<not> w)))))
        "
  by (simp add: NotStateUntilStateAndsem) 
from 1 2 3 4 show ?thesis by auto
qed
  

lemma PiEmpty:
 "\<turnstile> (init w) \<Pi> empty = (init (\<not> w)) \<U> ((init w) \<and> wnext (\<box> (init (\<not> w))))"
using PiEmptysem Valid_def by blast


lemma StatePiBoxStatesem:
 "\<sigma> \<Turnstile> (init w) \<Pi> f = (init w) \<Pi> (f \<and> \<box> (init w))"
proof -
 have 1: "(\<sigma> \<Turnstile> (init w) \<Pi> f) =
          ((\<exists> x \<in> iset \<sigma>. w \<langle>x\<rangle>) \<and>  ( (ifilter (\<lambda>y. w \<langle>y\<rangle>) \<sigma>) \<Turnstile>  f))
        " 
   by (metis Pistate)
 have 2: "(\<sigma> \<Turnstile> (init w) \<Pi> (f \<and> \<box> (init w))) = 
          ((\<exists> x \<in> iset \<sigma>. w \<langle>x\<rangle>) \<and>  ( (ifilter (\<lambda>y. w \<langle>y\<rangle>) \<sigma>) \<Turnstile> f \<and> \<box> (init w)))
         " 
   by (metis Pistate)
 have 3: "( (ifilter (\<lambda>y. w \<langle>y\<rangle>) \<sigma>) \<Turnstile> f \<and> \<box> (init w))
         = (f (ifilter (\<lambda>y. w \<langle>y\<rangle>) \<sigma>) \<and>
    (\<forall>n\<le>ilen (ifilter (\<lambda>y. w \<langle>y\<rangle>) \<sigma>). w \<langle>inth (ifilter (\<lambda>y. w \<langle>y\<rangle>) \<sigma>) n\<rangle>))" 
   by (simp add: always_defs init_defs)
 have 4: "((\<exists> x \<in> iset \<sigma>. w \<langle>x\<rangle>) \<and> (f (ifilter (\<lambda>y. w \<langle>y\<rangle>) \<sigma>) \<and>
    (\<forall>n\<le>ilen (ifilter (\<lambda>y. w \<langle>y\<rangle>) \<sigma>). w \<langle>inth (ifilter (\<lambda>y. w \<langle>y\<rangle>) \<sigma>) n\<rangle>))) =
        ((\<exists> x \<in> iset \<sigma>. w \<langle>x\<rangle>) \<and> f (ifilter (\<lambda>y. w \<langle>y\<rangle>) \<sigma>))
         "
    by (meson ifilter_inth_aa)
 show ?thesis using "1" "2" "3" "4" by auto
qed

lemma StatePiBoxState:
 "\<turnstile> (init w) \<Pi> f = (init w) \<Pi> (f \<and> \<box> (init w))"
using StatePiBoxStatesem Valid_def by blast

lemma StatePiUntil1:
 "\<turnstile> ( (init (\<not> w)) \<U> ( (init w) \<and> (init w) \<Pi> f )) =
           (wprev (\<box> (init (\<not> w))));( (init w) \<and> (init w) \<Pi> f )  "
using StateUntilEqvWPrevChop by blast
  
lemma StatePiUntilsem2:
 "(\<sigma> \<Turnstile> (wprev (\<box> (init (\<not> w))));( (init w) \<and> (init w) \<Pi> f )) =
          (\<sigma> \<Turnstile> ((wprev (\<box> (init (\<not> w)))); ((init w) \<and> empty)) ; ((init w) \<and> (init w) \<Pi> f ) )"
by (auto simp add: chop_defs init_defs empty_defs)
   fastforce

lemma StatePiUntil2:
 "\<turnstile> (wprev (\<box> (init (\<not> w))));( (init w) \<and> (init w) \<Pi> f ) =
          ( ((wprev (\<box> (init (\<not> w)))); ((init w) \<and> empty)) ; ((init w) \<and> (init w) \<Pi> f ) )"
by (simp add: StatePiUntilsem2 Valid_def)

lemma StatePiUntil3:
 "\<turnstile> ((wprev (\<box> (init (\<not> w)))); ((init w) \<and> empty)) =
          ( ((init (\<not> w)) \<U> ((init w) \<and> wnext (\<box> (init (\<not> w))))) ; ((init w) \<and> empty)) " 
proof -
 have 1: " \<turnstile> (wprev (\<box> (init (\<not> w)))); ((init w) \<and> empty) =
             (init (\<not> w)) \<U> ((init w) \<and> empty) "
        by (meson Prop11 StateUntilEqvWPrevChop)
 have 2: "\<turnstile> ((init w) \<and> empty) = ((init w) \<and> wnext (\<box> (init (\<not> w))));((init w) \<and> empty)"
    by (auto simp add: Valid_def init_defs empty_defs wnext_defs always_defs chop_defs) 
       (metis Suc_pred eq_imp_le ilen_gr_zero le_neq_trans)
 show ?thesis by (metis "1" "2" UntilChopDist inteq_reflection)
qed


lemma StatePiUntilsem4:
"(\<sigma> \<Turnstile> ((init (\<not> w)) \<U> ((init w) \<and> wnext (\<box> (init (\<not> w))))) ; ((init w) \<and> empty)) =
          (\<sigma> \<Turnstile> ((init w) \<Pi> empty); ((init w) \<and> empty)) " 
 by (metis PiEmpty inteq_reflection)

lemma StatePiUntil4:
 "\<turnstile> ((init (\<not> w)) \<U> ((init w) \<and> wnext (\<box> (init (\<not> w))))) ; ((init w) \<and> empty) =
          ( ((init w) \<Pi> empty); ((init w) \<and> empty))"
by (simp add: StatePiUntilsem4 Valid_def)

lemma StatePiUntilsem:
 "\<sigma> \<Turnstile> (init w) \<Pi> f = (init (\<not> w)) \<U> ( (init w) \<and> (init w) \<Pi> f )"
proof -
 have 2: "(\<sigma> \<Turnstile> (init (\<not> w)) \<U> ( (init w) \<and> (init w) \<Pi> f )) =
          (\<sigma> \<Turnstile> (wprev (\<box> (init (\<not> w))));( (init w) \<and> (init w) \<Pi> f ))
         " 
   using StateUntilEqvWPrevChopsem[of "LIFT(\<not> w)" "LIFT((init w) \<and> (init w) \<Pi> f)" "\<sigma>"]
     by simp
 have 7: "(\<sigma> \<Turnstile> (wprev (\<box> (init (\<not> w))));( (init w) \<and> (init w) \<Pi> f )) = 
          (\<sigma> \<Turnstile> (((init w) \<Pi> empty); ((init w) \<and> empty));( (init w) \<and> (init w) \<Pi> f ))  
         "
     by (metis PiEmpty StatePiUntil2 StatePiUntil3 inteq_reflection) 
 have 8: "(\<sigma> \<Turnstile> (((init w) \<Pi> empty); ((init w) \<and> empty));( (init w) \<and> (init w) \<Pi> f )) =
          (\<sigma> \<Turnstile> (((init w) \<Pi> empty));( (init w) \<and> (init w) \<Pi> f ))
         " 
   by (auto simp add: chop_defs init_defs empty_defs)
      fastforce
 have 9: "(\<sigma> \<Turnstile> (((init w) \<Pi> empty));( (init w) \<and> (init w) \<Pi> f )) =
          (\<sigma> \<Turnstile> (init w) \<Pi> (empty;f))
         "    
    using PiChopDistsema PiChopDistsemb by blast
 have 10: "(\<sigma> \<Turnstile> (init w) \<Pi> (empty;f)) = (\<sigma> \<Turnstile> (init w) \<Pi> f)"
     by (metis chop_defs empty_defs prefix_ilen_good suffix_zero pi_d_def zero_order(1)) 
 show ?thesis 
 by (simp add: "10" "2" "7" "8" "9")
qed

lemma StatePiUntil:
 "\<turnstile> (init w) \<Pi> f = (init (\<not> w)) \<U> ( (init w) \<and> (init w) \<Pi> f )"
using StatePiUntilsem by blast


lemma StateAndPiEmpty:
 "\<turnstile>  ((init w) \<and> (init w) \<Pi> empty) = (w \<and> empty) ; (wnext (\<box> (init (\<not> w))))"
proof -
 have 1: "\<turnstile>  ((init w) \<and> (init w) \<Pi> empty) = 
             ((init w) \<and> (init (\<not> w)) \<U> ( (init w) \<and> (wnext (\<box> (init (\<not> w))))) ) "
   using PiEmpty by fastforce 
 have 2: "\<turnstile> ((init w) \<and> (init (\<not> w)) \<U> ( (init w) \<and> (wnext (\<box> (init (\<not> w))))) )
             = ((init w) \<and> (wnext (\<box> (init (\<not> w))))) "
    by (auto simp add: until_d_def Valid_def init_defs)
      force
 have 3: "\<turnstile> ((init w) \<and> (wnext (\<box> (init (\<not> w))))) = (w \<and> empty) ; (wnext (\<box> (init (\<not> w))))"
   by (metis InitAndEmptyEqvAndEmpty StateAndEmptyChop inteq_reflection) 
 show ?thesis 
 using "1" "2" "3" by fastforce
qed

lemma PiPowerExpandsem:
 " (\<sigma> \<Turnstile> (\<exists>k. (init w) \<Pi> (power f k))) = 
    
     (\<sigma> \<Turnstile> (init w) \<Pi> empty \<or> (\<exists>k. (init w) \<Pi> (f;(power f (k))))) "
proof -
 have 1: "(\<sigma> \<Turnstile> (\<exists>k. (init w) \<Pi> (power f k))) =
          (\<exists> k. (\<sigma> \<Turnstile> (init w) \<Pi> (power f k)))  "
     by simp 
 have 2: "(\<exists> k. (\<sigma> \<Turnstile> (init w) \<Pi> (power f k))) =
          ( (\<sigma> \<Turnstile> (init w) \<Pi> (power f 0)) \<or>     (\<exists> k. 1 \<le> k \<and> (\<sigma> \<Turnstile> (init w) \<Pi> (power f (k))))) "
         by (metis One_nat_def diff_Suc_1 le_SucE le_add1 plus_1_eq_Suc)
 have 3: "(\<sigma> \<Turnstile> (init w) \<Pi> (power f 0)) = (\<sigma> \<Turnstile> (init w) \<Pi> empty)" 
   by simp
 have 4: "(\<exists> k. 1 \<le> k \<and> (\<sigma> \<Turnstile> (init w) \<Pi> (power f (k)))) =
          (\<exists> k.  (\<sigma> \<Turnstile> (init w) \<Pi> (power f (Suc k))))" 
     by (metis le_add1 ordered_cancel_comm_monoid_diff_class.add_diff_inverse plus_1_eq_Suc)
 have 5: "(\<exists> k.  (\<sigma> \<Turnstile> (init w) \<Pi> (power f (Suc k)))) =
          (\<sigma> \<Turnstile> (\<exists>k. (init w) \<Pi> (power f (Suc k))))
         " 
      by simp
 have 6: " ((\<sigma> \<Turnstile> (init w) \<Pi> empty) \<or> (\<sigma> \<Turnstile> (\<exists>k. (init w) \<Pi> (power f (Suc k))))) = 
           (\<sigma> \<Turnstile> (init w) \<Pi> empty \<or> (\<exists>k. (init w) \<Pi> (f;power f (k))) )  "
    by auto
 show ?thesis 
 using "1" "2" "3" "4" "5" "6" by blast
qed


lemma PiPowerExpandsem1:
 " \<forall> \<sigma>. \<sigma> \<Turnstile>  (\<exists>k. (init w) \<Pi> (power f k)) = 
     ( (init w) \<Pi> empty   \<or> (\<exists>k. (init w) \<Pi> (power f (Suc k))  ) )     "
proof 
  fix \<sigma>
  show "\<sigma> \<Turnstile>  (\<exists>k. (init w) \<Pi> (power f k)) = 
              ( (init w) \<Pi> empty   \<or> (\<exists>k. (init w) \<Pi> (power f (Suc k))  ) )" 
   proof -
     have 1: "(\<sigma> \<Turnstile>  (\<exists>k. (init w) \<Pi> (power f k)) = 
              ( (init w) \<Pi> empty   \<or> (\<exists>k. (init w) \<Pi> (power f (Suc k))  ) )) 
               = ( (\<sigma> \<Turnstile>  (\<exists>k. (init w) \<Pi> (power f k))) = 
                   (\<sigma> \<Turnstile> (init w) \<Pi> empty   \<or> (\<exists>k. (init w) \<Pi> (power f (Suc k))  ) ) )
              " 
       by auto
     have 2: "( (\<sigma> \<Turnstile>  (\<exists>k. (init w) \<Pi> (power f k))) = 
                   (\<sigma> \<Turnstile> (init w) \<Pi> empty   \<or> (\<exists>k. (init w) \<Pi> (power f (Suc k))  ) ) )" 
        using PiPowerExpandsem[of "w" "f" "\<sigma>"]  by simp
     show ?thesis 
     using "1" "2" by blast
   qed
qed

lemma PiPowerExpand:
 " \<turnstile>  (\<exists>k. (init w) \<Pi> (power f k)) = 
     ( (init w) \<Pi> empty   \<or> (\<exists>k. (init w) \<Pi> (power f (Suc k))  ) )     "
using PiPowerExpandsem1[of "w" "f" ] by (auto simp add: Valid_def PiPowerExpandsem1)


lemma exists_expand_sem:
 "(\<sigma> \<Turnstile> (\<exists>k. (power ((init w) \<Pi> f \<and> fin w) k))) = 
  ((\<sigma> \<Turnstile> (power ((init w) \<Pi> f \<and> fin w) 0)) \<or> 
   (\<sigma> \<Turnstile> (\<exists>k. (power ((init w) \<Pi> f \<and> fin w) (Suc k)))))"
by (metis (no_types, lifting) not0_implies_Suc unl_Rex)  

lemma exists_expand:
 "\<turnstile>  (\<exists>k. (power ((init w) \<Pi> f \<and> fin w) k)) = 
     ((power ((init w) \<Pi> f \<and> fin w) 0) \<or> (\<exists>k. (power ((init w) \<Pi> f \<and> fin w) (Suc k))))  "
using exists_expand_sem Valid_def by fastforce



subsubsection \<open>TruePiEqv\<close>

lemma TruePiEqvsem:
 "\<sigma> \<Turnstile> #True \<Pi> f = f"
by (simp add: pi_d_def pifilt_true) auto

lemma TruePiEqv:
 "\<turnstile> (#True) \<Pi> f = f"
using  TruePiEqvsem by  (auto simp add: Valid_def)

subsubsection \<open>BoxImpEqvPi\<close>

lemma BoxImpEqvPi:
 "\<turnstile> \<box> f \<longrightarrow> g = f \<Pi> g "
by (simp add: Valid_def always_defs  pi_d_def pifilt_def sfxfilt_def) 
   (metis ifilter_True ilen_gr_zero inth_and_iset ilen_suffixes
     imap_first_suffixes inth_suffixes)


subsubsection \<open>PiEqvDiamondUPi\<close>

lemma PiEqvDiamondUPi:
 "\<turnstile> f \<Pi> g = ( \<diamond> f \<and> f \<Pi>\<^sup>u g)"
by (simp add: Valid_def upi_d_def sometimes_defs pi_d_def,blast)

subsubsection \<open>PiEqvUntilPi\<close>

lemma PiEqvUntilPi:
 "\<turnstile> (init w) \<Pi> g = (init (\<not> w)) \<U> ((init w) \<Pi> g)"
by (metis StatePiUntil UntilUntil int_eq)



subsubsection \<open>UPiEqvBoxOrPi\<close>

lemma UPiEqvBoxOrPi:
 "\<turnstile> f \<Pi>\<^sup>u g = (\<box> (\<not> f) \<or> f \<Pi> g)"
by (simp add: Valid_def upi_d_def always_defs pi_d_def,blast)

subsection \<open>Theorems\<close>

lemma UPiImpRule:
 assumes "\<turnstile> g1 \<longrightarrow> g2"
 shows   "\<turnstile> f \<Pi>\<^sup>u g1 \<longrightarrow> f \<Pi>\<^sup>u g2"
using assms 
by (meson MP PiK PiN)

lemma UPiEqvRule:
 assumes "\<turnstile> g1 = g2"
 shows   "\<turnstile> f \<Pi>\<^sup>u g1 = f \<Pi>\<^sup>u g2"
proof -
 have 1: "\<turnstile>  g1 \<longrightarrow> g2"
   using assms by (simp add: int_iffD1)
 have 2: "\<turnstile> f \<Pi>\<^sup>u g1 \<longrightarrow> f \<Pi>\<^sup>u g2"
   using "1" UPiImpRule by blast
 have 3: "\<turnstile> g2 \<longrightarrow> g1"
   using assms by (simp add: int_iffD2)
 have 4: "\<turnstile> f \<Pi>\<^sup>u g2 \<longrightarrow> f \<Pi>\<^sup>u g1"
   using "3" UPiImpRule by blast
 from 3 4 show ?thesis 
 by (simp add: "2" int_iffI)
qed

lemma PiEqvNotUPiNot:
 "\<turnstile>  f \<Pi> g = (\<not> (f \<Pi>\<^sup>u (\<not> g)))"
by (simp add: upi_d_def)

lemma NotPiEqvNotUPi:
 "\<turnstile>  f \<Pi> (\<not> g) = (\<not> (f \<Pi>\<^sup>u g))"
by (simp add: upi_d_def)

lemma UPiEqvNotPiNot:
 "\<turnstile> f \<Pi>\<^sup>u g = (\<not> (f \<Pi> (\<not> g)))"
by (simp add: upi_d_def)

lemma NotUPiEqvNotPi:
 "\<turnstile> f \<Pi>\<^sup>u (\<not>g) = (\<not> (f \<Pi> g))"
by (simp add: upi_d_def)

lemma PiImpRule:
 assumes "\<turnstile> g1 \<longrightarrow> g2"
 shows   "\<turnstile> f \<Pi> g1 \<longrightarrow> f \<Pi> g2"
proof -
 have 1: "\<turnstile> \<not> g2 \<longrightarrow> \<not> g1"
  by (simp add: assms) 
 have 2: "\<turnstile> f \<Pi>\<^sup>u (\<not> g2) \<longrightarrow> f \<Pi>\<^sup>u (\<not>g1)"
   using 1 UPiImpRule by blast
 have 3: "\<turnstile> \<not>(f \<Pi>\<^sup>u (\<not>g1)) \<longrightarrow> \<not>(f \<Pi>\<^sup>u (\<not>g2))"
  using 2 by fastforce
 from 3 show ?thesis using PiEqvNotUPiNot by fastforce
qed

lemma PiEqvRule:
 assumes "\<turnstile> g1 = g2"
 shows   "\<turnstile> f \<Pi> g1 = f \<Pi> g2"
proof -
 have 1: "\<turnstile> g1 \<longrightarrow> g2"
  using assms by (simp add: int_iffD1)
 have 2: "\<turnstile> f \<Pi> g1 \<longrightarrow> f \<Pi> g2"
   using "1" PiImpRule by blast
 have 3: "\<turnstile> g2 \<longrightarrow> g1 "
   using assms by (simp add: int_iffD2)
 have 4:"\<turnstile> f \<Pi> g2 \<longrightarrow> f \<Pi> g1"
  using "3" PiImpRule by blast
 from 2 4 show ?thesis by (simp add: int_iffI)
qed


lemma UPiAndPiImpPiAnd:
 "\<turnstile> f1 \<Pi>\<^sup>u f \<and> f1 \<Pi> (\<not> g) \<longrightarrow> f1 \<Pi> (f \<and> \<not>g)"
proof -
 have 1: "\<turnstile> (\<not>(f \<longrightarrow> g)) = (f \<and> \<not>g)"
  by fastforce
 have 2: "\<turnstile> (\<not> (f1 \<Pi>\<^sup>u (f \<longrightarrow>g))) = f1 \<Pi> (\<not>(f \<longrightarrow> g))"
   by (simp add: NotPiEqvNotUPi int_iffD1 int_iffD2 int_iffI)
 have 3: "\<turnstile>  \<not>(  f1 \<Pi>\<^sup>u f \<longrightarrow> f1 \<Pi>\<^sup>u g ) \<longrightarrow> \<not> (f1 \<Pi>\<^sup>u (f \<longrightarrow>g))"
   by (simp add: PiK) 
 have 4: "\<turnstile> (\<not>(  f1 \<Pi>\<^sup>u f \<longrightarrow> f1 \<Pi>\<^sup>u g ) ) = (f1 \<Pi>\<^sup>u f \<and> f1 \<Pi> (\<not> g))"
   using NotPiEqvNotUPi[of "f1" "g"] by fastforce  
 have 5:"\<turnstile> f1 \<Pi> (\<not>(f \<longrightarrow> g)) = f1 \<Pi> (f \<and> \<not>g)" 
    using 1 by (simp add: PiEqvRule)
 from 1 2 3 4 5 show ?thesis by fastforce
qed
 
lemma UPiAndPiImpPiAndA:
 "\<turnstile> f1 \<Pi>\<^sup>u f \<and> f1 \<Pi> g \<longrightarrow> f1 \<Pi> (f \<and> g)"
using UPiAndPiImpPiAnd[of "f1" "f" "LIFT(\<not>g)"] by fastforce


lemma PiAndPiImpPiAnd:
 "\<turnstile> f1 \<Pi> f \<and> f1 \<Pi> g \<longrightarrow> f1 \<Pi> (f \<and> g)"
proof -
 have 1: "\<turnstile>  f1 \<Pi>\<^sup>u f \<and> f1 \<Pi> g \<longrightarrow> f1 \<Pi> (f \<and> g) "
   using UPiAndPiImpPiAndA by fastforce
 have 2: "\<turnstile> f1 \<Pi> f \<longrightarrow> f1 \<Pi>\<^sup>u f"
  using PiDc by blast 
 from 1 2 show ?thesis by fastforce
qed


lemma PiAnd:
 "\<turnstile> f \<Pi> (g1 \<and> g2) = (f \<Pi> g1 \<and> f \<Pi> g2)"
proof - 
 have 1: "\<turnstile> f \<Pi> (g1 \<and> g2) \<longrightarrow> f \<Pi> g1"
   by (meson PiImpRule Prop12 int_iffD1 lift_and_com)
 have 2: "\<turnstile> f \<Pi> (g1 \<and> g2) \<longrightarrow> f \<Pi> g2"
   by (meson PiImpRule Prop12 int_iffD1 lift_and_com)
 have 3: "\<turnstile> f \<Pi> (g1 \<and> g2) \<longrightarrow> f \<Pi> g1 \<and> f \<Pi> g2" 
   using 1 2 by fastforce
 have 4: "\<turnstile> f \<Pi> g1 \<and> f \<Pi> g2 \<longrightarrow> f \<Pi> (g1 \<and> g2)"
   by (simp add: PiAndPiImpPiAnd)
 from  3 4 show ?thesis by fastforce
qed

 
  

lemma UPiAnd:
 "\<turnstile> f \<Pi>\<^sup>u (g1 \<and> g2) = ( f \<Pi>\<^sup>u g1 \<and> f \<Pi>\<^sup>u g2 )"
proof -
 have 1: "\<turnstile> f \<Pi> (\<not>g1 \<or> \<not> g2) = (f \<Pi> (\<not>g1) \<or> f \<Pi> (\<not> g2)) "
  by (simp add: PiOr)
 have 2: "\<turnstile> (\<not>(f \<Pi> (\<not>g1 \<or> \<not> g2))) = (\<not>(f \<Pi> (\<not>g1) \<or> f \<Pi> (\<not> g2)))"
   using 1 by fastforce
 have 3: "\<turnstile> (\<not>(f \<Pi> (\<not>g1 \<or> \<not> g2))) = f \<Pi>\<^sup>u (\<not>(\<not>g1 \<or> \<not> g2))"
   by (meson NotUPiEqvNotPi Prop11) 
 have 4: "\<turnstile> (\<not>(\<not>g1 \<or> \<not> g2)) = (g1 \<and> g2)"
   by fastforce
 have 5: "\<turnstile> f \<Pi>\<^sup>u (\<not>(\<not>g1 \<or> \<not> g2)) = f \<Pi>\<^sup>u (g1 \<and> g2)"
  using 4 by (simp add: UPiEqvRule)
 have 6: "\<turnstile> (\<not>(f \<Pi> (\<not>g1) \<or> f \<Pi> (\<not> g2))) = (\<not>(f \<Pi> (\<not>g1)) \<and> \<not>(f \<Pi>(\<not>g2)))"
   by fastforce
 have 7: "\<turnstile>\<not>(f \<Pi> (\<not>g1)) = f \<Pi>\<^sup>u g1"
   by (simp add: NotPiEqvNotUPi)  
 have 8: "\<turnstile> \<not>(f \<Pi> (\<not>g2)) = f \<Pi>\<^sup>u g2 "
   by (simp add: NotPiEqvNotUPi)  
 have 9: "\<turnstile> (\<not>(f \<Pi> (\<not>g1)) \<and> \<not>(f \<Pi>(\<not>g2))) = (f \<Pi>\<^sup>u g1 \<and> f \<Pi>\<^sup>u g2)"
   using 6 7 8 by fastforce
 from 2 3 5 6 9 show ?thesis by fastforce
qed 


lemma UpiAndImp:
 "\<turnstile> f \<Pi>\<^sup>u (g1 \<longrightarrow> g2) \<and> f \<Pi> g1 \<longrightarrow> f \<Pi> g2"
proof -
 have 2: "\<turnstile> f \<Pi>\<^sup>u ((\<not>g2) \<longrightarrow> (\<not>g1)) \<longrightarrow> ( f \<Pi>\<^sup>u (\<not> g2) \<longrightarrow> f \<Pi>\<^sup>u (\<not> g1) )"
   using PiK by blast
 have 3: "\<turnstile> (f \<Pi>\<^sup>u (\<not>g2) \<longrightarrow> f \<Pi>\<^sup>u (\<not>g1)) = ((\<not> (f \<Pi>\<^sup>u (\<not>g1))) \<longrightarrow> (\<not>(f \<Pi>\<^sup>u (\<not>g2))))"
   by auto
 have 4: "\<turnstile> (\<not> (f \<Pi>\<^sup>u (\<not>g2))) = f \<Pi> g2" 
    by (simp add: upi_d_def)
 have 5: "\<turnstile> (\<not> (f \<Pi>\<^sup>u (\<not>g1))) = f \<Pi> g1" 
    by (simp add: upi_d_def)
 have 6: "\<turnstile> f \<Pi>\<^sup>u ((\<not>g2) \<longrightarrow> (\<not>g1)) = f \<Pi>\<^sup>u (g1 \<longrightarrow> g2)"
    by simp
 have 7: "\<turnstile> (f \<Pi>\<^sup>u (g1 \<longrightarrow> g2) \<and> f \<Pi> g1 \<longrightarrow> f \<Pi> g2) =
            (f \<Pi>\<^sup>u (g1 \<longrightarrow> g2)  \<longrightarrow> (f \<Pi> g1 \<longrightarrow> f \<Pi> g2))  "
   by auto
 show ?thesis 
 using "2" "4" "5" by fastforce
qed

lemma BoxImpUPiBox:
 "\<turnstile> \<box> (init w) \<longrightarrow> f \<Pi>\<^sup>u (\<box> (init w))"
proof -
 have 1: "\<turnstile> f \<Pi> (\<diamond> (init (\<not>w))) \<longrightarrow> \<diamond> (init (\<not>w))"
     by (simp add: PiDiamondImp) 
 have 2: "\<turnstile> \<not>  \<diamond> (init (\<not>w)) \<longrightarrow> \<not> (f \<Pi> (\<diamond> (init (\<not>w))))" 
   using "1" by auto
 have 3: "\<turnstile> (\<not>  \<diamond> (init (\<not>w))) = \<box> (init w)" 
   by (metis "2" Initprop(2) Prop10 always_d_def inteq_reflection)
 have 4: "\<turnstile> (\<not> (f \<Pi> (\<diamond> (init (\<not>w))))) = f \<Pi>\<^sup>u  (\<box> (init w))" 
   by (simp add: upi_d_def)
      (metis "3" int_simps(4) inteq_reflection) 
 show ?thesis 
 using "2" "3" "4" by fastforce
qed

lemma WPrevPi:
 "\<turnstile> (init w) \<Pi> f = (wprev (\<box> (init (\<not> w)))); ((init w) \<and> (init w) \<Pi> f)"
   using StatePiUntil StatePiUntil1 by fastforce


lemma PiChopstarhelp2a:
 "\<turnstile>  (w \<and> empty);(power ( (f;(w \<and> empty)) \<and> more) k) =
            (power (((w \<and> empty);f) \<and> more) k ) ;(w \<and> empty)"
proof (induction k)
case 0
then show ?case 
by (metis ChopEmpty EmptyChop inteq_reflection pow_0)
next
case (Suc k)
then show ?case 
  proof -
   have 1: "\<turnstile> (w \<and> empty);power (f;(w \<and> empty) \<and> more) (Suc k)  =
               (w \<and> empty);((f;(w \<and> empty) \<and> more); power (f;(w \<and> empty) \<and> more) k) "
     by simp
   have 11: "\<turnstile> (f \<and> more);(w \<and> empty) = (fin w \<and> (f \<and> more)) "
      by (metis FinEqvTrueChopAndEmpty TrueChopAndEmptyEqvChopAndEmpty inteq_reflection)
   have 12: "\<turnstile> (f;(w \<and> empty)) = (fin w \<and> f)" 
          by (meson AndFinEqvChopAndEmpty Prop04 lift_and_com)
   have 13: "\<turnstile> (f;(w \<and> empty) \<and> more) = ((fin w \<and> f) \<and> more)" 
       using "12" by auto
   have 14: "\<turnstile> ((fin w \<and> f) \<and> more) = (fin w \<and> (f \<and> more))"
      by fastforce
   have 2: "\<turnstile> (f;(w \<and> empty) \<and> more) = (f \<and> more);(w \<and> empty)"
         using "11" "13" by fastforce
   have 20: "\<turnstile>((f;(w \<and> empty) \<and> more); power (f;(w \<and> empty) \<and> more) k) =
              ((f \<and> more);(w \<and> empty)); power (f;(w \<and> empty) \<and> more) k"
        by (simp add: "2" LeftChopEqvChop)      
   have 21: "\<turnstile> ((f \<and> more);(w \<and> empty)); power (f;(w \<and> empty) \<and> more) k =
               (f \<and> more);((w \<and> empty); power (f;(w \<and> empty) \<and> more) k)"
       using ChopAssocB by blast
   have 22: "\<turnstile> (f \<and> more);((w \<and> empty); power (f;(w \<and> empty) \<and> more) k) =
              ((f \<and> more); (power ((w \<and> empty);f \<and> more) k;(w \<and> empty))) "
         by (simp add: RightChopEqvChop Suc.IH)
   have 3: "\<turnstile> (w \<and> empty);((f;(w \<and> empty) \<and> more); power (f;(w \<and> empty) \<and> more) k) =
              (w \<and> empty);((f \<and> more); (power ((w \<and> empty);f \<and> more) k;(w \<and> empty)))  " 
     by (metis "1" "2" "21" "22" inteq_reflection)
   have 4: "\<turnstile> (w \<and> empty);((f \<and> more); (power ((w \<and> empty);f \<and> more) k;(w \<and> empty))) =
              ((w \<and> empty);(f \<and> more)); ((power ((w \<and> empty);f \<and> more) k;(w \<and> empty)))"       
       using ChopAssoc by blast
   have 5: "\<turnstile> (w \<and> empty);(f \<and> more) = ((w \<and> empty);f \<and> more) "
       by (auto simp add: Valid_def chop_defs empty_defs more_defs)
   have 6: "\<turnstile> ((w \<and> empty);(f \<and> more)); ((power ((w \<and> empty);f \<and> more) k;(w \<and> empty))) =
             ((w \<and> empty);f \<and> more); ((power ((w \<and> empty);f \<and> more) k;(w \<and> empty))) "
       using "5" LeftChopEqvChop by blast
   have 7: "\<turnstile> ((w \<and> empty);f \<and> more); ((power ((w \<and> empty);f \<and> more) k;(w \<and> empty))) =
              ((power ((w \<and> empty);f \<and> more) (Suc k);(w \<and> empty)))" 
     by (simp add: ChopAssoc)
   show ?thesis
   by (metis "1" "3" "4" "5" "7" int_eq)
   qed
qed

lemma PiChopstarhelp2:
 "\<turnstile> (w \<and> empty);(f;(w \<and> empty))\<^sup>\<star> = ((w \<and> empty);f)\<^sup>\<star>;(w \<and> empty)"
proof -
 have 1: "\<turnstile> (w \<and> empty);(f;(w \<and> empty))\<^sup>\<star> = (w \<and> empty);(\<exists>k. power ( (f;(w \<and> empty)) \<and> more) k)  " 
   by (simp add: chopstar_d_def powerstar_d_def)
 have 2: "\<turnstile> (w \<and> empty);(\<exists>k. power ( (f;(w \<and> empty)) \<and> more) k) =
             (\<exists>k. (w \<and> empty);(power ( (f;(w \<and> empty)) \<and> more) k))     "
    using ChopExist by fastforce 
 have 3: "\<turnstile> (\<exists>k. (w \<and> empty);(power ( (f;(w \<and> empty)) \<and> more) k)) =
            (\<exists>k. (power ((w \<and> empty);f \<and> more) k);(w \<and> empty)) 
            " 
    by (simp add: ExEqvRule PiChopstarhelp2a)
 have 4: "\<turnstile> (\<exists>k. (power ((w \<and> empty);f \<and> more) k);(w \<and> empty)) =
            (\<exists>k. (power ((w \<and> empty);f \<and> more) k));(w \<and> empty)"
    using ExistChop by fastforce
 have 5: "\<turnstile> (\<exists>k. (power ((w \<and> empty);f \<and> more) k));(w \<and> empty) =
            ((w \<and> empty);f)\<^sup>\<star>;(w \<and> empty)" 
   by (simp add: chopstar_d_def powerstar_d_def)
 show ?thesis 
 using "1" "2" "3" "4" "5" by fastforce
qed  

lemma PiPowerSuca:
 "\<turnstile> (init w) \<Pi> (power f (Suc k)) = ((init w) \<Pi> f);((init w) \<and> (init w) \<Pi> (power f k))"
by (simp add: PiChopDist)

lemma PiPowerSucb:
 "\<turnstile> (init w) \<Pi> (power f (Suc k)) = ((init w) \<Pi> f \<and> fin w);((init w) \<and>  (init w) \<Pi> (power f k))"
by (metis (no_types, lifting) AndFinChopEqvStateAndChop ChopAssoc ChopImpDiamond FinChopEqvDiamond
    FinEqvTrueChopAndEmpty InitAndEmptyEqvAndEmpty PiPowerSuca Prop10 StateAndEmptyChop 
   inteq_reflection)


lemma PiPowerSucc:
  "\<turnstile> (init w) \<Pi> (power f (Suc k)) = 
     (power ((init w) \<Pi> f \<and> fin w) (Suc k));((init w) \<and> (init w) \<Pi> empty)"
proof (induction k)
case 0
then show ?case 
  by (metis ChopEmpty PiPowerSucb inteq_reflection pow_0 pow_Suc)
next
case (Suc k)
then show ?case
  by (metis (no_types, opaque_lifting) AndFinEqvChopAndEmpty ChopAssocB InitAndEmptyEqvAndEmpty 
      PiChopDist  StateAndEmptyChop TrueW  int_simps(1) inteq_reflection lift_and_com pow_Suc)
qed 

lemma PiPowerSucd:
  "\<turnstile> (init w) \<Pi> (power f (Suc k)) = (power ((init w) \<Pi> f \<and> fin w) (Suc k));((init w) \<Pi> empty)"
proof (induction k)
case 0
then show ?case 
by (metis (no_types, lifting) AndFinChopEqvStateAndChop AndFinEqvChopAndEmpty 
    InitAndEmptyEqvAndEmpty PiPowerSuca int_eq pow_0 pow_Suc)
next
case (Suc k)
then show ?case 
by (metis (no_types, lifting) ChopAssoc PiPowerSucc inteq_reflection pow_Suc)
qed

lemma PiChopstar:
 "\<turnstile> (init w) \<Pi> (f\<^sup>\<star>) = (init (\<not> w)) \<U> ((init w) \<and> (((init w) \<Pi> f) \<and> fin w)\<^sup>\<star>;wnext(\<box> (init (\<not> w))))"
proof -
 have 1: "\<turnstile> (init w) \<Pi> (f\<^sup>\<star>) = (init w) \<Pi> (\<exists>k. power f k)"
   by (metis ChopstarEqvPowerstar PiEqvRule powerstar_d_def) 
 have 2: "\<turnstile> (init w) \<Pi> (\<exists>k. power f k) = (\<exists>k. (init w) \<Pi> (power f k))"
   by (simp add: Valid_def pi_d_def) 
 have 3: "\<turnstile> (\<exists>k. (init w) \<Pi> (power f k)) = 
            ( (init w) \<Pi> empty \<or> (\<exists>k. (init w) \<Pi> (power f (Suc k))))"
  using PiPowerExpand by auto
 have 4: "\<turnstile> (\<exists>k. (init w) \<Pi> (power f (Suc k))) =
            (\<exists>k. (power ((init w) \<Pi> f \<and> fin w) (Suc k));((init w) \<Pi> empty))" 
   by (meson ExEqvRule PiPowerSucd)
 have 5: "\<turnstile> (init w) \<Pi> empty = empty ; ((init w) \<Pi> empty)"
   by (simp add: EmptyChop int_iffD1 int_iffD2 int_iffI)
 have 6: "\<turnstile> (\<exists>k. (power ((init w) \<Pi> f \<and> fin w) (Suc k));((init w) \<Pi> empty)) =
            (\<exists>k. (power ((init w) \<Pi> f \<and> fin w) (Suc k)));((init w) \<Pi> empty)"
    by (simp add: Semantics.ExistChop) 
 have 7: "\<turnstile> ( (init w) \<Pi> empty \<or> (\<exists>k. (power ((init w) \<Pi> f \<and> fin w) (Suc k));((init w) \<Pi> empty)))=
            ( empty; ((init w) \<Pi> empty) \<or> 
              (\<exists>k. (power ((init w) \<Pi> f \<and> fin w) (Suc k)));((init w) \<Pi> empty))"
         using "5" "6" by fastforce
 have 8: "\<turnstile> ( empty; ((init w) \<Pi> empty) \<or> 
              (\<exists>k. (power ((init w) \<Pi> f \<and> fin w) (Suc k)));((init w) \<Pi> empty)) =
            ( empty \<or> (\<exists>k. (power ((init w) \<Pi> f \<and> fin w) (Suc k))));((init w) \<Pi> empty)"
     by (meson OrChopEqv Prop11)
 have 9: "\<turnstile> empty = (power ((init w) \<Pi> f \<and> fin w) 0)"
       by simp 
 have 10: "\<turnstile> ( empty \<or> (\<exists>k. (power ((init w) \<Pi> f \<and> fin w) (Suc k)))) = 
             ( (power ((init w) \<Pi> f \<and> fin w) 0) \<or> (\<exists>k. (power ((init w) \<Pi> f \<and> fin w) (Suc k))))"
   by simp
 have 11: "\<turnstile> ( (power ((init w) \<Pi> f \<and> fin w) 0) \<or> (\<exists>k. (power ((init w) \<Pi> f \<and> fin w) (Suc k)))) =
             (\<exists>k. (power ((init w) \<Pi> f \<and> fin w) k))"
     using exists_expand[of "w" "f"] by fastforce   
 have 12: "\<turnstile> ( (init w) \<Pi> empty \<or> 
               (\<exists>k. (power ((init w) \<Pi> f \<and> fin w) (Suc k));((init w) \<Pi> empty))) =
            (\<exists>k. (power ((init w) \<Pi> f \<and> fin w) ( k))); ((init w) \<Pi> empty)"
    by (metis "11" "7" "8" "9" inteq_reflection)
 have 13: "\<turnstile> (\<exists>k. (power ((init w) \<Pi> f \<and> fin w) ( k))); ((init w) \<Pi> empty) =
             ((init w) \<Pi> f \<and> fin w)\<^sup>\<star> ;  ((init w) \<Pi> empty)"
   by (metis ChopstarEqvPowerstar LeftChopEqvChop inteq_reflection powerstar_d_def)
 have 14: "\<turnstile> ((init w) \<Pi> f \<and> fin w)\<^sup>\<star> ;  ((init w) \<Pi> empty) =
             ( ((init w) \<Pi> empty) \<or> 
               (((init w) \<Pi> f \<and> fin w);((init w) \<Pi> f \<and> fin w)\<^sup>\<star>);((init w) \<Pi> empty))"
    using CSChopEqvOrChopPlusChop by blast
 have 15: "\<turnstile> ((init w) \<Pi> empty) =  (init (\<not> w)) \<U> ( (init w) \<and> wnext (\<box> (init (\<not> w))))"
   by (simp add: PiEmpty) 
 have 16:"\<turnstile> (((init w) \<Pi> f \<and> fin w);((init w) \<Pi> f \<and> fin w)\<^sup>\<star>);((init w) \<Pi> empty) =
            (((init w) \<Pi> f \<and> fin w);((init w) \<Pi> f \<and> fin w)\<^sup>\<star>); wnext (\<box> (init (\<not> w)))"
   by (metis (no_types, lifting) AndFinEqvChopAndEmpty ChopAssoc InitAndEmptyEqvAndEmpty 
         PiChopstarhelp2 StateAndEmptyChop StateAndPiEmpty inteq_reflection)
 have 17: "\<turnstile> ((init w) \<Pi> f \<and> fin w) = ((init w) \<Pi> f);((init w) \<and> empty)"
   by (metis AndFinEqvChopAndEmpty InitAndEmptyEqvAndEmpty inteq_reflection)
 have 18: "\<turnstile> ((init w) \<Pi> f) = wprev(\<box> (init (\<not> w)));((init w) \<and> (init w) \<Pi> f)"
   by (simp add: WPrevPi)
 have 19: "\<turnstile> wprev(\<box> (init (\<not> w)));((init w) \<and> (init w) \<Pi> f) = 
             wprev(\<box> (init (\<not> w)));(((init w) \<and> empty) ; ((init w) \<Pi> f))"
   by (meson RightChopImpChop StateAndEmptyChop int_iffD1 int_iffD2 int_iffI)
 have 20: "\<turnstile> wprev(\<box> (init (\<not> w)));(((init w) \<and> empty) ; ((init w) \<Pi> f)) =
             (init (\<not> w)) \<U> ( (init w) \<and> ((init w) \<Pi> f)) "
   using "18" "19" StatePiUntil by fastforce   
 have 21: "\<turnstile> (((init w) \<Pi> f \<and> fin w);((init w) \<Pi> f \<and> fin w)\<^sup>\<star>); wnext (\<box> (init (\<not> w))) =
             (init (\<not> w)) \<U> (
                       (init w) \<and> 
                       ((((init w) \<Pi> f) \<and> fin w);((init w) \<Pi> f \<and> fin w)\<^sup>\<star>);wnext (\<box> (init (\<not> w)))
                             )"
    by (metis (no_types, lifting) "17" "18" AndFinChopEqvStateAndChop ChopAssocB 
        StateUntilEqvWPrevChop int_eq)
 have 22: "\<turnstile> ((init (\<not> w)) \<U> ( (init w) \<and> wnext (\<box> (init (\<not> w)))) \<or>
             (init (\<not> w)) \<U> ( 
                    (init w) \<and> 
                    ((((init w) \<Pi> f) \<and> fin w);((init w) \<Pi> f \<and> fin w)\<^sup>\<star>);wnext (\<box> (init (\<not> w)))
                            ) ) =
              (init (\<not> w)) \<U>  ( 
         ((init w) \<and> wnext (\<box> (init (\<not> w)))) \<or>
         ( (init w) \<and> ((((init w) \<Pi> f) \<and> fin w);((init w) \<Pi> f \<and> fin w)\<^sup>\<star>);wnext (\<box> (init (\<not> w))))
                               ) "
       using UntilOrDist by fastforce 
 have 23: "\<turnstile> ( 
        ((init w) \<and> wnext (\<box> (init (\<not> w)))) \<or>
        ( (init w) \<and> ((((init w) \<Pi> f) \<and> fin w);((init w) \<Pi> f \<and> fin w)\<^sup>\<star>);wnext (\<box> (init (\<not> w))))
             ) =
             ( (init w) \<and> (((init w) \<Pi> f) \<and> fin w)\<^sup>\<star>;wnext(\<box> (init (\<not> w))))"
   by (metis (no_types, lifting) CSChopEqvOrChopPlusChop ChopOrEqv StateAndEmptyChop 
       inteq_reflection) 
 have 24: "\<turnstile> (init w) \<Pi> (f\<^sup>\<star>) = ( (init w) \<Pi> empty \<or> (\<exists>k. (init w) \<Pi> (power f (Suc k))))" 
   using "1" "2" "3" by fastforce
 have 25: "\<turnstile> ( (init w) \<Pi> empty \<or> (\<exists>k. (init w) \<Pi> (power f (Suc k)))) =
             ( (init w) \<Pi> empty \<or> (\<exists>k. (power ((init w) \<Pi> f \<and> fin w) (Suc k));((init w) \<Pi> empty)))" 
   using "4" by fastforce
 have 26: "\<turnstile> ( (init w) \<Pi> empty \<or> 
               (\<exists>k. (power ((init w) \<Pi> f \<and> fin w) (Suc k));((init w) \<Pi> empty))) =
             ((init w) \<Pi> f \<and> fin w)\<^sup>\<star> ;  ((init w) \<Pi> empty)"
    using "12" "13" by fastforce
 have 27:"\<turnstile> ((init w) \<Pi> f \<and> fin w)\<^sup>\<star> ;  ((init w) \<Pi> empty) =
            (init (\<not> w)) \<U> ((init w) \<and> (((init w) \<Pi> f) \<and> fin w)\<^sup>\<star>;wnext(\<box> (init (\<not> w))))"
    by (metis "14" "15" "16" "21" "22" "23" inteq_reflection)
 from 24 25 26 27 show ?thesis by fastforce
qed 








end