(*  Title     : A Shallow Encoding of ITL in Isabelle/HOL
    Authors   : Antonio Cau     <cau.researcher at gmail.com>
                Ben Moszkowski
                David Smallwood <drs at dmu.ac.uk>
    Maintainer: Antonio Cau     <cau.researcher at gmail.com>        
    License   : BSD
*)

section \<open>Finite ITL Semantics\<close>   

theory Semantics
imports Interval "HOL-TLA.Intensional" "HOL-TLA.Stfun"
begin
(*
sledgehammer_params [minimize=true,preplay_timeout=10,timeout=60,verbose=true,
                    provers="
  cvc4 z3 e spass vampire
"]
*)

text \<open> 
  This theory mechanises a \emph{shallow} embedding of finite ITL using the
  \<open>Interval\<close> and \<open>Intensional\<close> theories. A shallow embedding
  represents ITL using Isabelle/HOL predicates, while a \emph{deep}
  embedding \cite{ITLHOLv1} would represent ITL formulas  as mutually
  inductive datatypes. See, e.g., \cite{Wildmoser04} for a discussion
  about deep vs. shallow embeddings in Isabelle/HOL.  
  The choice of a shallow over a deep embedding is motivated \cite{Merz98,GrovMerz}
   by the following 
  factors: a shallow embedding is usually less involved, and existing Isabelle
  theories and tools can be applied more directly to enhance automation; due to
  the lifting in the \<open>Intensional\<close> theory, a shallow embedding can reuse
  standard logical operators, whilst a deep embedding requires a different
  set of operators for formulas. Finally, since our 
  target is system verification rather than proving meta-properties of the logic,
  which requires a deep embedding, a shallow embedding is more fit for purpose.
\<close>

subsection "Types of Formulas"

text \<open>
  To mechanise the ITL semantics, the following type abbreviations are used:
\<close>

type_synonym ('a,'b) formfun = "'a interval \<Rightarrow> 'b"
type_synonym 'a formula      = "('a,bool) formfun"
type_synonym ('a,'b) stfun   = "'a \<Rightarrow> 'b"
type_synonym 'a stpred       = "('a,bool) stfun"

instance 
 "fun" :: (type,type) world ..

instance
 "prod" :: (type,type) world ..

instance
 "interval" :: (type) world ..

text \<open>
  Pair, function, and interval are instantiated to be of type class world.
  This allows use of the lifted Intensional logic for formulas, and 
  standard logical connectives can therefore be used.
\<close>

subsection \<open>Semantics of ITL\<close>

text \<open> The semantics of ITL is defined. Note chopstar is a derived operator, i.e., it is 
defined recursively in terms of chop.\<close>

definition skip_d :: "('a ::world) formula "
where "skip_d \<equiv> \<lambda>s. ilen s =1"

definition chop_d :: " ('a ::world) formula \<Rightarrow> ('a ::world) formula \<Rightarrow> ('a ::world) formula"
where "chop_d F1 F2 \<equiv> \<lambda> s. \<exists>n. 0\<le> n \<and> n\<le>ilen s \<and> ((prefix n s) \<Turnstile> F1) \<and> ((suffix n s) \<Turnstile> F2 )"  

definition current_val_d :: "('a::world,'b) stfun \<Rightarrow> ('a,'b) formfun"
where "current_val_d f \<equiv> \<lambda> s. (inth s 0) \<Turnstile> f"

definition next_val_d :: "('a::world,'b) stfun \<Rightarrow> ('a,'b) formfun"
where "next_val_d f \<equiv> \<lambda> s. if ilen s >0 then ((inth s 1) \<Turnstile> f) else (\<some> (x::'b ). x=x) " 

definition fin_val_d :: "('a::world,'b) stfun \<Rightarrow> ('a,'b) formfun"
where "fin_val_d f \<equiv> \<lambda> s. (inth s (ilen s)) \<Turnstile> f"

definition penult_val_d :: "('a::world,'b) stfun \<Rightarrow> ('a,'b) formfun"
where "penult_val_d f \<equiv> \<lambda> s. if ilen s >0 then (inth s ((ilen s)-1) \<Turnstile> f) else (\<some> (x::'b ). x=x)" 


text \<open> This is the concrete syntax for the (abstract) operators above. \<close>

syntax
 "_skip_d"         :: "lift"                ("(skip)")
 "_chop_d"         :: "[lift,lift] \<Rightarrow> lift" ("(_;_)" [84,84] 83) 
 "_current_val_d"  :: "lift \<Rightarrow> lift"        ("($_)" [100] 99)
 "_next_val_d"     :: "lift \<Rightarrow> lift"        ("(_$)" [100] 99)
 "_fin_val_d"      :: "lift \<Rightarrow> lift"        ("(!_)" [100] 99)
 "_penult_val_d"   :: "lift \<Rightarrow> lift"        ("(_!)" [100] 99)
 "TEMP"            :: "lift \<Rightarrow> 'b"          ("(TEMP _)")

syntax (ASCII)
 "_skip_d"         :: "lift"                ("(skip)")
 "_chop_d"         :: "[lift,lift] \<Rightarrow> lift" ("(_;_)" [84,84] 83)
 "_current_val_d"  :: "lift \<Rightarrow> lift"        ("($_)" [100] 99)
 "_next_val_d"     :: "lift \<Rightarrow> lift"        ("(_$)" [100] 99)
 "_fin_val_d"      :: "lift \<Rightarrow> lift"        ("(!_)" [100] 99)
 "_penult_val_d"   :: "lift \<Rightarrow> lift"        ("(_!)" [100] 99)
translations
 "_skip_d"        \<rightleftharpoons> "CONST skip_d"
 "_chop_d"        \<rightleftharpoons> "CONST chop_d"
 "_current_val_d" \<rightleftharpoons> "CONST current_val_d"
 "_next_val_d"    \<rightleftharpoons> "CONST next_val_d"
 "_fin_val_d"     \<rightleftharpoons> "CONST fin_val_d"
 "_penult_val_d"  \<rightleftharpoons> "CONST penult_val_d" 
 "TEMP F"         \<rightharpoonup> "(F:: (_ interval) \<Rightarrow> _)"


subsection "Abbreviations"

text \<open> Some standard temporal abbreviations, with their concrete syntax. \<close>

definition sometimes_d :: "('a::world) formula \<Rightarrow> 'a formula"
where "sometimes_d F \<equiv> LIFT(#True;F)"

definition di_d :: "('a::world) formula \<Rightarrow> 'a formula"
where "di_d F \<equiv> LIFT(F;#True)" 

definition da_d :: "('a::world) formula \<Rightarrow> 'a formula"
where "da_d F \<equiv> LIFT(#True;(F;#True))" 

definition next_d :: "('a::world) formula \<Rightarrow> 'a formula"
where "next_d F \<equiv> LIFT(skip;F)" 

definition prev_d :: "('a::world) formula \<Rightarrow> 'a formula"
where "prev_d F \<equiv> LIFT(F;skip)"

syntax
  "_sometimes_d" :: "lift \<Rightarrow> lift" ("(\<diamond>_)" [88] 87)
  "_di_d"        :: "lift \<Rightarrow> lift" ("(di _)" [88] 87)
  "_da_d"        :: "lift \<Rightarrow> lift" ("(da _)" [88] 87)
  "_next_d"      :: "lift \<Rightarrow> lift" ("(\<circle> _)" [88] 87)
  "_prev_d"      :: "lift \<Rightarrow> lift" ("(prev _)" [88] 87)


syntax (ASCII)
  "_sometimes_d" :: "lift \<Rightarrow> lift" ("(<>_)" [88] 87)
  "_di_d"        :: "lift \<Rightarrow> lift" ("(di _)" [88] 87) 
  "_da_d"        :: "lift \<Rightarrow> lift" ("(da _)" [88] 87)
  "_next_d"      :: "lift \<Rightarrow> lift" ("(next _)" [88] 87)
  "_prev_d"      :: "lift \<Rightarrow> lift" ("(prev _)" [88] 87)


translations 
  "_sometimes_d" \<rightleftharpoons> "CONST sometimes_d"
  "_di_d"        \<rightleftharpoons> "CONST di_d" 
  "_da_d"        \<rightleftharpoons> "CONST da_d"
  "_next_d"      \<rightleftharpoons> "CONST next_d" 
  "_prev_d"      \<rightleftharpoons> "CONST prev_d"


definition always_d :: "('a::world) formula \<Rightarrow> 'a formula "
where "always_d F \<equiv> LIFT(\<not>(\<diamond>(\<not>F)))"

definition bi_d :: "('a::world) formula \<Rightarrow> 'a formula "
where "bi_d F \<equiv> LIFT(\<not>(di(\<not>F)))"

definition ba_d :: "('a::world) formula \<Rightarrow> 'a formula "
where "ba_d F \<equiv> LIFT(\<not>(da(\<not>F)))"

definition wnext_d :: "('a::world) formula \<Rightarrow> 'a formula "
where "wnext_d F \<equiv> LIFT(\<not>(\<circle>(\<not>F)))"

definition wprev_d :: "('a::world) formula \<Rightarrow> 'a formula "
where "wprev_d F \<equiv> LIFT(\<not>(prev(\<not>F)))"

definition more_d :: "('a::world) formula "
where "more_d \<equiv> LIFT(\<circle>(#True))" 


syntax
 "_always_d"   :: "lift \<Rightarrow> lift" ("(\<box>_)" [88] 87)
 "_bi_d"       :: "lift \<Rightarrow> lift" ("(bi _)" [88] 87)
 "_ba_d"       :: "lift \<Rightarrow> lift" ("(ba _)" [88] 87)
 "_wnext_d"    :: "lift \<Rightarrow> lift" ("(wnext _)" [88] 87)
 "_wprev_d"    :: "lift \<Rightarrow> lift" ("(wprev _)" [88] 87)
 "_more_d"     :: "lift"         ("(more)")
 

syntax (ASCII)
 "_always_d"   :: "lift \<Rightarrow> lift" ("([]_)" [88] 87)
 "_bi_d"       :: "lift \<Rightarrow> lift" ("(bi _)" [88] 87)
 "_ba_d"       :: "lift \<Rightarrow> lift" ("(ba _)" [88] 87)
 "_wnext_d"    :: "lift \<Rightarrow> lift" ("(wnext _)" [88] 87)
 "_wprev_d"    :: "lift \<Rightarrow> lift" ("(wprev _)" [88] 87)
 "_more_d"     :: "lift"         ("(more)")
 

translations
 "_always_d" \<rightleftharpoons> "CONST always_d"
 "_bi_d"     \<rightleftharpoons> "CONST bi_d" 
 "_ba_d"     \<rightleftharpoons> "CONST ba_d"
 "_wnext_d"  \<rightleftharpoons> "CONST wnext_d"
 "_wprev_d"  \<rightleftharpoons> "CONST wprev_d"
 "_more_d"   \<rightleftharpoons> "CONST more_d" 

definition empty_d :: "('a::world) formula "
where "empty_d \<equiv> LIFT(\<not>(more))" 

definition dm_d :: "('a::world) formula \<Rightarrow> 'a formula"
where "dm_d F \<equiv> LIFT(#True;(more \<and> F))" 

syntax
 "_empty_d"    :: "lift"         ("(empty)")
 "_dm_d"       :: "lift \<Rightarrow> lift" ("(dm _)" [88] 87)

syntax (ASCII)
 "_empty_d"    :: "lift"         ("(empty)")
 "_dm_d"       :: "lift \<Rightarrow> lift" ("(dm _)" [88] 87)

translations
 "_empty_d" \<rightleftharpoons> "CONST empty_d"
 "_dm_d"    \<rightleftharpoons> "CONST dm_d"


definition bm_d :: "('a::world) formula \<Rightarrow> 'a formula "
where "bm_d F \<equiv> LIFT(\<not>(dm(\<not>F)))" 

definition init_d :: "('a::world) formula \<Rightarrow> 'a formula "
where "init_d F \<equiv> LIFT((empty \<and> F);#True)" 

definition fin_d :: "('a::world) formula \<Rightarrow> 'a formula "
where "fin_d F \<equiv> LIFT(\<box>(empty \<longrightarrow> F))" 

definition halt_d :: "('a::world) formula \<Rightarrow> 'a formula "
where "halt_d F \<equiv> LIFT(\<box>(empty = F))" 

definition initonly_d :: "('a::world) formula \<Rightarrow> 'a formula "
where "initonly_d F \<equiv> LIFT(bi(empty = F))" 

definition keep_d :: "('a::world) formula \<Rightarrow> 'a formula "
where "keep_d F \<equiv> LIFT(ba(skip \<longrightarrow> F))" 

definition yields_d :: "('a::world) formula \<Rightarrow> 'a formula \<Rightarrow> 'a formula "
where "yields_d F1 F2 \<equiv> LIFT(\<not>(F1;(\<not>F2)))" 

definition ifthenelse_d :: "('a::world) formula \<Rightarrow> 'a formula \<Rightarrow> 'a formula \<Rightarrow> 'a formula "
where "ifthenelse_d F G H \<equiv> LIFT((F \<and> G) \<or> (\<not>F \<and> H) ) "

primrec power_d :: "('a::world) formula \<Rightarrow> nat \<Rightarrow> 'a formula "
where  pow_0  : "(power_d F 0) = LIFT(empty)"
     | pow_Suc: "(power_d F (Suc n)) = LIFT((F);(power_d F n)) "


syntax
 "_bm_d"         :: "lift \<Rightarrow> lift"             ("(bm _)" [88] 87)
 "_init_d"       :: "lift \<Rightarrow> lift"             ("(init _)" [88] 87)
 "_fin_d"        :: "lift \<Rightarrow> lift"             ("(fin _)" [88] 87)
 "_halt_d"       :: "lift \<Rightarrow> lift"             ("(halt _)" [88] 87)
 "_initonly_d"   :: "lift \<Rightarrow> lift"             ("(initonly _)" [88] 87)
 "_keep_d"       :: "lift \<Rightarrow> lift"             ("(keep _)" [88] 87)
 "_yields_d"     :: "[lift,lift] \<Rightarrow> lift"      ("(_ yields _)" [88,88] 87)
 "_ifthenelse_d" :: "[lift,lift,lift] \<Rightarrow> lift" ("(if\<^sub>i _ then _ else _ )"  [88,88,88] 87)
 "_power_d"      :: "[lift,nat] \<Rightarrow> lift"       ("(power _ _)" [88,88] 87) 
 
syntax (ASCII)
 "_bm_d"         :: "lift \<Rightarrow> lift"             ("(bm _)" [88] 87)
 "_init_d"       :: "lift \<Rightarrow> lift"             ("(init _)" [88] 87)
 "_fin_d"        :: "lift \<Rightarrow> lift"             ("(fin _)" [88] 87)
 "_halt_d"       :: "lift \<Rightarrow> lift"             ("(halt _)" [88] 87)
 "_initonly_d"   :: "lift \<Rightarrow> lift"             ("(initonly _)" [88] 87)
 "_keep_d"       :: "lift \<Rightarrow> lift"             ("(keep _)" [88] 87)
 "_yields_d"     :: "[lift,lift] \<Rightarrow> lift"      ("(_ yields _)" [88,88] 87)
 "_ifthenelse_d" :: "[lift,lift,lift] \<Rightarrow> lift" ("(if\<^sub>i _ then _ else _ )"  [88,88,88] 87)
 "_power_d"      :: "[lift,nat] \<Rightarrow> lift"       ("(power _ _)" [88,88] 87)

translations
 "_bm_d"         \<rightleftharpoons> "CONST bm_d"
 "_init_d"       \<rightleftharpoons> "CONST init_d"
 "_fin_d"        \<rightleftharpoons> "CONST fin_d"
 "_halt_d"       \<rightleftharpoons> "CONST halt_d"
 "_initonly_d"   \<rightleftharpoons> "CONST initonly_d" 
 "_keep_d"       \<rightleftharpoons> "CONST keep_d"
 "_yields_d"     \<rightleftharpoons> "CONST yields_d"
 "_ifthenelse_d" \<rightleftharpoons> "CONST ifthenelse_d" 
 "_power_d"      \<rightleftharpoons> "CONST power_d"

definition len_d :: "nat \<Rightarrow> ('a::world) formula "
where "len_d n \<equiv> LIFT(power skip n)"

definition powerstar_d :: "('a::world) formula \<Rightarrow> 'a formula"
where "powerstar_d F \<equiv> LIFT(\<exists> k. power F k)"

syntax
 "_len_d"          :: "nat \<Rightarrow> lift"         ("(len _)" [88] 87) 
 "_powerstar_d"    :: "lift \<Rightarrow> lift"        ("(powerstar _)" [85] 85)

syntax (ASCII)
 "_len_d"          :: "nat \<Rightarrow> lift"         ("(len _)" [88] 87) 
 "_powerstar_d"    :: "lift \<Rightarrow> lift"        ("(powerstar _)" [85] 85)

translations
 "_len_d"          \<rightleftharpoons> "CONST len_d" 
 "_powerstar_d"    \<rightleftharpoons> "CONST powerstar_d"

definition chopstar_d :: "('a::world) formula \<Rightarrow> 'a formula"
where "chopstar_d F \<equiv> LIFT(powerstar (F \<and> more))"

syntax
 "_chopstar_d"     :: "lift \<Rightarrow> lift"        ("(_\<^sup>\<star>)" [85] 85)

syntax (ASCII)
 "_chopstar_d"     :: "lift \<Rightarrow> lift"        ("(chopstar _)" [85] 85)

translations
 "_chopstar_d"     \<rightleftharpoons> "CONST chopstar_d"


definition ifthen_d :: "('a::world) formula  \<Rightarrow> 'a formula \<Rightarrow> 'a formula "
where "ifthen_d F G \<equiv> LIFT(if\<^sub>i F then G else #True ) "

definition while_d :: "('a::world) formula  \<Rightarrow> 'a formula \<Rightarrow> 'a formula "
where "while_d F G \<equiv> LIFT( ( F \<and> G)\<^sup>\<star> \<and> (fin ((\<not>F))) ) "

syntax
 "_ifthen_d" :: "[lift,lift] \<Rightarrow> lift" ("(if\<^sub>i _ then _  )"  [88,88] 87)
 "_while_d"  :: "[lift,lift] \<Rightarrow> lift" ("(while _ do _  )"  [88,88] 87)

syntax (ASCII)
 "_ifthen_d" :: "[lift,lift] \<Rightarrow> lift" ("(if\<^sub>i _ then _  )"  [88,88] 87)
 "_while_d"  :: "[lift,lift] \<Rightarrow> lift" ("(while _ do _  )"  [88,88] 87)

translations
 "_ifthen_d" \<rightleftharpoons> "CONST ifthen_d"
 "_while_d"  \<rightleftharpoons> "CONST while_d"

definition repeat_d :: "('a::world) formula  \<Rightarrow> 'a formula \<Rightarrow> 'a formula "
where "repeat_d F G \<equiv> LIFT(F;while (\<not> G) do F ) "

syntax
 "_repeat_d"  :: "[lift,lift] \<Rightarrow> lift" ("(repeat _ until _  )"  [88,88] 87)

syntax (ASCII)
 "_repeat_d"  :: "[lift,lift] \<Rightarrow> lift" ("(repeat _ until _  )"  [88,88] 87)

translations
 "_repeat_d" \<rightleftharpoons> "CONST repeat_d"


definition next_assign_d :: "('a::world,'b) stfun \<Rightarrow> ('a,'b) formfun \<Rightarrow> 'a formula "
where "next_assign_d v e \<equiv> LIFT( v$ = e)" 

definition prev_assign_d :: "('a::world,'b) stfun \<Rightarrow> ('a,'b) formfun \<Rightarrow> 'a formula "
where "prev_assign_d v e \<equiv> LIFT( v! = e)"  

definition always_eq_d  :: "('a::world,'b) stfun \<Rightarrow> ('a,'b) formfun \<Rightarrow> 'a formula "
where "always_eq_d v e \<equiv> \<lambda> s. s \<Turnstile> \<box>($v = e)"  

definition temporal_assign_d  :: "('a::world,'b) stfun \<Rightarrow> ('a,'b) formfun \<Rightarrow> 'a formula "
where "temporal_assign_d v e \<equiv> \<lambda> s. s \<Turnstile> !v = e" 

definition gets_d :: "('a::world,'b) stfun \<Rightarrow> ('a,'b) formfun \<Rightarrow> 'a formula "
where "gets_d v e \<equiv> \<lambda> s. s \<Turnstile> keep( temporal_assign_d v e)" 

definition stable_d :: "('a::world,'b) stfun \<Rightarrow> 'a formula "
where " stable_d v \<equiv> \<lambda> s. s \<Turnstile> gets_d v (current_val_d v) "

definition padded_d :: "('a::world,'b) stfun \<Rightarrow> 'a formula "
where " padded_d v \<equiv> \<lambda> s. s \<Turnstile> (stable_d v);skip \<or> empty" 

definition padded_temp_assign_d :: "('a::world,'b) stfun \<Rightarrow> ('a,'b) formfun \<Rightarrow> 'a formula "
where "padded_temp_assign_d v e \<equiv> \<lambda> s. s \<Turnstile> (temporal_assign_d v e) \<and> (padded_d v)" 


syntax
 "_next_assign_d"        :: "[lift,lift] \<Rightarrow> lift" ("(_ := _)" [50,51] 50)
 "_prev_assign_d"        :: "[lift,lift] \<Rightarrow> lift" ("(_ =: _)" [50,51] 50)
 "_always_eq_d"          :: "[lift,lift] \<Rightarrow> lift" ("(_ \<approx> _)" [50,51] 50)
 "_temporal_assign_d"    :: "[lift,lift] \<Rightarrow> lift" ("(_ \<leftarrow> _)" [50,51] 50)
 "_gets_d"               :: "[lift,lift] \<Rightarrow> lift" ("(_ gets _)" [50,51] 50)
 "_stable_d"             :: "lift \<Rightarrow> lift"        ("(stable _)" [51] 50)
 "_padded_d"             :: "lift \<Rightarrow> lift"        ("(padded _)" [51] 50)
 "_padded_temp_assign_d" :: "[lift,lift] \<Rightarrow> lift" ("(_ <\<sim> _)" [50,51] 50)

syntax (ASCII)
 "_next_assign_d"        :: "[lift,lift] \<Rightarrow> lift" ("(_ := _)" [50,51] 50)
 "_prev_assign_d"        :: "[lift,lift] \<Rightarrow> lift" ("(_ =: _)" [50,51] 50)
 "_always_eq_d"          :: "[lift,lift] \<Rightarrow> lift" ("(_ alweqv _)" [50,51] 50)
 "_temporal_assign_d"    :: "[lift,lift] \<Rightarrow> lift" ("(_ <-- _)" [50,51] 50)
 "_gets_d"               :: "[lift,lift] \<Rightarrow> lift" ("(_ gets _)" [50,51] 50)
 "_stable_d"             :: "lift \<Rightarrow> lift"        ("(stable _)" [51] 50)
 "_padded_d"             :: "lift \<Rightarrow> lift"        ("(padded _)" [51] 50)
 "_padded_temp_assign_d" :: "[lift,lift] \<Rightarrow> lift" ("(_ <~ _)" [50,51] 50)

translations
 "_next_assign_d"        \<rightleftharpoons> "CONST next_assign_d"
 "_prev_assign_d"        \<rightleftharpoons> "CONST prev_assign_d"
 "_always_eq_d"          \<rightleftharpoons> "CONST always_eq_d"
 "_temporal_assign_d"    \<rightleftharpoons> "CONST temporal_assign_d"
 "_gets_d"               \<rightleftharpoons> "CONST gets_d"
 "_stable_d"             \<rightleftharpoons> "CONST stable_d"
 "_padded_d"             \<rightleftharpoons> "CONST padded_d"
 "_padded_temp_assign_d" \<rightleftharpoons> "CONST padded_temp_assign_d"

lemmas itl_def = skip_d_def chop_d_def current_val_d_def next_val_d_def fin_val_d_def penult_val_d_def
       sometimes_d_def di_d_def da_d_def next_d_def prev_d_def always_d_def bi_d_def ba_d_def wnext_d_def
       wprev_d_def more_d_def empty_d_def dm_d_def bm_d_def init_d_def fin_d_def halt_d_def initonly_d_def
       keep_d_def yields_d_def ifthenelse_d_def power_d_def len_d_def powerstar_d_def chopstar_d_def 
       ifthen_d_def while_d_def repeat_d_def next_assign_d_def prev_assign_d_def always_eq_d_def 
       temporal_assign_d_def gets_d_def stable_d_def padded_d_def padded_temp_assign_d_def
       

subsection \<open>Properties of Operators\<close>

text \<open> The following lemmas show that above operators have the expected semantics. \<close>

lemma skip_defs :
 "(w \<Turnstile> skip) = ( ilen w =1) "
by (simp add: itl_def)

lemma chop_defs :
 "(w \<Turnstile> F1 ; F2) = (\<exists> n . n\<le>ilen w \<and> ((prefix n w)\<Turnstile> F1) \<and> ( (suffix n w) \<Turnstile> F2)  ) "
by (simp add: itl_def)

lemma sometimes_defs : 
  "(w \<Turnstile> \<diamond> F) = (\<exists> n. n\<le>ilen w \<and> ((suffix n w) \<Turnstile> F))"
by (simp add: itl_def chop_defs)

lemma always_defs : 
 " (w \<Turnstile> \<box> F) =(\<forall> n. n\<le>ilen w \<longrightarrow> ((suffix n w) \<Turnstile> F))"
by (simp add: itl_def sometimes_defs)

lemma di_defs : 
  "(w \<Turnstile> di F) = (\<exists> n.  n\<le>ilen w \<and> ((prefix n w) \<Turnstile> F))"
by (simp add: itl_def chop_defs)

lemma bi_defs : 
  "(w \<Turnstile> bi F) = (\<forall> n. n\<le>ilen w \<longrightarrow> ((prefix n w) \<Turnstile> F))"
by (simp add: itl_def di_defs)

lemma da_defs : 
  "(w \<Turnstile> da F) = (\<exists> n na. n+na \<le> ilen w \<and> ((sub n (n+na) w) \<Turnstile> F))"
proof (auto simp add: itl_def chop_defs)
 show " \<And>n na.
       n \<le> ilen w \<Longrightarrow>
       na \<le> ilen w - n \<Longrightarrow> F (prefix na (suffix n w)) \<Longrightarrow> 
       \<exists>n na. n + na \<le> ilen w \<and> F (sub n (n + na) w) "
 by (metis  sub_def Nat.le_diff_conv2 add.commute add_diff_cancel_left')
 show " \<And>n na.
       n + na \<le> ilen w \<Longrightarrow> F (sub n (n + na) w) \<Longrightarrow> 
       \<exists>n\<le>ilen w. \<exists>na\<le>ilen w - n. F (prefix na (suffix n w)) "
 by (metis  sub_def add_leD1 ilen_sub prefix_ilen_bound suffix_ilen le_add1)
qed

lemma ba_defs :
 "(w \<Turnstile> ba F) = (\<forall> n na.  na+n \<le> ilen w \<longrightarrow> ((sub n (n+na) w) \<Turnstile> F))"
by (auto simp add: ba_d_def da_defs)

lemma next_defs :
 " (w \<Turnstile> \<circle> F) = (ilen w > 0 \<and> ((suffix 1 w)\<Turnstile> F) ) "
using Suc_le_eq min.absorb1 by (simp add: itl_def chop_defs skip_defs)  force

lemma wnext_defs :
 "(w \<Turnstile> wnext F) = (ilen w = 0 \<or> ((suffix 1 w)\<Turnstile> F) ) "
by (simp add: wnext_d_def next_defs)

lemma prev_defs :
 " (w \<Turnstile> prev F) = (ilen w > 0 \<and> ((prefix ((ilen w)-1) w)\<Turnstile> F) ) "
by (simp add: itl_def chop_defs skip_defs)
 (metis One_nat_def Suc_leI diff_diff_cancel diff_is_0_eq' diff_le_self 
        neq0_conv zero_neq_one)

lemma wprev_defs :
 " (w \<Turnstile> wprev F) = (ilen w = 0 \<or> ((prefix ((ilen w)-1) w)\<Turnstile> F) ) "
by (metis (mono_tags, lifting) less_le prev_defs unl_lift wprev_d_def zero_le)

lemma more_defs :
 " (w \<Turnstile> more) = (ilen w>0)"
by (simp add: more_d_def next_defs)

lemma empty_defs :
 " (w \<Turnstile> empty) = (ilen w = 0)"
by (simp add: empty_d_def more_defs)

lemma init_defs :
 " (w \<Turnstile> init F) = ( (prefix 0 w) \<Turnstile> F )"
using min.absorb1 by (simp add: init_d_def empty_defs chop_defs)  force

lemma initalt_defs :
 " (w \<Turnstile> bi( empty \<longrightarrow> F)) = ( (prefix 0 w) \<Turnstile> F )" 
  using min.absorb1 
  by (simp add: bi_defs empty_defs) 
  
lemma fin_defs :
 " (w \<Turnstile> fin F) = ( (suffix (ilen w) w) \<Turnstile> F)"
by (simp add: fin_d_def empty_defs always_defs) 

lemma finalt_defs :
 " (w \<Turnstile> #True;(F \<and> empty)) = ( (suffix (ilen w) w) \<Turnstile> F)"
by (simp add: chop_defs empty_defs) fastforce

lemma halt_defs :
 " (w \<Turnstile> halt(F)) = (\<forall>n\<le>ilen w. (ilen w = n) = F (suffix n w))"
by (simp add: halt_d_def empty_defs always_defs)

lemma initonly_defs :
 " (w \<Turnstile> initonly(F)) = (\<forall>n\<le>ilen w. (n = 0) = F (prefix n w))"
using min.absorb1 by (simp add: initonly_d_def bi_defs empty_defs) 

lemma ifthenelse_defs:
 " (w \<Turnstile> if\<^sub>i F then G else H) = 
   ( ((w \<Turnstile> F) \<and> (w \<Turnstile> G)) \<or> (( \<not>(w \<Turnstile> F) \<and> (w \<Turnstile> H))) )"
by (simp add: itl_def)

lemma len_defs :
 " (w \<Turnstile> len n) = (ilen w =n) "
proof 
 (induct n arbitrary: w)
 case 0
 then show ?case by (simp add: len_d_def empty_defs)
 next
 case (Suc n)
 then show ?case by (simp add: len_d_def chop_defs skip_defs) fastforce
qed


lemma currentval_defs :
 " (s \<Turnstile> $v) = (v (inth s 0))"
by (simp add: current_val_d_def)

lemma nextval_defs :
 "(s \<Turnstile> v$) = (if ilen s >0 then (v (inth s 1)) else (\<some> x. x=x)) "
by (simp add: itl_def)

lemma finval_defs :
 " (s \<Turnstile> !v) = (v (inth s (ilen s)))"
by (simp add: itl_def)

lemma penultval_defs :
 " (s \<Turnstile> v!) = (if ilen s >0 then (v (inth s ((ilen s)-1))) else (\<some> x. x=x))"
by (simp add: itl_def) 

lemma next_assign_defs :
assumes " ilen s >0"
shows   " (s \<Turnstile> v := e) = v ( inth s 1) = e s "
using assms by (auto simp: itl_def)

lemma prev_assign_defs :
assumes " ilen s >0 "
 shows  " (s \<Turnstile> v =: e) = v ( inth s ((ilen s)-1)) = e s "
using assms by (auto simp: itl_def)

lemma always_eqv_defs :
 " (s \<Turnstile> v \<approx> e) = (\<forall> i\<le> ilen s. v ( inth s i) = e (suffix i s))"
by (simp add: always_eq_d_def always_defs current_val_d_def)

lemma temporal_assign_defs :
 " (s \<Turnstile> v \<leftarrow> e) = (v ( inth s (ilen s)) = e s)"
by (simp add: itl_def)

lemma gets_defs :
 " (s \<Turnstile> v gets e) = (\<forall> i < ilen s. v ( inth s (Suc i)) = e (sub i (i+1) s) )"
using Suc_le_eq min.absorb1 add_le_imp_le_diff prefix_suffix_ilen_good
by (auto simp add: gets_d_def keep_d_def ba_defs skip_defs sub_def temporal_assign_defs) 

lemma stable_defs_helpa:
assumes "(\<forall>i<ilen s. v ( inth s (Suc i)) = v ( inth s i))"  
        " i \<le> ilen s" 
 shows  "(v ( inth s i) = v ( inth s 0))" 
using assms
proof (induct s arbitrary:i)
case (INil x)
then show ?case by simp
next
case (ICons x1a s)
then show ?case 
  proof (cases i)
  case 0
  then show ?thesis by blast
  next
  case (Suc nat)
  then show ?thesis 
     by (metis ICons.hyps ICons.prems(1) ICons.prems(2) Suc_le_mono Suc_mono inth_Suc 
         inth_zero ilen.simps(2) plus_1_eq_Suc zero_less_Suc)
  qed
qed


lemma stable_defs_helpb:
assumes "(\<forall>i\<le>ilen s. v ( inth s i) = v ( inth s 0))"  
        " i < ilen s" 
 shows  "v ( inth s (Suc i)) = v ( inth s i)" 
using assms
proof (induct s arbitrary:i)
case (INil x)
then show ?case by simp
next
case (ICons x1a s)
then show ?case 
   proof (cases i)
   case 0
   then show ?thesis using Suc_leI ICons.prems(1) ICons.prems(2) by blast
   next
   case (Suc nat)
   then show ?thesis using ICons.prems(1) ICons.prems(2) Suc_leI less_imp_le_nat by presburger
   qed
qed

lemma stable_defs_help:
 "(\<forall>i<ilen s. v ( inth s (Suc i)) = v ( inth s i)) \<longleftrightarrow> 
  (\<forall>i\<le>ilen s. v ( inth s i) = v ( inth s 0))"
proof -
 have 1: "(\<forall>i<ilen s. v ( inth s (Suc i)) = v ( inth s i)) \<longrightarrow>
          (\<forall>i\<le>ilen s. v ( inth s i) = v ( inth s 0)) " 
  using stable_defs_helpa by auto
 have 2: "(\<forall>i\<le>ilen s. v ( inth s i) = v ( inth s 0)) \<longrightarrow>
          (\<forall>i<ilen s. v ( inth s (Suc i)) = v ( inth s i))" 
   using stable_defs_helpb by blast
 show ?thesis using "1" "2" by blast
qed


lemma stable_defs:
 "(s \<Turnstile> stable v) = (\<forall> i\<le>ilen s. (v (inth s i)) = (v (inth s 0)))" 
by (simp add: stable_d_def gets_defs current_val_d_def sub_def stable_defs_help)

lemma padded_defs :
 " (s \<Turnstile> padded v) = ((\<forall> i< ilen s. (v (inth s i)) = (v (inth s 0))) \<or> ilen s =0) "
proof (simp add: padded_d_def stable_defs chop_defs skip_defs empty_defs)
  show " ((\<exists>n\<le>ilen s.
         (\<forall>i. i \<le> n \<and> i \<le> ilen s \<longrightarrow> v (inth s i) = v (inth s 0)) \<and> ilen s - n = Suc 0) \<or>
     ilen s = 0) =
    ((\<forall>i<ilen s. v ( inth s i) = v ( inth s 0)) \<or> ilen s = 0)"
  proof rule+
   show " \<And>i. (\<exists>n\<le>ilen s.
             (\<forall>i. i \<le> n \<and> i \<le> ilen s \<longrightarrow> v (inth s i) = v (inth s 0)) \<and> ilen s - n = Suc 0) \<or>
         ilen s = 0 \<Longrightarrow>
         i < ilen s \<Longrightarrow> v (inth s i) = v (inth s 0)"
   by (metis One_nat_def Suc_leI Suc_le_mono le_add_diff_inverse2 less_imp_le_nat not_less_zero
       plus_1_eq_Suc)
   show " (\<forall>i<ilen s. v (inth s i) = v (inth s 0)) \<or> ilen s = 0 \<Longrightarrow>
    (\<exists>n\<le>ilen s.
        (\<forall>i. i \<le> n \<and> i \<le> ilen s \<longrightarrow> v (inth s i) = v (inth s 0)) \<and> ilen s - n = Suc 0) \<or>
    ilen s = 0  "
   by (metis Suc_leI Suc_pred diff_diff_cancel diff_le_self gr_zeroI le_imp_less_Suc)
  qed
qed

lemma padded_temporal_assign_defs :
 " (s \<Turnstile> v <\<sim> e) = 
  ((s \<Turnstile> padded v) \<and> (v ( inth s (ilen s)) = e s ))"
by (auto simp add: padded_temp_assign_d_def padded_defs temporal_assign_defs)

lemmas itl_defs = skip_defs chop_defs sometimes_defs always_defs di_defs bi_defs da_defs ba_defs next_defs
       wnext_defs prev_defs wprev_defs more_defs empty_defs init_defs initalt_defs fin_defs finalt_defs 
       halt_defs initonly_defs ifthenelse_defs len_defs currentval_defs nextval_defs finval_defs 
       penultval_defs next_assign_defs prev_assign_defs always_eqv_defs temporal_assign_defs 
       gets_defs stable_defs padded_defs padded_temporal_assign_defs 

subsection \<open>Soundness of Finite ITL Axioms\<close>

subsubsection \<open>ChopAssoc\<close>

lemma ChopAssocSemHelpa:
assumes "(\<exists>i ia . i\<le>ilen \<sigma> \<and> ia \<le> ilen \<sigma> -i \<and> (prefix i \<sigma> \<Turnstile> f) \<and>
    (prefix ia (suffix i \<sigma>) \<Turnstile> g) \<and> (suffix (ia + i) \<sigma> \<Turnstile> h))" 
shows " (\<exists>j ja . j \<le>ilen \<sigma> \<and> ja \<le> j \<and> (prefix ja (prefix j \<sigma>) \<Turnstile> f) \<and> 
    (suffix ja (prefix j \<sigma>) \<Turnstile> g) \<and> (suffix j \<sigma> \<Turnstile> h))" 
proof -
 have 1: "(\<exists>i ia . i\<le>ilen \<sigma> \<and> ia \<le> ilen \<sigma> -i \<and> (prefix i \<sigma> \<Turnstile> f) \<and>
    (prefix ia (suffix i \<sigma>) \<Turnstile> g) \<and> (suffix (ia + i) \<sigma> \<Turnstile> h))" 
 using assms by auto
 obtain i ia where 2: "i\<le>ilen \<sigma> \<and> ia \<le> ilen \<sigma> -i \<and> (prefix i \<sigma> \<Turnstile> f) \<and>
    (prefix ia (suffix i \<sigma>) \<Turnstile> g) \<and> (suffix (ia + i) \<sigma> \<Turnstile> h)" 
   using "1" by auto
 have 3: "(suffix (ia+i) \<sigma> \<Turnstile> h) "
    using "2" by auto
 have 4: "ia +i \<le>ilen \<sigma> " 
    using "2" Nat.le_diff_conv2 by blast
 have 5: " i \<le> ia+i"
    by simp 
 have 6: "(suffix i (prefix (ia +i) \<sigma>) \<Turnstile> g)" 
    using "2" "4" suffix_prefix_swap by force
 have 7: "(prefix i (prefix (ia +i) \<sigma>) \<Turnstile> f)" 
    by (simp add: "2" add.commute)
 show ?thesis using "2" "4" "5" "6" "7" by blast
qed

lemma ChopAssocSemHelpb:
assumes " (\<exists>j ja . j \<le>ilen \<sigma> \<and> ja \<le> j \<and> (prefix ja (prefix j \<sigma>) \<Turnstile> f) \<and> 
    (suffix ja (prefix j \<sigma>) \<Turnstile> g) \<and> (suffix j \<sigma> \<Turnstile> h))" 
shows "(\<exists>i ia . i\<le>ilen \<sigma> \<and> ia \<le> ilen \<sigma> -i \<and> (prefix i \<sigma> \<Turnstile> f) \<and>
    (prefix ia (suffix i \<sigma>) \<Turnstile> g) \<and> (suffix (ia + i) \<sigma> \<Turnstile> h))" 
proof -
 have 1: "(\<exists>j ja . j \<le>ilen \<sigma> \<and> ja \<le> j \<and> (prefix ja (prefix j \<sigma>) \<Turnstile> f) \<and> 
    (suffix ja (prefix j \<sigma>) \<Turnstile> g) \<and> (suffix j \<sigma> \<Turnstile> h))"
  using assms by auto
 obtain j ja where 2: "j \<le>ilen \<sigma> \<and> ja \<le> j \<and> (prefix ja (prefix j \<sigma>) \<Turnstile> f) \<and> 
    (suffix ja (prefix j \<sigma>) \<Turnstile> g) \<and> (suffix j \<sigma> \<Turnstile> h) "  
 using "1" by auto
 have 3: "ja\<le>ilen \<sigma>" 
   using "2" le_trans by blast
 have 4: "j -ja \<le> ilen \<sigma> -ja" 
   by (simp add: "2" diff_le_mono)
 have 5: "(prefix ja \<sigma> \<Turnstile> f)" 
    by (metis "2" pref_pref_3 le_add_diff_inverse)
 have 6: "(prefix (j -ja) (suffix ja \<sigma>) \<Turnstile> g)" 
   by (simp add: "2" suffix_prefix_swap)
 have 7: "(suffix ((j -ja) + ja) \<sigma> \<Turnstile> h)" 
   by (simp add: "2")
 show ?thesis using "3" "4" "5" "6" "7" by blast
qed

lemma ChopAssocSemHelp:
 "(\<exists>i ia . i\<le>ilen \<sigma> \<and> ia \<le> ilen \<sigma> -i \<and> (prefix i \<sigma> \<Turnstile> f) \<and>
    (prefix ia (suffix i \<sigma>) \<Turnstile> g) \<and> (suffix (ia + i) \<sigma> \<Turnstile> h)) =
  (\<exists>j ja . j \<le>ilen \<sigma> \<and> ja \<le> j \<and> (prefix ja (prefix j \<sigma>) \<Turnstile> f) \<and> 
    (suffix ja (prefix j \<sigma>) \<Turnstile> g) \<and> (suffix j \<sigma> \<Turnstile> h))" 
using ChopAssocSemHelpa[of "\<sigma>" "f" "g" "h"] 
      ChopAssocSemHelpb[of "\<sigma>" "f" "g" "h"] by auto 

lemma ChopAssocSemHelp2:
 " (\<sigma> \<Turnstile>  f ; (g ; h)) = (\<sigma> \<Turnstile> (f;g);h)"
proof -
 have " (\<sigma> \<Turnstile>  f ; (g ; h)) =
        ((\<exists>i\<le>ilen \<sigma>. (prefix i \<sigma> \<Turnstile> f) \<and> (\<exists>ia\<le>ilen (suffix i \<sigma>). 
          (prefix ia (suffix i \<sigma>) \<Turnstile> g) \<and> (suffix (ia + i) \<sigma> \<Turnstile> h))))" 
 by (simp add: chop_defs)
 also have " ... = 
             (\<exists>i ia . i\<le>ilen \<sigma> \<and> ia \<le> ilen \<sigma> -i \<and> (prefix i \<sigma> \<Turnstile> f) \<and>
         (prefix ia (suffix i \<sigma>) \<Turnstile> g) \<and> (suffix (ia + i) \<sigma> \<Turnstile> h))" 
 by fastforce
 also have " ... =
            (\<exists>j ja . j \<le>ilen \<sigma> \<and> ja \<le> j \<and> (prefix ja (prefix j \<sigma>) \<Turnstile> f) \<and>
            (suffix ja (prefix j \<sigma>) \<Turnstile> g) \<and> (suffix j \<sigma> \<Turnstile> h))" 
 using ChopAssocSemHelp[of "\<sigma>" f "g" "h"]  by blast
 also have " ... =
            (\<exists>i\<le>ilen \<sigma>. (\<exists>ia\<le>ilen (prefix i \<sigma>).  (prefix ia (prefix i \<sigma>) \<Turnstile> f) \<and>
             (suffix ia (prefix i \<sigma>) \<Turnstile> g)) \<and> (suffix i \<sigma> \<Turnstile> h)) " 
 by fastforce
 also have " ... =
             (\<sigma> \<Turnstile> (f;g);h)" by (simp add: chop_defs)
 finally show "(\<sigma> \<Turnstile>  f ; (g ; h)) = (\<sigma> \<Turnstile> (f;g);h) " .
qed

lemma ChopAssocSem:
  " (\<sigma> \<Turnstile>  f ; (g ; h) =  (f;g);h)"
using ChopAssocSemHelp2 using unl_lift2 by blast

subsubsection \<open>OrChopImp\<close>

lemma OrChopImpSem:
  " (\<sigma> \<Turnstile>  ( f \<or> g);h   \<longrightarrow>  f;h \<or> g;h ) "
by (simp add: itl_defs) blast

subsubsection \<open>ChopOrImp\<close>

lemma ChopOrImpSem:
  " (\<sigma> \<Turnstile> f;(g \<or> h) \<longrightarrow>   f;g \<or>  f;h ) "
by (simp add: itl_defs) blast

subsubsection \<open>EmptyChop\<close>

lemma EmptyChopSem:
  " (\<sigma> \<Turnstile> empty ; f = f ) "
using min.absorb1 by (simp add: itl_defs) force

subsubsection \<open>ChopEmpty\<close>

lemma ChopEmptySem:
  " (\<sigma> \<Turnstile> f;empty = f ) "
by (simp add: itl_defs) auto

subsubsection \<open>StateImpBi\<close>

lemma StateImpBiSem:
  " (\<sigma> \<Turnstile> init f \<longrightarrow>   bi (init f) ) "
by (simp add: itl_defs)

subsubsection \<open>NextImpNotNextNot\<close>

lemma NextImpNotNextNotSem:
  " (\<sigma> \<Turnstile> \<circle> f \<longrightarrow>  \<not> (\<circle> (\<not> f)) ) "
by (simp add: itl_defs) 

subsubsection \<open>BiBoxChopImpChop\<close>

lemma BiBoxChopImpChopSem:
  " (\<sigma> \<Turnstile> bi ( f \<longrightarrow> f1) \<and> \<box>(g \<longrightarrow> g1) \<longrightarrow> f;g \<longrightarrow> f1;g1 ) "
by (simp add: itl_defs) fastforce

subsubsection \<open>BoxInduct\<close>

lemma box_induct_help_1 :
assumes  "(f \<sigma>)"
         "  (\<forall>i. Suc 0 \<le> ilen \<sigma> - i \<longrightarrow>
         i \<le> ilen \<sigma> \<longrightarrow> (suffix i \<sigma> \<Turnstile> f) \<longrightarrow> f (suffix (Suc i) \<sigma> ) ) "
shows    " (\<forall> j. j \<le> ilen \<sigma> \<longrightarrow> f (suffix j \<sigma> )) "
proof
    fix j
    show " j \<le> ilen \<sigma> \<longrightarrow> f (suffix j \<sigma>)" 
     using assms 
     proof
       (induct j arbitrary: \<sigma>)
       case 0
       then show ?case by simp
       next
       case (Suc j)
       then show ?case 
       by (metis Nat.le_diff_conv2 One_nat_def Suc_eq_plus1_left Suc_leD)
       qed
qed

lemma BoxInductSem:
  " (\<sigma> \<Turnstile> \<box> (f \<longrightarrow> wnext f) \<and> f \<longrightarrow> \<box> f)"
proof -
 have 1: "(\<sigma> \<Turnstile> \<box> (f \<longrightarrow> wnext f) \<and> f \<longrightarrow> \<box> f) =
          ((\<forall>n\<le>ilen \<sigma>. f (suffix n \<sigma>) \<longrightarrow> ilen \<sigma> = n \<or> f (suffix (Suc n) \<sigma>)) \<and> f \<sigma> \<longrightarrow> 
           (\<forall>n\<le>ilen \<sigma>. f (suffix n \<sigma>)))"
       by (simp add: always_defs wnext_defs)
 from 1 show ?thesis using box_induct_help_1[of "f" "\<sigma>"] by fastforce    
qed
 
subsubsection \<open>ChopStarEqv\<close>

lemma ChopExist:
 "\<turnstile>  (\<exists> k. f;g k) = f;(\<exists> k. g k)"
by (auto simp add: itl_defs Valid_def)

lemma ExistChop:
 "\<turnstile> (\<exists> k. (g k);f) =  (\<exists> k. g k);f"
by (auto simp add: itl_defs Valid_def)

lemma powersem1:
 "(\<sigma> \<Turnstile>  (\<exists> k. power f k) = ( empty \<or> (\<exists> k. power f (Suc k)))) "
proof auto
  show " \<And>x. \<sigma> \<Turnstile> (power f x)  \<Longrightarrow> \<forall>k. \<not>(\<sigma> \<Turnstile> (f;power f k) )  \<Longrightarrow> \<sigma> \<Turnstile> empty  "
  by (metis not0_implies_Suc pow_0 pow_Suc)
  show " \<sigma> \<Turnstile> empty  \<Longrightarrow> \<exists>x. \<sigma> \<Turnstile> (power f x)  "
  by (metis pow_0)
  show " \<And>k. \<sigma> \<Turnstile> (f;power f k)  \<Longrightarrow> \<exists>x. \<sigma> \<Turnstile> (power f x)   " 
  by (metis pow_Suc)
qed

lemma powersem:
 " \<turnstile> (\<exists> k. power f k) = (  empty \<or> (f);(\<exists> k. (power f k)))"
proof -
 have 1: "\<turnstile> (\<exists> k. power f k) = ( empty \<or> (\<exists> k. power f (Suc k))) "
 using powersem1 by blast
 have 2: "\<turnstile> (\<exists> k. power f (Suc k)) = (\<exists> k. (f);power f k) "
 by simp
 have 3: "\<turnstile> (\<exists> k. (f);(power f k)) = (f);(\<exists> k. (power f k)) " 
 using ChopExist by blast
 from 1 2 3 show ?thesis by fastforce
qed

lemma PowerstarEqvSem:
  " (\<sigma> \<Turnstile> (powerstar f) =  (empty \<or>  f;(powerstar f) ))"
proof -
 have 1: "(\<sigma> \<Turnstile> (powerstar f)) = 
          ( \<sigma> \<Turnstile> (\<exists> k. power f k)) "
 by (simp add: powerstar_d_def)
 have 2: "( \<sigma> \<Turnstile> (\<exists> k. power f k)) =
          (\<sigma> \<Turnstile> (empty \<or> f;(\<exists> k. (power f k))))"
 using powersem by (metis inteq_reflection)
 from 1 2 show ?thesis by (simp add: powerstar_d_def)
qed
 
lemma ChopstarEqvSem:
  " (\<sigma> \<Turnstile> f\<^sup>\<star> =  (empty \<or>  (f \<and> more); f\<^sup>\<star>) )"
by (metis PowerstarEqvSem chopstar_d_def)


subsection \<open>Quantification over State (Flexible) Variables\<close>

text \<open>
  The hidden state approach, as used in the embedding of TLA in 
  Isabelle/HOL TLA embedding \cite{Merz98,GrovMerz}, is used. 
  Here \cite{Merz98,GrovMerz}, a state space is defined by its projections, 
  and everything else is
  unknown. Thus, a variable is a projection of the state space, and has the same
  type as a state function. Moreover, strong typing is achieved, since the projection
  function may have any result type. To achieve this, the state space is represented
  by an undefined type, which is an instance of the \<open>world\<close> class to enable
  use with the \<open>Intensional\<close> theory.
\<close>

(*
typedecl state

instance state :: world ..
*)

type_synonym 'a statefun = "(state,'a) stfun"
type_synonym statepred   = "bool statefun"
type_synonym 'a tempfun  = "(state,'a) formfun"
type_synonym temporal    = "state formula"

text \<open> 
  Similar to \cite{Merz98,GrovMerz} we define a
  state to be an anonymous type whose only purpose is to provide Skolem IConstants.
  Similarly, we do not define a type of state variables separate from that
  of arbitrary state functions, again in order to simplify the definition
  of flexible quantification later on. Nevertheless, we need to distinguish
  state variables. Note we deviate from \cite{Merz98,GrovMerz} in that we do not use 
  axioms but use definitions and lemmas.
\<close>


subsection \<open>Temporal Quantifiers\<close>

definition exist_state_d :: " ('a statefun \<Rightarrow>  temporal )\<Rightarrow> temporal" (binder "Eex " 10)
where "exist_state_d  F  \<equiv> (\<lambda>s.  (\<exists> x.  s \<Turnstile> F x ))"

syntax
 "_Eex" :: "[idts, lift] \<Rightarrow> lift"      ("(3\<exists>\<exists> _./ _)" [0,10] 10)

translations
 "_Eex v A"  == "Eex v. A" 

definition forall_state_d :: " ('a statefun \<Rightarrow>  temporal )\<Rightarrow> temporal" (binder "Aall " 10)
where "forall_state_d F \<equiv> LIFT(\<not>(\<exists>\<exists> x. \<not>(F x)))"

syntax 
 "_Aall" :: "[idts, lift] \<Rightarrow> lift"      ("(3\<forall>\<forall> _./ _)" [0,10] 10)

translations
 "_Aall v A" == "Aall v. A"



end
