(*  title     : An encoding of ITL in Isabelle/HOL
    Authors   : Antonio Cau     <cau.researcher at gmail.com>
                Ben Moszkowski
                David Smallwood <drs at dmu.ac.uk>
    Maintainer: Antonio Cau     <cau.researcher at gmail.com>        
    License   : BSD
*)

section \<open>Time Reversal\<close>

theory TimeReversal
imports
 Theorems FOTheorems
begin

text \<open>
 Time reversal operator is defined in \cite{moszkowski14:_compos}. 
\<close>

subsection \<open>Definition\<close>

definition reverse_d :: "('a::world, 'b) formfun \<Rightarrow> ('a, 'b) formfun"
where "reverse_d F \<equiv> \<lambda> s. irev s \<Turnstile> F" 

syntax
 "_reverse_d"      :: "lift \<Rightarrow> lift"        ("(_\<^sup>r)" [85] 85)

syntax (ASCII)
 "_reverse_d"      :: "lift \<Rightarrow> lift"        ("(reverse _)" [85] 85)

translations
 "_reverse_d"     \<rightleftharpoons> "CONST reverse_d" 


subsection \<open>Time reversal Rules\<close>

lemma EExRev :
 "\<turnstile> (\<exists>\<exists> x. F x)\<^sup>r = (\<exists>\<exists> x. (F x)\<^sup>r)"
by (simp add: Valid_def exist_state_d_def reverse_d_def)

lemma rev_const :
 "\<turnstile> (#c)\<^sup>r = #c"
by (auto simp: reverse_d_def)

lemma rev_fun1 :
 "\<turnstile> (f<x>)\<^sup>r = f<x\<^sup>r>"
by (auto simp: reverse_d_def)

lemma rev_fun2:
 "\<turnstile> (f<x,y>)\<^sup>r = f<x\<^sup>r,y\<^sup>r>"
by (auto simp: reverse_d_def)

lemma rev_fun3:
 "\<turnstile> (f<x,y,z>)\<^sup>r = f<x\<^sup>r,y\<^sup>r,z\<^sup>r>"
by (auto simp: reverse_d_def)

lemma rev_forall:
 "\<turnstile> (\<forall> x. P x)\<^sup>r = (\<forall> x. (P x)\<^sup>r)"
by (auto simp: reverse_d_def)

lemma rev_exists:
 "\<turnstile> (\<exists> x. P x)\<^sup>r = (\<exists> x. (P x)\<^sup>r)"
by (auto simp: reverse_d_def)

lemma rev_exists1:
 "\<turnstile> (\<exists>! x. P x)\<^sup>r = (\<exists>! x. (P x)\<^sup>r)"
by (auto simp: reverse_d_def)

lemma rev_current:
 "\<turnstile> ($v)\<^sup>r = (!v)"
by (auto simp: irev_inth itl_def reverse_d_def) 

lemma rev_next:
 "\<turnstile>  (v$)\<^sup>r = (v!)"
by (auto simp: irev_inth itl_def reverse_d_def) 

lemma rev_penult:
 "\<turnstile>  (v!)\<^sup>r = (v$)"
by (auto simp: irev_inth itl_def reverse_d_def) 

lemma rev_fin:
 "\<turnstile> (!v)\<^sup>r = ($v)"
by (auto simp: irev_inth itl_def reverse_d_def) 

lemma EqvReverseReverse:
 "\<turnstile> (f\<^sup>r)\<^sup>r = f"
by (simp add: Valid_def reverse_d_def)

lemma ReverseEqv:
  "(\<turnstile> f) \<longleftrightarrow> (\<turnstile> f\<^sup>r)"
by (metis Valid_def irev_swap reverse_d_def)

lemma RevSkip:
 "\<turnstile> skip\<^sup>r = skip"
by (simp add: Valid_def reverse_d_def skip_defs)

lemma RevChop:
"\<turnstile> (f;g)\<^sup>r = (g\<^sup>r;f\<^sup>r)"
proof (auto simp add: Valid_def itl_def reverse_d_def)
 show "\<And>w n. n \<le> ilen w \<Longrightarrow>
           f (prefix n (irev w)) \<Longrightarrow>
           g (suffix n (irev w)) \<Longrightarrow> 
       \<exists>n\<le>ilen w. g (irev (prefix n w)) \<and> f (irev (suffix n w)) " 
 by (metis diff_diff_cancel irev_prefix irev_suffix suffix_ilen_bound suffix_ilen)
 show "\<And>w n. n \<le> ilen w \<Longrightarrow>
           g (irev (prefix n w)) \<Longrightarrow>
           f (irev (suffix n w)) \<Longrightarrow> 
       \<exists>n\<le>ilen w. f (prefix n (irev w)) \<and> g (suffix n (irev w))  " 
 by (metis irev_prefix irev_suffix suffix_ilen_bound suffix_ilen)
qed

lemma RMoreEqvMore:
 "\<turnstile> more\<^sup>r = more"
by (simp add: Valid_def itl_def reverse_d_def )

lemma REmptyEqvEmpty:
 "\<turnstile> empty\<^sup>r = empty"
by (metis RMoreEqvMore empty_d_def int_eq rev_fun1)

lemma PowerCommute:
 "\<turnstile> ((f \<and> more);(power (f \<and> more) n)) = (power (f \<and> more) n);(f \<and> more)"
proof 
 (induct n)
 case 0
 then show ?case by (metis ChopEmpty EmptyChop inteq_reflection pow_0)
 next
 case (Suc n)
 then show ?case by (metis ChopAssoc inteq_reflection pow_Suc)
qed

lemma REqvRule:
 assumes "\<turnstile> f =g"
 shows " \<turnstile> (f\<^sup>r) = (g\<^sup>r)"
using assms 
using inteq_reflection by force

lemma RevPowerChop:
 "\<turnstile> (power (f \<and> more)  n)\<^sup>r = (power ((f \<and> more)\<^sup>r) n)"
proof 
 (induct n)
 case 0
 then show ?case using REmptyEqvEmpty by auto
 next
 case (Suc n)
 then show ?case 
 by (metis PowerCommute RevChop inteq_reflection pow_Suc)
qed

lemma RevChopstar:
 "\<turnstile> (f\<^sup>\<star>)\<^sup>r = (f\<^sup>r)\<^sup>\<star>"
proof -
 have 1: "\<turnstile> (f\<^sup>\<star>) = (\<exists>n. power (f \<and> more) n)" 
      by (simp add: chopstar_d_def powerstar_d_def Valid_def) 
 have 2: "\<turnstile> (f\<^sup>\<star>)\<^sup>r = (\<exists> n. power (f \<and> more) n)\<^sup>r " 
      using REqvRule "1" by blast
 have 3: "\<turnstile> (\<exists> n. power (f \<and> more) n)\<^sup>r = (\<exists> n. (power (f \<and> more) n)\<^sup>r)" 
      by (simp add: rev_exists)
 have 4: "\<turnstile> (\<exists> n. (power (f \<and> more) n)\<^sup>r) = (\<exists> n. (power ((f \<and> more)\<^sup>r) n))" 
      by (simp add: RevPowerChop ExEqvRule)  
 have 5: "\<turnstile> (f \<and> more)\<^sup>r = (f\<^sup>r \<and> more)"
  by (metis RMoreEqvMore inteq_reflection rev_fun2)
 hence 6: "\<turnstile> (\<exists> n. (power ((f \<and> more)\<^sup>r) n)) = (\<exists> n. (power ((f\<^sup>r \<and> more)) n))"
   by (metis "4" inteq_reflection)
 have 7: "\<turnstile> (\<exists> n. (power ((f\<^sup>r \<and> more)) n)) = (f\<^sup>r)\<^sup>\<star>" 
      by (simp add: chopstar_d_def powerstar_d_def Valid_def)
 from 2 3 4 6 7 show ?thesis by fastforce
qed

lemmas all_rev = rev_const rev_fun1 rev_fun2 rev_fun3 rev_forall rev_exists
  rev_exists1 rev_current rev_next rev_penult rev_fin RevSkip RevChop RevChopstar

lemmas all_rev_unl = all_rev[THEN intD]
lemmas all_rev_eq = all_rev[THEN inteq_reflection]


subsection \<open>Properties of Time Reversal\<close>

lemma RNot:
 "\<turnstile> (\<not>f)\<^sup>r = (\<not> f\<^sup>r)"
by (simp add: rev_fun1)

lemma RRNot:
 "\<turnstile> (\<not>(f\<^sup>r))\<^sup>r = (\<not>f)"
by (metis EqvReverseReverse int_eq rev_fun1)

lemma RTrue:
 "\<turnstile> (#True)\<^sup>r = #True"
using rev_const by fastforce

lemma ROr:
 "\<turnstile> (f \<or> g)\<^sup>r = (f\<^sup>r \<or> g\<^sup>r)"
by (simp add: rev_fun2)

lemma RROr:
 "\<turnstile> (f\<^sup>r \<or> g\<^sup>r)\<^sup>r = (f \<or> g)"
proof -
 have 1: "\<turnstile> (f\<^sup>r \<or> g\<^sup>r)\<^sup>r = ((f\<^sup>r)\<^sup>r \<or> (g\<^sup>r)\<^sup>r)" using ROr by blast
 have 2: "\<turnstile> ((f\<^sup>r)\<^sup>r \<or> (g\<^sup>r)\<^sup>r) = (f \<or> g)" using EqvReverseReverse by (metis inteq_reflection)
 from 1 2 show ?thesis by fastforce
qed

lemma RAnd:
 " \<turnstile> (f \<and> g)\<^sup>r = (f\<^sup>r \<and> g\<^sup>r)"
by (simp add: rev_fun2)

lemma RRAnd:
 " \<turnstile> (f\<^sup>r \<and> g\<^sup>r)\<^sup>r = (f \<and> g)"
proof -
 have 1: "\<turnstile> (f\<^sup>r \<and> g\<^sup>r)\<^sup>r = ((f\<^sup>r)\<^sup>r \<and> (g\<^sup>r)\<^sup>r)" using RAnd by blast
 have 2: "\<turnstile> ((f\<^sup>r)\<^sup>r \<and> (g\<^sup>r)\<^sup>r) = (f \<and> g)" using EqvReverseReverse by (metis inteq_reflection)
 from 1 2 show ?thesis by fastforce
qed

lemma RImpRule:
 assumes "\<turnstile> f \<longrightarrow> g"
 shows   "\<turnstile> f\<^sup>r \<longrightarrow> g\<^sup>r"
using assms by (simp add: Valid_def reverse_d_def)


lemma RAndEmptyEqvAndEmpty:
 "\<turnstile> (f \<and> empty)\<^sup>r = (f \<and> empty)"
by (simp add: Valid_def itl_defs reverse_d_def, metis INil_ilen irev.simps(1))

lemma RNextEqvPrev:
 "\<turnstile> (\<circle> f)\<^sup>r = prev (f\<^sup>r)"
by (metis RevChop RevSkip inteq_reflection next_d_def prev_d_def)

lemma RRNextEqvPrev:
 "\<turnstile> (\<circle> (f\<^sup>r))\<^sup>r = prev (f)"
proof -
 have 1: "\<turnstile> (\<circle> (f\<^sup>r))\<^sup>r = prev ((f\<^sup>r)\<^sup>r)" using RNextEqvPrev by blast
 have 2: "\<turnstile> prev ((f\<^sup>r)\<^sup>r) = prev f" using EqvReverseReverse  by (metis inteq_reflection)
 from 1 2 show ?thesis by fastforce
qed

lemma RWNextEqvWPrev:
 "\<turnstile> (wnext f)\<^sup>r = wprev(f\<^sup>r)"
by (simp add: all_rev_eq(12) all_rev_eq(13) all_rev_eq(2) next_d_def prev_d_def wnext_d_def 
    wprev_d_def)

lemma RRWNextEqvWPrev:
 "\<turnstile> (wnext (f\<^sup>r))\<^sup>r = wprev(f)"
proof -
 have 1: "\<turnstile> (wnext (f\<^sup>r))\<^sup>r = wprev ((f\<^sup>r)\<^sup>r)" using RWNextEqvWPrev by blast
 have 2: "\<turnstile> wprev ((f\<^sup>r)\<^sup>r) = wprev f" using EqvReverseReverse by (metis inteq_reflection)
 from 1 2 show ?thesis by fastforce
qed

lemma RPrevEqvNext:
"\<turnstile> (prev f)\<^sup>r = \<circle> (f\<^sup>r)"
by (metis RevChop RevSkip inteq_reflection next_d_def prev_d_def)


lemma RRPrevEqvNext:
"\<turnstile> (prev (f\<^sup>r))\<^sup>r = \<circle> (f)"
proof -
 have 1: "\<turnstile> (prev (f\<^sup>r))\<^sup>r = \<circle> ((f\<^sup>r)\<^sup>r)" using RPrevEqvNext by blast
 have 2: "\<turnstile> \<circle> ((f\<^sup>r)\<^sup>r) = \<circle> f" using EqvReverseReverse by (metis inteq_reflection)
 from 1 2 show ?thesis by fastforce
qed

lemma RWPrevEqvWNext:
 "\<turnstile> (wprev f)\<^sup>r = wnext(f\<^sup>r)"
by (metis EqvReverseReverse RRWNextEqvWPrev int_eq)

lemma RRWPrevEqvWNext:
 "\<turnstile> (wprev (f\<^sup>r))\<^sup>r = wnext(f)"
proof -
 have 1: "\<turnstile> (wprev (f\<^sup>r))\<^sup>r = wnext ((f\<^sup>r)\<^sup>r)" using RWPrevEqvWNext by blast
 have 2: "\<turnstile> wnext ((f\<^sup>r)\<^sup>r) = wnext f" using EqvReverseReverse by (metis inteq_reflection)
 from 1 2 show ?thesis by fastforce
qed

lemma RDiamondEqvDi:
 "\<turnstile> (\<diamond>f)\<^sup>r = di (f\<^sup>r)"
by (simp add: di_d_def sometimes_d_def, metis RevChop RTrue inteq_reflection)

lemma RRDiamondEqvDi:
 "\<turnstile> (\<diamond>(f\<^sup>r))\<^sup>r = di (f)"
proof -
 have 1: "\<turnstile> (\<diamond> (f\<^sup>r))\<^sup>r = di ((f\<^sup>r)\<^sup>r)" using RDiamondEqvDi by blast
 have 2: "\<turnstile> di ((f\<^sup>r)\<^sup>r) = di f" using EqvReverseReverse by (metis inteq_reflection)
 from 1 2 show ?thesis by fastforce
qed

lemma RBoxEqvBi:
 "\<turnstile> (\<box> f)\<^sup>r = bi (f\<^sup>r)"
by  (simp add: always_d_def bi_d_def, metis RDiamondEqvDi int_eq rev_fun1 )


lemma RRBoxEqvBi:
 "\<turnstile> (\<box> (f\<^sup>r))\<^sup>r = bi (f)"
proof -
 have 1: "\<turnstile> (\<box> (f\<^sup>r))\<^sup>r = bi ((f\<^sup>r)\<^sup>r)" using RBoxEqvBi by blast
 have 2: "\<turnstile> bi ((f\<^sup>r)\<^sup>r) = bi f" using EqvReverseReverse by (metis inteq_reflection)
 from 1 2 show ?thesis by fastforce
qed

lemma RDiEqvDiamond:
 "\<turnstile> (di f)\<^sup>r = \<diamond> (f\<^sup>r)"
by (simp add: di_d_def sometimes_d_def, metis RevChop RTrue inteq_reflection)

lemma RRDiEqvDiamond:
 "\<turnstile> (di (f\<^sup>r))\<^sup>r = \<diamond> (f)"
proof -
 have 1: "\<turnstile> (di (f\<^sup>r))\<^sup>r = \<diamond> ((f\<^sup>r)\<^sup>r)" using RDiEqvDiamond by blast
 have 2: "\<turnstile> \<diamond> ((f\<^sup>r)\<^sup>r) = \<diamond> f" using EqvReverseReverse by (metis inteq_reflection)
 from 1 2 show ?thesis by fastforce
qed

lemma RBiEqvBox:
 "\<turnstile> (bi f)\<^sup>r = \<box> (f\<^sup>r)"
by (simp add: always_d_def bi_d_def, metis RDiEqvDiamond rev_fun1 int_eq)

lemma RRBiEqvBox:
 "\<turnstile> (bi (f\<^sup>r))\<^sup>r = \<box> (f)"
proof -
 have 1: "\<turnstile> (bi (f\<^sup>r))\<^sup>r = \<box> ((f\<^sup>r)\<^sup>r)" using RBiEqvBox by blast
 have 2: "\<turnstile> \<box> ((f\<^sup>r)\<^sup>r) = \<box> f" using EqvReverseReverse by (metis inteq_reflection)
 from 1 2 show ?thesis by fastforce
qed

lemma RDaEqvDa:
 "\<turnstile> (da f)\<^sup>r = da(f\<^sup>r)"
proof -
 have 1: "\<turnstile> (#True;(f;#True))\<^sup>r = (f;#True)\<^sup>r; #True\<^sup>r" using RevChop by blast
 have 2: "\<turnstile> (f;#True)\<^sup>r; #True\<^sup>r =  (f;#True)\<^sup>r; #True" using RTrue RightChopEqvChop by blast
 have 3: "\<turnstile> (f;#True)\<^sup>r; #True = (#True\<^sup>r;f\<^sup>r);#True " by (simp add: RevChop LeftChopEqvChop)
 have 4: "\<turnstile> (#True\<^sup>r;f\<^sup>r);#True = (#True;f\<^sup>r);#True " by (metis "3" RTrue int_eq)
 have 5: "\<turnstile> (#True;f\<^sup>r);#True = #True;(f\<^sup>r;#True)" using ChopAssocB by blast
 have 6: "\<turnstile> (#True;(f;#True))\<^sup>r = #True;(f\<^sup>r;#True)" using "1" "2" "3" "4" "5" by fastforce
 from 6 show ?thesis by (simp add: da_d_def)
qed

lemma RRDaEqvDa:
 "\<turnstile> (da (f\<^sup>r))\<^sup>r = da(f)"
proof -
 have 1: "\<turnstile> (da (f\<^sup>r))\<^sup>r = da ((f\<^sup>r)\<^sup>r)" using RDaEqvDa by blast
 have 2: "\<turnstile> da ((f\<^sup>r)\<^sup>r) = da f" using EqvReverseReverse by (metis inteq_reflection)
 from 1 2 show ?thesis by fastforce
qed

lemma RBaEqvBa:
 "\<turnstile> (ba f)\<^sup>r = ba(f\<^sup>r)"
by (simp add: ba_d_def, metis  RDaEqvDa int_eq rev_fun1)


lemma RRBaEqvBa:
 "\<turnstile> (ba (f\<^sup>r))\<^sup>r = ba(f)"
proof -
 have 1: "\<turnstile> (ba (f\<^sup>r))\<^sup>r = ba ((f\<^sup>r)\<^sup>r)" using RBaEqvBa by blast
 have 2: "\<turnstile> ba ((f\<^sup>r)\<^sup>r) = ba f" using EqvReverseReverse by (metis inteq_reflection)
 from 1 2 show ?thesis by fastforce
qed

lemma ChopCsImpCSChop:
 "\<turnstile> f;f\<^sup>\<star> \<longrightarrow> f\<^sup>\<star>;f"
by (meson CSChopEqvChopOrRule CSChopEqvOrChopPlusChop ChopAssocB 
          ChopPlusElimWithoutMore EmptyYields Prop03 Prop04 Prop06)

lemma CSChopImpChopCS:
 "\<turnstile> f\<^sup>\<star>;f \<longrightarrow> f;f\<^sup>\<star>"
proof -
 have 1: "\<turnstile> (f\<^sup>r);(f\<^sup>r)\<^sup>\<star> \<longrightarrow> (f\<^sup>r)\<^sup>\<star>;(f\<^sup>r)" 
      using ChopCsImpCSChop by blast
 hence 2: "\<turnstile> ((f\<^sup>r);(f\<^sup>r)\<^sup>\<star> \<longrightarrow> (f\<^sup>r)\<^sup>\<star>;(f\<^sup>r) )\<^sup>r " 
      using ReverseEqv by blast
 have 3: "\<turnstile> (((f\<^sup>r);(f\<^sup>r)\<^sup>\<star> \<longrightarrow> (f\<^sup>r)\<^sup>\<star>;(f\<^sup>r) )\<^sup>r) = ( ((f\<^sup>r);(f\<^sup>r)\<^sup>\<star>)\<^sup>r \<longrightarrow> ((f\<^sup>r)\<^sup>\<star>;(f\<^sup>r))\<^sup>r)" 
      by (simp add: rev_fun2)
 have 4: "\<turnstile> ((f\<^sup>r);(f\<^sup>r)\<^sup>\<star>)\<^sup>r = ((f\<^sup>r)\<^sup>\<star> )\<^sup>r; (f\<^sup>r)\<^sup>r" 
      by (simp add: RevChop)
 have 5: "\<turnstile> ((f\<^sup>r)\<^sup>\<star> )\<^sup>r; (f\<^sup>r)\<^sup>r = ((f\<^sup>r)\<^sup>r)\<^sup>\<star>;(f\<^sup>r)\<^sup>r" 
      by (simp add: LeftChopEqvChop RevChopstar)
 have 6: "\<turnstile> (f\<^sup>r)\<^sup>r = f" 
      using EqvReverseReverse by blast
 have 7: "\<turnstile> ((f\<^sup>r)\<^sup>r)\<^sup>\<star>;(f\<^sup>r)\<^sup>r = f\<^sup>\<star>;f" 
      using "6" CSEqvCS ChopEqvChop by blast
 have 8: "\<turnstile> ((f\<^sup>r);(f\<^sup>r)\<^sup>\<star>)\<^sup>r = f\<^sup>\<star>;f" 
      using "7" "5" using "4" by fastforce
 have 9: "\<turnstile> ((f\<^sup>r)\<^sup>\<star>;(f\<^sup>r))\<^sup>r = (f\<^sup>r)\<^sup>r;((f\<^sup>r)\<^sup>\<star>)\<^sup>r" 
      by (simp add: RevChop)
 have 10: "\<turnstile> (f\<^sup>r)\<^sup>r;((f\<^sup>r)\<^sup>\<star>)\<^sup>r = (f\<^sup>r)\<^sup>r; ((f\<^sup>r)\<^sup>r)\<^sup>\<star>" 
      by (simp add: RevChopstar RightChopEqvChop)
 have 11: "\<turnstile> (f\<^sup>r)\<^sup>r; ((f\<^sup>r)\<^sup>r)\<^sup>\<star> = f;f\<^sup>\<star>" 
      using "6" ChopPlusEqvChopPlus by blast
 have 12: "\<turnstile> ((f\<^sup>r);(f\<^sup>r)\<^sup>\<star>)\<^sup>r = f;f\<^sup>\<star>" 
      using "9" "10" "11" by (metis "4" "5" ChopCsImpCSChop RImpRule int_eq int_iffI)
 from 2 3 8 12 show ?thesis by fastforce
qed

lemma CSChopEqvChopCS:
 "\<turnstile> f;f\<^sup>\<star> = f\<^sup>\<star>;f"
using ChopCsImpCSChop CSChopImpChopCS by fastforce

lemma TrueChopSkipEqvSkipChopTrue:
 "\<turnstile>  #True;skip = skip;#True"
proof -
 have 1: "\<turnstile>skip;skip\<^sup>\<star> = skip\<^sup>\<star>;skip " using CSChopEqvChopCS by blast
 have 2: "\<turnstile> skip\<^sup>\<star> = #True" using CSSkip by simp
 have 3: "\<turnstile> skip;skip\<^sup>\<star> = skip;#True" using "2" using RightChopEqvChop by blast
 have 4: "\<turnstile> skip\<^sup>\<star>;skip = #True;skip" using "2" using LeftChopEqvChop by blast
 from 1 3 4 show ?thesis by fastforce
qed

lemma RInitEqvFin:
 "\<turnstile> (init f)\<^sup>r = fin(f)"
proof -
 have 1: "\<turnstile> (init f)\<^sup>r = ((f \<and> empty);#True)\<^sup>r" 
      by (metis AndChopCommute REqvRule init_d_def)
 have 2: "\<turnstile> ((f \<and> empty);#True)\<^sup>r = (#True;(f \<and> empty)\<^sup>r)" 
      using RTrue by (metis RevChop int_eq)
 have 3: "\<turnstile> #True;(f \<and> empty)\<^sup>r = #True;(f\<^sup>r \<and> empty)" 
      by (metis RAnd REmptyEqvEmpty RightChopEqvChop int_eq)
 have 4: "\<turnstile> #True;(f\<^sup>r \<and> empty) = #True;(f \<and> empty)" 
     using RAndEmptyEqvAndEmpty 
     by (metis REmptyEqvEmpty RightChopEqvChop all_rev_eq(3) int_eq)
 have 5: "\<turnstile> #True;(f \<and> empty) = fin(f)"  
      using FinEqvTrueChopAndEmpty by fastforce
from 1 2 3 4 5 show ?thesis by fastforce
qed



lemma RFinEqvInit:
 "\<turnstile> (fin f)\<^sup>r = init (f)"
proof -
 have 1: "\<turnstile> fin f = #True;(f \<and> empty)" 
      using FinEqvTrueChopAndEmpty by auto
 have 2: "\<turnstile> (fin f)\<^sup>r = (#True;(f \<and> empty))\<^sup>r" 
      using "1" REqvRule by blast
 have 3: "\<turnstile> (#True;(f \<and> empty))\<^sup>r = (f \<and> empty)\<^sup>r;#True" 
      using RTrue by (metis RevChop int_eq)
 have 4: "\<turnstile> (f\<and> empty)\<^sup>r;#True = (f\<^sup>r \<and> empty);#True" 
      using LeftChopEqvChop RAnd REmptyEqvEmpty  by (metis int_eq)
 have 5: "\<turnstile> (f\<and> empty)\<^sup>r;#True = (f\<and> empty);#True"
      by (simp add: RAndEmptyEqvAndEmpty LeftChopEqvChop)
 have 6: "\<turnstile> (f \<and> empty);#True = init(f)" 
      by (simp add: AndChopCommute init_d_def)
 from 1 2 3 4 5 6 show ?thesis by fastforce
qed



lemma RHaltEqvInitonly:
 "\<turnstile> (halt f)\<^sup>r = initonly (f\<^sup>r) " 
proof -
 have 1: "\<turnstile> (halt f)\<^sup>r = (\<box> ( empty = f ))\<^sup>r" by (simp add: halt_d_def)
 have 2: "\<turnstile> (\<box> ( empty = f ))\<^sup>r = bi ( (empty = f)\<^sup>r)" by (simp add: RBoxEqvBi)
 have 3: "\<turnstile> (empty = f)\<^sup>r = (empty = f\<^sup>r)" by (metis REmptyEqvEmpty inteq_reflection rev_fun2)
 hence 4: "\<turnstile> bi ( (empty = f)\<^sup>r) = bi(empty = f\<^sup>r)" by (simp add: BiEqvBi)
 have 5: "\<turnstile> bi(empty = f\<^sup>r) = initonly(f\<^sup>r)" by (simp add: initonly_d_def)
 from 1 2 4 5 show ?thesis by fastforce
qed

lemma RInitonlyEqvHalt:
 "\<turnstile> (initonly f)\<^sup>r = halt(f\<^sup>r)" 
proof -
 have 1: "\<turnstile> (initonly f)\<^sup>r = (bi (empty = f))\<^sup>r" by (simp add: initonly_d_def)
 have 2: "\<turnstile> (bi (empty = f))\<^sup>r = \<box>((empty = f)\<^sup>r)" by (simp add: RBiEqvBox)
 have 3: "\<turnstile> (empty = f)\<^sup>r = (empty = f\<^sup>r)" by (metis REmptyEqvEmpty inteq_reflection rev_fun2)
 hence 4: "\<turnstile> \<box> ((empty = f)\<^sup>r) = \<box>(empty = f\<^sup>r)" by (simp add: BoxEqvBox)
 have 5: "\<turnstile> \<box>(empty = f\<^sup>r) = halt(f\<^sup>r)" by (simp add: halt_d_def)
 from 1 2 4 5 show ?thesis by fastforce
qed

lemma RRHaltEqvInitonly:
 "\<turnstile> (halt (f\<^sup>r))\<^sup>r = initonly (f) "
proof -
 have 1: "\<turnstile> (halt (f\<^sup>r))\<^sup>r = initonly ((f\<^sup>r)\<^sup>r)" using RHaltEqvInitonly by blast
 have 2: "\<turnstile> initonly ((f\<^sup>r)\<^sup>r) = initonly(f) " using EqvReverseReverse by (metis inteq_reflection)
 from 1 2 show ?thesis by fastforce
qed

lemma RRInitonlyEqvHalt :
 "\<turnstile> (initonly (f\<^sup>r))\<^sup>r = halt(f)"
proof -
 have 1: "\<turnstile> (initonly (f\<^sup>r))\<^sup>r = halt((f\<^sup>r)\<^sup>r)" using RInitonlyEqvHalt by blast
 have 2: "\<turnstile> halt((f\<^sup>r)\<^sup>r) = halt(f)" using EqvReverseReverse by (metis inteq_reflection)
 from 1 2 show ?thesis by fastforce
qed

lemma RKeepEqvKeep :
 "\<turnstile> (keep f)\<^sup>r = keep(f\<^sup>r)"
proof -
 have 1: "\<turnstile> (keep f)\<^sup>r = (ba(skip \<longrightarrow> f))\<^sup>r" by (simp add: keep_d_def)
 have 2: "\<turnstile> (ba(skip \<longrightarrow> f))\<^sup>r = ba((skip \<longrightarrow> f)\<^sup>r)" by (simp add:RBaEqvBa)
 have 3: "\<turnstile> (skip \<longrightarrow> f)\<^sup>r = (skip \<longrightarrow> f\<^sup>r)" by (metis all_rev_eq(12) rev_fun2)
 hence 4: "\<turnstile> ba((skip \<longrightarrow> f)\<^sup>r) = ba(skip \<longrightarrow> f\<^sup>r)" by (simp add: BaEqvBa)
 have 5: "\<turnstile> ba(skip \<longrightarrow> f\<^sup>r) = keep(f\<^sup>r)" by (simp add: keep_d_def)
 from 1 2 4 5 show ?thesis by fastforce
qed

lemma RRKeepEqvKeep :
 "\<turnstile> (keep (f\<^sup>r))\<^sup>r = keep(f)"
proof -
 have 1: "\<turnstile> (keep (f\<^sup>r))\<^sup>r = keep( (f\<^sup>r)\<^sup>r )" using RKeepEqvKeep by blast
 have 2: "\<turnstile> keep( (f\<^sup>r)\<^sup>r ) = keep(f)" using EqvReverseReverse by (metis inteq_reflection)
 from 1 2 show ?thesis by fastforce
qed

lemma NextDiamondEqvDiamondNext:
 " \<turnstile> \<circle>( \<diamond> f) = \<diamond>(\<circle>f)"
proof -
 have  1: "\<turnstile> #True;skip = skip;#True  " by (rule TrueChopSkipEqvSkipChopTrue)
 hence 2: "\<turnstile> (#True;skip);f =( skip;#True);f " using LeftChopEqvChop by blast
 have  3: "\<turnstile> (#True;skip);f =  #True;(skip;f)  " by (simp add: ChopAssocB)
 have  4: "\<turnstile> ( skip;#True);f =  skip;(#True;f) "  by (simp add: ChopAssocB)
 from 2 3 4 show ?thesis by (metis int_eq next_d_def sometimes_d_def)
qed

lemma WeakNextBoxInduct:
 assumes "\<turnstile> wnext  (\<box> f) \<longrightarrow> f "
 shows   "\<turnstile> f "
proof -
 have   1: "\<turnstile> wnext  (\<box> f) \<longrightarrow> f" using assms by blast
 hence  2: "\<turnstile> \<not> f  \<longrightarrow> \<not> ( wnext  (\<box> f))  "  by fastforce 
 hence  3: "\<turnstile> \<not> f  \<longrightarrow>   \<circle>  (\<not> (\<box> f)) "  by (simp add: wnext_d_def) 
 have   4: "\<turnstile> (\<not> (\<box> f)) =   (\<diamond> (\<not> f)) " by (auto simp: always_d_def)
 hence  5: "\<turnstile> \<circle>(\<not> (\<box> f)) =  \<circle> (\<diamond> (\<not> f)) " using NextEqvNext by blast
 have   6: "\<turnstile> \<not> f  \<longrightarrow>   \<circle>  (\<diamond> (\<not> f)) " using "3" "5" by fastforce
 have   7: "\<turnstile> \<circle>  (\<diamond> (\<not> f)) = \<diamond>( \<circle> (\<not> f))  " using NextDiamondEqvDiamondNext by blast
 have   8: "\<turnstile> \<not> f  \<longrightarrow>   \<diamond>( \<circle> (\<not> f))  " using "6" "7" by fastforce
 have   9: "\<turnstile> \<diamond>(\<not> f)  \<longrightarrow>   \<diamond>(\<diamond>( \<circle> (\<not> f))) " using "8" DiamondImpDiamond by blast
 have  10: "\<turnstile> \<diamond>(\<diamond>( \<circle> (\<not> f))) = \<diamond>( \<circle> (\<not> f)) " using DiamondDiamondEqvDiamond by blast
 have  11: "\<turnstile> \<diamond>(\<not> f)  \<longrightarrow> \<diamond>( \<circle> (\<not> f)) " using "9" "10" by fastforce
 have  12: "\<turnstile> \<diamond>(\<not> f)  \<longrightarrow> \<circle> (\<diamond> (\<not> f)) " using "7" "11" by fastforce
 hence 13: "\<turnstile> \<not>( \<diamond>(\<not> f))"  using NextLoop by blast
 hence 14: "\<turnstile> \<box>f " by (simp add: always_d_def)
 have  15: "\<turnstile> \<box>f \<longrightarrow> f " using BoxElim by blast
 from 14 15 show ?thesis using MP by blast 
qed

lemma RassignEqvTAssign:
 "\<turnstile> ($v = e)\<^sup>r = (v \<leftarrow> e\<^sup>r)"
proof -
 have 1: "\<turnstile> ($v = e)\<^sup>r = (($v)\<^sup>r = e\<^sup>r)" by (simp add: rev_fun2)
 have 2: "\<turnstile> (($v)\<^sup>r = e\<^sup>r) = ((!v) = e\<^sup>r)" by (simp add: all_rev_eq(8))
 have 3: "\<turnstile> ((!v) = e\<^sup>r) = (v \<leftarrow> e\<^sup>r)" by (simp add: intI temporal_assign_d_def)
 from 1 2 3 show ?thesis by fastforce
qed

lemma RTAssignEqvAssign:
 "\<turnstile> (v \<leftarrow> e)\<^sup>r = ($v = e\<^sup>r)" 
proof -
 have 1: "\<turnstile> (v \<leftarrow> e)\<^sup>r = (!v = e)\<^sup>r" by (simp add: REqvRule intI temporal_assign_d_def)
 have 2: "\<turnstile> (!v = e)\<^sup>r = ($v = e\<^sup>r)" by (metis all_rev_eq(11) rev_fun2)
 from 1 2 show ?thesis by fastforce
qed

lemma RNextAssignEqvPrevAssign:
 "\<turnstile> (v := e)\<^sup>r = (v =: e\<^sup>r)"
proof -
 have 1: "\<turnstile> (v := e)\<^sup>r = (v$ = e)\<^sup>r" by (simp add: REqvRule intI next_assign_d_def)
 have 2: "\<turnstile> (v$ = e)\<^sup>r = (v! = e\<^sup>r)" by (metis all_rev_eq(9) rev_fun2)
 have 3: "\<turnstile> (v! = e\<^sup>r) = (v =: e\<^sup>r)" by (simp add: prev_assign_d_def)
 from 1 2 3 show ?thesis by fastforce
qed

lemma RPrevAssignEqvNextAssign:
 "\<turnstile> (v =: e)\<^sup>r = (v := e\<^sup>r)"
proof -
 have 1: "\<turnstile> (v =: e)\<^sup>r = (v! = e)\<^sup>r" by (simp add: REqvRule intI prev_assign_d_def)
 have 2: "\<turnstile> (v! = e)\<^sup>r = (v$ = e\<^sup>r)" by (metis all_rev_eq(10) rev_fun2)
 have 3: "\<turnstile> (v$ = e\<^sup>r) = (v := e\<^sup>r)" by (simp add: next_assign_d_def)
 from 1 2 3 show ?thesis by fastforce
qed

lemma RGetsEqvBaSkipImp:
 "\<turnstile> (v gets e)\<^sup>r = ba(skip \<longrightarrow> ($v = e\<^sup>r))"
proof -
 have 1: "\<turnstile> (v gets e)\<^sup>r = (ba(skip \<longrightarrow> (!v = e)))\<^sup>r" 
      using gets_d_def temporal_assign_d_def keep_d_def REqvRule 
      by (metis Prop04  ba_d_def int_simps(15))
 have 2: "\<turnstile> (ba(skip \<longrightarrow> (!v = e)))\<^sup>r = ba ( (skip \<longrightarrow> (!v = e))\<^sup>r )" 
      by (simp add: RBaEqvBa)
 have 3: "\<turnstile> (skip \<longrightarrow> (!v = e))\<^sup>r = (skip \<longrightarrow> ($v = e\<^sup>r))" 
      by (simp add: all_rev_eq(11) all_rev_eq(12) all_rev_eq(3))
 hence 4: "\<turnstile> ba ((skip \<longrightarrow> (!v = e))\<^sup>r) = ba (skip \<longrightarrow> ($v = e\<^sup>r))" 
      by (simp add: BaEqvBa)
 from 1 2 4 show ?thesis by fastforce
qed
     
lemma RIfThenElse:
 "\<turnstile> (if\<^sub>i f0 then f1 else f2)\<^sup>r = if\<^sub>i (f0\<^sup>r) then (f1\<^sup>r) else (f2\<^sup>r)"
by (simp add: all_rev_eq(2) all_rev_eq(3) ifthenelse_d_def)

lemma RWhile:
 "\<turnstile> (init f \<and> while f0 do f1)\<^sup>r = ( fin(f) \<and> ((f0\<^sup>r) \<and> (f1\<^sup>r))\<^sup>\<star> \<and> init (\<not>(f0)) )" 
proof -
 have 1: "\<turnstile> (init f \<and> while f0 do f1)\<^sup>r = ( init f \<and> (f0 \<and> f1)\<^sup>\<star> \<and> fin (\<not>f0) )\<^sup>r" 
       by (simp add: while_d_def)
 have 2: "\<turnstile> ( init f \<and> (f0 \<and> f1)\<^sup>\<star> \<and> fin (\<not>f0) )\<^sup>r = ((init f)\<^sup>r \<and> ((f0 \<and> f1)\<^sup>\<star>)\<^sup>r \<and> (fin (\<not>f0))\<^sup>r)"
       by (simp add: all_rev_eq(3))      
 have 3: "\<turnstile> (init f)\<^sup>r = fin(f)" 
       by (simp add: RInitEqvFin)
 have 4: "\<turnstile> ((f0 \<and> f1)\<^sup>\<star>)\<^sup>r = ((f0\<^sup>r) \<and> (f1\<^sup>r))\<^sup>\<star>" 
       by (metis RevChopstar all_rev_eq(3))
 have 5: "\<turnstile> (fin (\<not>f0))\<^sup>r = init (\<not>(f0))" 
       by (metis RFinEqvInit)
 have 6: "\<turnstile> ((init f)\<^sup>r \<and> ((f0 \<and> f1)\<^sup>\<star>)\<^sup>r \<and> (fin (\<not>f0))\<^sup>r) = 
            ( fin(f) \<and> ((f0\<^sup>r) \<and> (f1\<^sup>r))\<^sup>\<star> \<and> init (\<not>(f0)) ) " using "3" "4" "5" by fastforce
 from 1 2 6 show ?thesis by fastforce
qed

lemma AAxRev:
 "\<turnstile> (\<forall>\<forall> x. F x)\<^sup>r = (\<forall>\<forall> x. (F x)\<^sup>r)"
proof -
 have 1: "\<turnstile> (\<forall>\<forall> x. F x) = (\<not>(\<exists>\<exists> x. \<not>(F x)))" using AAxDef by blast
 have 2: "\<turnstile> (\<forall>\<forall> x. F x)\<^sup>r = (\<not>(\<exists>\<exists> x. \<not>(F x)))\<^sup>r" using REqvRule "1" by blast
 have 3: "\<turnstile> (\<not>(\<exists>\<exists> x. \<not>(F x)))\<^sup>r = (\<not>((\<exists>\<exists> x.(\<not> (F x)))\<^sup>r))" by (simp add: rev_fun1)
 have 4: "\<turnstile> ((\<exists>\<exists> x.(\<not> (F x)))\<^sup>r) = ((\<exists>\<exists> x.(\<not> (F x))\<^sup>r))" by (simp add: EExRev)
 hence 5: "\<turnstile> (\<not>((\<exists>\<exists> x.(\<not> (F x)))\<^sup>r)) = (\<not>(\<exists>\<exists> x.(\<not> (F x))\<^sup>r)) " by auto
 have 51: "\<And> x. \<turnstile> (\<not> (F x))\<^sup>r = (\<not>( (F x)\<^sup>r))" by (simp add: rev_fun1)
 hence 52: "\<turnstile> (\<exists>\<exists> x.(\<not> (F x))\<^sup>r) = (\<exists>\<exists> x.\<not>( (F x)\<^sup>r))" 
    using EExEqvRule[of "(\<lambda>x. LIFT( (\<not> (F x)))\<^sup>r)" "(\<lambda>x. LIFT((\<not>( (F x)\<^sup>r)))) " ] 
    by fastforce
 hence 6: "\<turnstile> (\<not>(\<exists>\<exists> x.(\<not> (F x))\<^sup>r)) = (\<not>(\<exists>\<exists> x.\<not>( (F x)\<^sup>r)))"   by fastforce
 have 7: "\<turnstile> (\<not>(\<exists>\<exists> x.\<not>( (F x)\<^sup>r))) = (\<forall>\<forall> x. (F x)\<^sup>r) " using AAxDef by fastforce
 from 1 2 3 5 6 7 show ?thesis by fastforce
qed

end